//==================================================
// FILE:	DigiDocGen.c
// PROJECT:     Digi Doc
// DESCRIPTION: DigiDoc helper routines for XML generation
// AUTHOR:  Ants Madisson, Business IT Partner
//==================================================
// Copyright (C) AS Sertifitseerimiskeskus
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// GNU Lesser General Public Licence is available at
// http://www.gnu.org/copyleft/lesser.html
//==========< HISTORY >=============================
//      18.02.2008      Ants Madisson
//                      Creation
//==================================================


#include <libdigidoc/DigiDocDefs.h>
#include <libdigidoc/DigiDocLib.h>
#include <libdigidoc/DigiDocDebug.h>
#include <libdigidoc/DigiDocConfig.h>
#include <libdigidoc/DigiDocConvert.h>
#include <libdigidoc/DigiDocCert.h>
#include <libdigidoc/DigiDocSAXParser.h>

#include <libdigidoc/DigiDocGen.h>
#include <libdigidoc/BDocGen.h>
#include <libdigidoc/DigiDocError.h>
#include <string.h>
#include <time.h>

#include <libdigidoc/BDocZip.h>

#include <libxml/globals.h>
#include <libxml/xmlerror.h>
#include <libxml/parser.h>
#include <libxml/parserInternals.h> /* only for xmlNewInputFromFile() */
#include <libxml/tree.h>
#include <libxml/debugXML.h>
#include <libxml/xmlmemory.h>
#include <libxml/c14n.h>


#include <fcntl.h>

#define DD_TEMP_FILE_MAX 200

//============================================================
// Adds a notary signature to signed document buffer
// pMBufXML - output buffer
// pSigDoc - signed doc pointer
// pNotInfo - notary signature info 
// returns error code
//============================================================
int addNotaryInfoXMLBDoc(DigiDocMemBuf *pMBufXML, const SignedDoc *pSigDoc, const SignatureInfo* pSigInfo, const NotaryInfo* pNotInfo)
{
  int err = ERR_OK, i;
  DigiDocMemBuf mbuf1;
  const DigiDocMemBuf *pMBuf;
  CertValue *pCertValue = 0;

  RETURN_IF_NULL_PARAM(pSigInfo);
  RETURN_IF_NULL_PARAM(pNotInfo);
  RETURN_IF_NULL_PARAM(pMBufXML);
  mbuf1.pMem = 0;
  mbuf1.nLen = 0;
  // unsigned prop begin
  ddocMemAppendData(pMBufXML, "<UnsignedProperties xmlns=\"", -1);
  ddocMemAppendData(pMBufXML,NAMESPACE_XADES_132,  -1);
  ddocMemAppendData(pMBufXML,"\">\n<UnsignedSignatureProperties>", -1);
#ifdef WITH_TS // it it is actually possible
  err = ddocSigInfo_SignatureTimeStamp_toXML(pSigInfo, &mbuf1);
  if(!err && mbuf1.pMem) {
    ddocMemAppendData(pMBufXML,(const char*)mbuf1.pMem, -1);
    ddocMemAppendData(pMBufXML,"\n", -1);
  }
  ddocMemBuf_free(&mbuf1);
#endif


  // <CompleteCertificateRefs>
  err = bdocCompleteCertificateRefs_toXML((SignedDoc*)pSigDoc, (SignatureInfo*)pSigInfo, &mbuf1);
  ddocMemAppendData(pMBufXML,(const char*)mbuf1.pMem, -1);
  ddocMemBuf_free(&mbuf1);
  // <CompleteRevocationRefs>
  err = ddocCompleteRevocationRefs_toXML((SignedDoc*)pSigDoc, (SignatureInfo*)pSigInfo, &mbuf1);
  ddocMemAppendData(pMBufXML,(const char*)mbuf1.pMem, -1);
  ddocMemBuf_free(&mbuf1);
  
  // if we want timestamp - run-time
#ifdef WITH_TS // it it is actually possible
  err = ddocSigInfo_SigAndRefsTimeStamp_toXML(pSigInfo, &mbuf1);
  //AM 17.11.08 added mbuf1.nLen > 0 so newline wont be added when timestamp isnt used in bdoc
	if(err == 0 && mbuf1.pMem) {
    ddocMemAppendData(pMBufXML,(const char*)mbuf1.pMem, -1);
    ddocMemAppendData(pMBufXML,"\n", -1);
  }
  ddocMemBuf_free(&mbuf1);
#endif

  // responder cert
  ddocMemAppendData(pMBufXML,"<CertificateValues>\n", -1);
  // TODO format cert without header
  for(i = 0; i < ddocCertValueList_GetCertValuesCount(pSigInfo->pCertValues); i++) {
    pCertValue = ddocCertValueList_GetCertValue(pSigInfo->pCertValues, i);
    if(pCertValue && pCertValue->nType != CERTID_VALUE_SIGNERS_CERT) {
      ddocCertValue_toXML(pCertValue, &mbuf1);
      ddocMemAppendData(pMBufXML, (const char*)mbuf1.pMem, -1);
      ddocMemAppendData(pMBufXML, "\n", -1); 
      ddocMemBuf_free(&mbuf1);
    }
  }
  ddocMemAppendData(pMBufXML, "</CertificateValues>\n", -1);
  // revocation values
  ddocMemAppendData(pMBufXML, "<RevocationValues>", -1);
  // Ver 1.76 - in format 1.3 we use here additionally element <OCSPValues>
  ddocMemAppendData(pMBufXML, "<OCSPValues>", -1);
  // OCSP response
  ddocMemAppendData(pMBufXML, "<EncapsulatedOCSPValue Id=\"", -1);
  ddocMemAppendData(pMBufXML, pNotInfo->szId, -1);
  ddocMemAppendData(pMBufXML, "\">", -1);	
  pMBuf = ddocNotInfo_GetOCSPResponse(pNotInfo);
  RETURN_IF_NULL(pMBuf);
  // fix for backward compatibility with 2.1.5
  // remove trailing \n after base64 content
  ddocEncodeBase64(pMBuf, &mbuf1);
	//AM 17.11.08 moved these 2 lines after ddocEncodeBase64
	if(mbuf1.pMem && ((char*)mbuf1.pMem)[strlen((const char*)mbuf1.pMem)-1] == '\n')
    ((char*)mbuf1.pMem)[strlen((const char*)mbuf1.pMem)-1] = 0;
  ddocMemAppendData(pMBufXML, (const char*)mbuf1.pMem, -1);
  ddocMemBuf_free(&mbuf1);
  ddocMemAppendData(pMBufXML, "</EncapsulatedOCSPValue>\n", -1);
  // Ver 1.76 - in format 1.3 we use here additionally element <OCSPValues>
  ddocMemAppendData(pMBufXML,"</OCSPValues>", -1);
  ddocMemAppendData(pMBufXML,"</RevocationValues>", -1);
  // unsigned prop end
  ddocMemAppendData(pMBufXML, "</UnsignedSignatureProperties>\n</UnsignedProperties>", -1);
  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}

//============================================================
// Adds a signature to signed document file
// hFile - output file
// pSigInfo - signature info 
// cert - signers certificate
// returns error code
//============================================================
int addSignatureInfoXMLBDoc(DigiDocMemBuf *pMBufXML, SignedDoc* pSigDoc, SignatureInfo* pSigInfo)
{
  int err = ERR_OK;
  unsigned char buf2[500], *buf1 = 0;
  int len2, len1;
  EVP_PKEY* pubKey;
  SignatureValue *pSigVal;
  DigiDocMemBuf mbuf1, *pMBuf2 = 0;

  RETURN_IF_NULL_PARAM(pMBufXML);
  RETURN_IF_NULL_PARAM(pSigDoc);
  RETURN_IF_NULL_PARAM(pSigInfo);
  mbuf1.pMem = 0;
  mbuf1.nLen = 0;
  //TODO: xmlns = http://uri.etsi.org/01903/v1.3.2#
  ddocMemAppendData(pMBufXML,"<Signature Id=\"", -1);
  ddocMemAppendData(pMBufXML,pSigInfo->szId, -1);
  ddocMemAppendData(pMBufXML,"\" xmlns=\"http://www.w3.org/2000/09/xmldsig#\">\n", -1);
  buf1 = (unsigned char*)createXMLSignedInfoBDoc(pSigDoc, pSigInfo);
	RETURN_IF_NULL(buf1);
  ddocMemAppendData(pMBufXML,(const char*)buf1, -1);
  free(buf1);
  len1 = 300;
  buf1 = (unsigned char*)malloc(len1);
  RETURN_IF_BAD_ALLOC(buf1);
  memset(buf1, 0, len1);
  // <SignatureValue>
  pSigVal = ddocSigInfo_GetSignatureValue(pSigInfo);
  //RETURN_IF_NULL(pSigVal);
  ddocSignatureValue_toXML(pSigVal, &mbuf1);
  ddocMemAppendData(pMBufXML,(const char*)mbuf1.pMem, -1);
  ddocMemBuf_free(&mbuf1);
  // cert data...
  // RSA KEY value
  ddocMemAppendData(pMBufXML,"<KeyInfo>\n", -1);
  err = GetPublicKey(&pubKey, ddocSigInfo_GetSignersCert(pSigInfo));
  // FIXME
  //AM 11.02.09 
  if(!err && pubKey->type==NID_rsaEncryption){
	  ddocMemAppendData(pMBufXML,"<KeyValue>\n<RSAKeyValue>\n", -1);
	  len1 = BN_bn2bin(pubKey->pkey.rsa->n, buf1);
	  // in version 1.1 we output modulus as it is
	  // starting from 1.2 we convert it to big-endian
	  /*len2 = sizeof(buf2);
		memset(buf2, 0, len2);
		encode(buf1, len1, buf2, &len2);
		printf("Old modulus: %s\n", buf2);
		if(!strcmp(pSigDoc->szFormatVer, DIGIDOC_XML_1_1_VER)) {
		swapBytes((byte*)buf1, len1);
		}*/
	  len2 = sizeof(buf2);
	  memset(buf2, 0, len2);
	  encode(buf1, len1, buf2, &len2);
	  //printf("New modulus: %s len: %d\n", buf2, len1);
	  ddocMemAppendData(pMBufXML, "<Modulus>", -1);
	  ddocMemAppendData(pMBufXML,buf2, -1);
	  ddocMemAppendData(pMBufXML,"</Modulus>\n", -1);
	  // exponent
	  memset(buf1, 0, len1);
	  len1 = BN_bn2bin(pubKey->pkey.rsa->e, buf1);
	  len2 = sizeof(buf2);
	  memset(buf2, 0, len2);
	  encode(buf1, len1, buf2, &len2);
	  ddocMemAppendData(pMBufXML,"<Exponent>", -1);
	  ddocMemAppendData(pMBufXML,buf2, -1);
	  ddocMemAppendData(pMBufXML,"</Exponent>\n", -1);
	  ddocMemAppendData(pMBufXML,"</RSAKeyValue>\n</KeyValue>\n", -1);
  }
  EVP_PKEY_free(pubKey);
  RETURN_IF_NOT(err == ERR_OK, err);
  // cert data
  ddocMemAppendData(pMBufXML,"<X509Data><X509Certificate>", -1);
  free(buf1);
  
  err = getCertPEM(ddocSigInfo_GetSignersCert(pSigInfo), 0, (char**)&buf1);
  RETURN_IF_NULL(buf1);
  ddocMemAppendData(pMBufXML,buf1, -1);
  free(buf1);
  ddocMemAppendData(pMBufXML,"</X509Certificate></X509Data></KeyInfo>\n", -1);
  // VS in releases prior to 1.76 we had incorrect <QualifyingProperties> atributes
  ddocMemAppendData(pMBufXML,"<Object><QualifyingProperties xmlns=\"http://uri.etsi.org/01903/v1.3.2#\" Target=\"#", -1);
  ddocMemAppendData(pMBufXML,pSigInfo->szId, -1);
  ddocMemAppendData(pMBufXML,"\">\n", -1);
  //ddocMemAppendData(pMBufXML,(const char*)buf2, -1);

  // signed properties	
  buf1 = (unsigned char*)createXMLSignedProperties(pSigDoc, pSigInfo, 1);
  RETURN_IF_NULL(buf1);
  ddocMemAppendData(pMBufXML,(const char*)buf1, -1);
  free(buf1);
  // unsigned properties
  if(pSigInfo->pNotary)
    addNotaryInfoXMLBDoc(pMBufXML, pSigDoc, pSigInfo, pSigInfo->pNotary);
  // qualifying properties end
  ddocMemAppendData(pMBufXML,"</QualifyingProperties></Object>\n", -1);
  // signature end
  ddocMemAppendData(pMBufXML,"</Signature>\n", -1);
  if (err != ERR_OK) SET_LAST_ERROR(err);	
  ddocMemBuf_free(&mbuf1);
  return err;
}
void stradd(char* str, const char *astr, int pos)
{	int i;
	for(i=0; i<strlen(astr); i++)
		str[pos+i] = astr[i];
}
/*char *uriEncode(char *str)
{	
	char ret[300];
	int pos;
	int len, i;

	len = strlen(str);
	//ret = char[len+1];
	//cout << ret[0];
	pos=0;
	for(i=0; i<len; i++)	
		switch(str[i])
		{		
		case '%':
			stradd(ret,"%25",pos);
			pos+=3;
			break;		
		case ' ':		
			stradd(ret,"%20",pos);
			pos+=3;		
			break;		
		case '^':		
			stradd(ret,"%5E",pos);
			pos+=3;	
			break;		
		case '&':		
			//stradd(ret,"%26",pos);
			//pos+=3;
			stradd(ret,"&amp;",pos);
			pos+=5;	
			break;		
		case '`':		
			stradd(ret,"%60",pos);
			pos+=3;	
			break;		
		case '{':	
			stradd(ret,"%7B",pos);
			pos+=3;
			break;		
		case '}':		
			stradd(ret,"%7D",pos);
			pos+=3;	
			break;	
		case '|':		
			stradd(ret,"%7C",pos);
			pos+=3;	
			break;		
		case ']':		
			stradd(ret,"%5D",pos);
			pos+=3;	
			break;		
		case '[':		
			stradd(ret,"%5B",pos);
			pos+=3;	
			break;		
		case '"':		
			stradd(ret,"%22",pos);
			pos+=3;	
			break;		
		case '<':	
			stradd(ret,"%3C",pos);
			pos+=3;
			break;		
		case '>':	
			stradd(ret,"%3E",pos);
			pos+=3;	
			break;	
		case '\\':	
			stradd(ret,"%5C",pos);
			pos+=3;	
			break;		
		case '#':		
			stradd(ret,"%23",pos);
			pos+=3;	
			break;		
		case '?':		
			stradd(ret,"%3F",pos);
			pos+=3;	
			break;		
		case '/':		
		  if(i!=0) {stradd(ret,"%2F",pos); pos+=3;}	
		  else {stradd(ret,"/",pos); pos+=1;}	
			break;			
		default:
			ret[pos++]=str[i];	
			break;	
	}	
	ret[pos]='\0';
	return ret;
}	*/

char* legalChars = "-_.!~*'();/?&:@=+$,"; // leave out & as it must be transformed to &amp;

unsigned isLegalChar(unsigned char c)
{
	int i, l = strlen(legalChars);
	for(i = 0; i < l; i++)
		if(legalChars[i] == c)
			return i;
	return -1; // not found
}

unsigned char *uriEncode(char *str)
{
	unsigned char *ret, c;
	int i, j = 0, k = 0, l = strlen(str);

	// calc mem size requirements
	for(i = 0; (i < l); i++) {
		c = (unsigned char)(str[i] & 0x000000FF);
		if(!isalnum(c) && isLegalChar(c) == -1) {
			k += 3;
		} else {
			k++;
		}
	}
	k++;
	ret = malloc(k);
	if(!ret) return NULL;
	memset(ret, 0, k);
	for(i = 0; (i < l) && (j < k); i++) {
		c = (unsigned char)(str[i] & 0x000000FF);
		if(!isalnum(c) && isLegalChar(c) == -1) { // to be converted
			sprintf(ret+j, "%%%2x", c);
			j += 3;
		} else { // normal chars and &
			ret[j] = c;
			j++;
		}
	}
	return ret;
}



//============================================================
// Creates a <SignedInfo> XML block for a signature
// This is the actual data to be signed
// pSigInfo - signature info data
// buf - output buffer
// returns number of output bytes written
//============================================================
EXP_OPTION char* createXMLSignedInfoBDoc(const SignedDoc* pSigDoc, const SignatureInfo* pSigInfo)
{
  char buf1[300], *pRet = 0, *p1, *p2;
  int i, err, l1;
  xmlNodePtr pnSigInfo, pN1, pN2;
  static xmlChar nl[] = "\n";
  xmlDocPtr doc;
  xmlChar* pBuf;
  //FILE* hFile;
  DataFile* pDF;
  DigiDocMemBuf mbuf1;
  //AM 13.02.09 ecdsa-sha1 support
  EVP_PKEY* pubKey;
  
  mbuf1.pMem = 0;
  mbuf1.nLen = 0;
  RETURN_OBJ_IF_NULL(pSigInfo, 0);
  // XML doc
  doc = xmlNewDoc((const xmlChar*)"1.0");
  doc->children = xmlNewDocNode(doc, NULL, (const xmlChar*)"", NULL);
  // <SignedInfo>
  pnSigInfo = xmlNewChild(doc->children, NULL, (const xmlChar*)"SignedInfo", nl);
  // in ver 1.1 we need this namespace def
  xmlNewNs(pnSigInfo, (const xmlChar*)NAMESPACE_XML_DSIG, NULL);
  // <CanonicalizationMethod>
  pN1 = xmlNewChild(pnSigInfo, NULL, (const xmlChar*)"CanonicalizationMethod", nl);
  xmlSetProp(pN1, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.w3.org/TR/2001/REC-xml-c14n-20010315");
  xmlNodeAddContent(pnSigInfo, nl);
  // <SignatureMethod>
  pN1 = xmlNewChild(pnSigInfo, NULL, (const xmlChar*)"SignatureMethod", nl);
  err = GetPublicKey(&pubKey, ddocSigInfo_GetSignersCert(pSigInfo));
#ifdef WITH_ECDSA
  if(pubKey->type==NID_X9_62_id_ecPublicKey)
	xmlSetProp(pN1, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.w3.org/2001/04/xmldsig-more#ecdsa-sha1");
  else { 
#endif
  if(BDOC_DIGEST==DIGEST_SHA256)
	//AM FIXME_SHA256 add sha256 when rsa-sha256 is possible
	xmlSetProp(pN1, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#rsa-sha1");
  else
	xmlSetProp(pN1, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#rsa-sha1");
#ifdef WITH_ECDSA
  }
#endif
  EVP_PKEY_free(pubKey);
  xmlNodeAddContent(pnSigInfo, nl);
  // <Reference>
  for(i = 0; i < pSigInfo->nDocs; i++) {
	pDF = getDataFileWithId(pSigDoc, pSigInfo->pDocs[i]->szDocId);
    // documents digest
    pN1 = xmlNewChild(pnSigInfo, NULL, (const xmlChar*)"Reference", nl);
    snprintf(buf1, sizeof(buf1), "%s", uriEncode(pSigInfo->pDocs[i]->szDocId));
    //buf = uriEncode(pSigInfo->pDocs[i]->szDocId); //AM uri encode is needed
	xmlSetProp(pN1, (const xmlChar*)"URI", (const xmlChar*)buf1);
    
    pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"DigestMethod", nl);
		if(BDOC_DIGEST==DIGEST_SHA256)
			xmlSetProp(pN2, (const xmlChar*)"Algorithm", (const xmlChar*)DIGEST_METHOD_SHA256);
    else
			xmlSetProp(pN2, (const xmlChar*)"Algorithm", (const xmlChar*)"http://www.w3.org/2000/09/xmldsig#sha1");
		xmlNodeAddContent(pN1, nl);
	// if no digest then calculate file digest
	if(!pSigInfo->pDocs[i]->nDigestLen || !pSigInfo->pDocs[i]->szDigest ||
		!pSigDoc->pDataFiles[i]->mbufDigest.nLen) {
		//pSigInfo->pDocs[i]->szDigestType
		if(BDOC_DIGEST==DIGEST_SHA256) {
			err = calculateFileDigestFromZip(pSigDoc->szFileName, pSigDoc->pDataFiles[i]->szFileName, (byte*)buf1, DIGEST_SHA256, DIGEST_LEN256);
			if(!err)
				setString((char**)&(pSigInfo->pDocs[i]->szDigest), buf1, DIGEST_LEN256);
			pSigInfo->pDocs[i]->nDigestLen = DIGEST_LEN256;
			err = ddocMemAssignData(&(pSigDoc->pDataFiles[i]->mbufDigest), buf1, DIGEST_LEN256);
		} else {
			err = calculateFileDigestFromZip(pSigDoc->szFileName, pSigDoc->pDataFiles[i]->szFileName, (byte*)buf1, DIGEST_SHA1, DIGEST_LEN);
			if(!err)
				setString((char**)&(pSigInfo->pDocs[i]->szDigest), buf1, DIGEST_LEN);
			pSigInfo->pDocs[i]->nDigestLen = DIGEST_LEN;
			err = ddocMemAssignData(&(pSigDoc->pDataFiles[i]->mbufDigest), buf1, DIGEST_LEN);
		}
		memset(buf1, 0, sizeof(buf1));
	}

    l1 = sizeof(buf1);
    encode(pSigInfo->pDocs[i]->szDigest, 
	   pSigInfo->pDocs[i]->nDigestLen, (byte*)buf1, &l1);
	ddocDebug(3, "createXMLSignedInfoBDoc", "Ref: %s digest: %s", pSigInfo->pDocs[i]->szDocId, buf1);
    pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"DigestValue", (const xmlChar*)buf1);
    xmlNodeAddContent(pN1, nl);
    xmlNodeAddContent(pnSigInfo, nl);
  }
  // signed properties digest
  pN1 = xmlNewChild(pnSigInfo, NULL, (const xmlChar*)"Reference", nl);
  snprintf(buf1, sizeof(buf1), "#%s-SignedProperties", pSigInfo->szId);
  xmlSetProp(pN1, (const xmlChar*)"URI", (const xmlChar*)buf1);
  // for References that contain <SignedProperties> digest
  xmlSetProp(pN1, (const xmlChar*)"Type", (const xmlChar*)"http://uri.etsi.org/01903#SignedProperties");

  pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"DigestMethod", nl);
	if(BDOC_DIGEST==DIGEST_SHA256)
		xmlSetProp(pN2, (const xmlChar*)"Algorithm", (const xmlChar*)DIGEST_METHOD_SHA256);
	else
		xmlSetProp(pN2, (const xmlChar*)"Algorithm", (const xmlChar*)DIGEST_METHOD_SHA1);
  xmlNodeAddContent(pN1, nl);

  ddocEncodeBase64(ddocDigestValue_GetDigestValue(pSigInfo->pSigPropDigest), &mbuf1);
	//AM 17.11.08 to remove newline after base64
	if(mbuf1.pMem && ((char*)mbuf1.pMem)[strlen((const char*)mbuf1.pMem)-1] == '\n')
    ((char*)mbuf1.pMem)[strlen((const char*)mbuf1.pMem)-1] = 0;
  pN2 = xmlNewChild(pN1, NULL, (const xmlChar*)"DigestValue", (const xmlChar*)mbuf1.pMem);
  ddocMemBuf_free(&mbuf1);
  xmlNodeAddContent(pN1, nl);
  xmlNodeAddContent(pnSigInfo, nl);
  // convert to string
  pBuf = NULL;
  err = xmlC14NDocDumpMemory(doc, NULL, 0, NULL, 0, &pBuf);	
  p1 = strstr((const char*)pBuf, "<SignedInfo");
  if(p1) {
    p2 = strstr((const char*)pBuf, "</SignedInfo>");
    if(p2) {
      p2[strlen("</SignedInfo>")] = 0;
      pRet = strdup(p1);
    }
  }
  xmlFree(pBuf);
  xmlFreeDoc(doc);
  return pRet;
}

//--------------------------------------------------
// Creates a new signed bdoc
// pSigDoc - signed doc info
// szSigDocFile - output XML file name. If the file exists,
// pMBufXML - output buffer if required to pass back in memory
// then it will be used to read in embedded DataFile contents.
// returns error code or ERR_OK for success
//--------------------------------------------------
EXP_OPTION int createSignedXMLBDoc(SignedDoc* pSigDoc, const char* szOldFile, const char* szSigDocFile, DigiDocMemBuf* pMBufXML, int hash)
{
  int err = ERR_OK, i, nFiles, len1, k, j;
  //FILE *hFile = 0;
  DataFile* pDf = NULL;
  char ** arrTempFiles = NULL;
  char buf[300];
  char buf1[300];
  char buf2[300]= "META-INF/";
  char buf3[300];
  char buf4[5]; //for id
  char fileNameBuf[300]; //for id
  char fixedFileName[300], *fName;
  char szSigFile[200] = "";
  char szMimeBuf[100] = "application/vnd.bdoc";
  DigiDocMemBuf mbuf1;
  DigiDocMemBuf manifBuf;
  DigiDocMemBuf sigBuf;
  DigiDocMemBuf hashBuf;
  DigiDocMemBuf digestBuf;
	//DigiDocMemBuf metaBuf;
  
  RETURN_IF_NULL_PARAM(pSigDoc);
  RETURN_IF_NULL_PARAM(szSigDocFile || pMBufXML);

  mbuf1.pMem = 0;
  mbuf1.nLen = 0;
  manifBuf.pMem = 0;
  manifBuf.nLen = 0;
  sigBuf.pMem = 0;
  sigBuf.nLen = 0;
  hashBuf.pMem = 0;
  hashBuf.nLen = 0;
  digestBuf.pMem = 0;
  digestBuf.nLen = 0;
  nFiles = getCountOfDataFiles(pSigDoc);
  ddocDebug(3, "createSignedXMLDoc", "Old file: %s new file: %s mbuf: %s, DFs: %d", 
	    szOldFile, (szSigDocFile ? szSigDocFile : "NULL"), (pMBufXML ? "Y" : "N"), nFiles);

  if(szSigDocFile) {
		err = ddocConvertFileName(buf1, sizeof(buf1), szSigDocFile);
		if(!err && szOldFile)
		  err = ddocConvertFileName(buf3, sizeof(buf3), szOldFile);
		if(err) return err;
  
		//filename is needed for deleting datafile
		setString(&pSigDoc->szFileName, buf1, -1);
		if(szOldFile && checkFileExists(szOldFile) && !hash){
			if(strcmp(buf3,buf1))
				err = makeDfZIP(buf3,buf1);
			else 
				err = makeDfZIP2(buf1);
			if (err != ERR_OK) return err;
	}

	//add mimetype
	strcat(szMimeBuf, "-");
  strcat(szMimeBuf, pSigDoc->szFormatVer);
	err = addFileToZip(pSigDoc->szFileName, NULL,"mimetype", 2, szMimeBuf);
	if (err != ERR_OK) return err;
	//generating meta.xml (based on Dublin Core Metadata Element Set, Version 1.1)
	/*ddocMemAppendData(&metaBuf, "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n", -1);
	ddocMemAppendData(&metaBuf, "<!DOCTYPE dublinCore PUBLIC \"-//OCLC//DTD Dublin core v.1//EN\" \"dublinCore.dtd\"> \n", -1);
  ddocGen_startElem(&metaBuf, "dublinCore");
	ddocMemAppendData(&metaBuf, "\n", -1);
  ddocGen_startElem(&metaBuf, "format");
  ddocMemAppendData(&metaBuf, "application/", -1);
  ddocMemAppendData(&metaBuf, pSigDoc->szFormat, -1);
	ddocMemAppendData(&metaBuf, "/", -1);
	ddocMemAppendData(&metaBuf, pSigDoc->szFormatVer, -1);
  ddocGen_endElem(&metaBuf, "format");
	ddocMemAppendData(&metaBuf, "\n", -1);
	if(ConfigItem_lookup("CLIENT") && ConfigItem_lookup("LIBRARY"))
		ddocGen_startElem(&metaBuf, "creator");
	  ddocMemAppendData(&metaBuf, ConfigItem_lookup("CLIENT"), -1);
		ddocMemAppendData(&metaBuf, "/", -1);
		ddocMemAppendData(&metaBuf, ConfigItem_lookup("LIBRARY"), -1);
		ddocGen_endElem(&metaBuf, "creator");
		ddocMemAppendData(&metaBuf, "\n", -1);
  ddocGen_startElem(&metaBuf, "date");
  time (&rawtime);
	ddocMemAppendData(&metaBuf, ctime (&rawtime), -1);
  ddocGen_endElem(&metaBuf, "date");		  
	ddocMemAppendData(&metaBuf, "\n", -1);
  ddocGen_endElem(&metaBuf, "dublinCore");
	addFileToZip(pSigDoc->szFileName, NULL,"meta.xml", 2, (char*)metaBuf.pMem);*/
	
		/*ddocMemAppendData(&metaBuf, "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n", -1);
	ddocMemAppendData(&metaBuf, "<!DOCTYPE dublinCore PUBLIC \"-//OCLC//DTD Dublin core v.1//EN\" \"dublinCore.dtd\"> \n", -1);
  ddocGen_startElem(&metaBuf, "dublinCore");
	ddocMemAppendData(&metaBuf, "\n", -1);
  ddocGen_startElem(&metaBuf, "format");
  ddocMemAppendData(&metaBuf, "application/", -1);
  ddocMemAppendData(&metaBuf, pSigDoc->szFormat, -1);
	ddocMemAppendData(&metaBuf, "/", -1);
	ddocMemAppendData(&metaBuf, pSigDoc->szFormatVer, -1);
  ddocGen_endElem(&metaBuf, "format");
	ddocMemAppendData(&metaBuf, "\n", -1);
	if(ConfigItem_lookup("CLIENT") && ConfigItem_lookup("LIBRARY"))
		ddocGen_startElem(&metaBuf, "creator");
	  ddocMemAppendData(&metaBuf, ConfigItem_lookup("CLIENT"), -1);
		ddocMemAppendData(&metaBuf, "/", -1);
		ddocMemAppendData(&metaBuf, ConfigItem_lookup("LIBRARY"), -1);
		ddocGen_endElem(&metaBuf, "creator");
		ddocMemAppendData(&metaBuf, "\n", -1);
  ddocGen_startElem(&metaBuf, "date");
  time (&rawtime);
	ddocMemAppendData(&metaBuf, ctime (&rawtime), -1);
  ddocGen_endElem(&metaBuf, "date");		  
	ddocMemAppendData(&metaBuf, "\n", -1);
  ddocGen_endElem(&metaBuf, "dublinCore");
	addFileToZip(pSigDoc->szFileName, NULL,"meta.xml", 2, (char*)metaBuf.pMem);*/
	}

  // now create the new document
  if(szSigDocFile || pMBufXML) {
		ddocMemAppendData(&manifBuf, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n", -1);
	//ddocMemAppendData(&manifBuf, "<!DOCTYPE manifest:manifest PUBLIC \"-//OpenOffice.org//DTD Manifest 1.0//EN\" \"Manifest.dtd\">\n", -1);	
		ddocMemAppendData(&manifBuf, "<manifest:manifest xmlns:manifest=\"urn:oasis:names:tc:opendocument:xmlns:manifest:1.0\">\n", -1);		
		ddocMemAppendData(&manifBuf, "<manifest:file-entry manifest:media-type=\"application/vnd.bdoc-", -1);
		ddocMemAppendData(&manifBuf, pSigDoc->szFormatVer, -1);		
		ddocMemAppendData(&manifBuf, "\" manifest:full-path=\"/\" />\n", -1);
    if(pMBufXML)
      ddocMemAppendData(pMBufXML, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n", -1);
	
	if(hash){ 
	  ddocMemAppendData(&hashBuf, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n", -1);
	  ddocMemAppendData(&hashBuf, "<DataFiles>\n", -1);
	}
    // DataFile objects
	for(i = 0; i < nFiles; i++) {
	  pDf = getDataFile(pSigDoc, i);
	  if(szSigDocFile && !hash) { 
				err = ddocConvertFileName(fileNameBuf, sizeof(fileNameBuf), pDf->szFileName);
				if(checkFileExists(pDf->szFileName)) { // don't add files not locally existing
				err= addFileToZip(pSigDoc->szFileName, fileNameBuf, (char*)getSimpleFileName(pDf->szFileName), 2, NULL);
				if (err != ERR_OK) return err;
				}
			}
			if(hash){
			  ddocMemAppendData(&hashBuf, "<DataFile DigestType=\"", -1);
			  ddocMemAppendData(&hashBuf,pDf->szDigestType, -1);
			  ddocMemAppendData(&hashBuf,"\" DigestValue=\"",-1);
			  ddocEncodeBase64(ddocDataFile_GetDigestValue(pDf), &digestBuf);
			  if(digestBuf.pMem && ((char*)digestBuf.pMem)[strlen((const char*)digestBuf.pMem)-1] == '\n')
			  ((char*)digestBuf.pMem)[strlen((const char*)digestBuf.pMem)-1] = 0;
			  ddocMemAppendData(&hashBuf,(char*)digestBuf.pMem, -1);
			  ddocMemAppendData(&hashBuf,"\" Filename=\"",-1);
			  len1 = sizeof(buf1);
			  // replaces '&' with '&amp;'
			  memset(fixedFileName, 0, sizeof(fixedFileName));
			  fName = (char*)getSimpleFileName(pDf->szFileName);
			  for(k = j = 0; k < (int)strlen(fName); k++) {
				if(fName[k] == '&') {
				  strncat(fixedFileName, "&amp;", sizeof(fixedFileName) - strlen(fixedFileName));
				  j += 5;
				} else {
				  fixedFileName[j] = fName[k];
				  j++;
				}
			  }
			  ddocMemAppendData(&hashBuf,fixedFileName, -1);
			  ddocMemAppendData(&hashBuf,"\" MimeType=\"",-1);
			  ddocMemAppendData(&hashBuf,pDf->szMimeType, -1);
			  ddocMemAppendData(&hashBuf,"\" Size=\"",-1);
			  snprintf(buf, sizeof(buf), "%d", (int)pDf->nSize);
			  ddocMemAppendData(&hashBuf,buf, -1);
			  ddocMemAppendData(&hashBuf,"\" />\n", -1);			  
			} else { 
			  // replaces '&' with '&amp;'
			  memset(fixedFileName, 0, sizeof(fixedFileName));
			  fName = (char*)getSimpleFileName(pDf->szFileName);
			  for(k = j = 0; k < (int)strlen(fName); k++) {
				if(fName[k] == '&') {
				  strncat(fixedFileName, "&amp;", sizeof(fixedFileName) - strlen(fixedFileName));
				  j += 5;
				} else {
				  fixedFileName[j] = fName[k];
				  j++;
				}
			  }
			  ddocMemAppendData(&manifBuf, "<manifest:file-entry manifest:media-type=\"", -1);
			  ddocMemAppendData(&manifBuf,pDf->szMimeType, -1);
			  ddocMemAppendData(&manifBuf,"\" manifest:full-path=\"",-1);
			  ddocMemAppendData(&manifBuf,fixedFileName, -1);
			  ddocMemAppendData(&manifBuf,"\" />\n", -1);
			}
	}
	if(hash){
	  ddocMemAppendData(&hashBuf, "</DataFiles>\n", -1);
	  addFileToZip(pSigDoc->szFileName, NULL, "META-INF/datafiles.xml", 2,(char*)hashBuf.pMem);
	}
    for(i = 0; i < pSigDoc->nSignatures; i++) {
			if(szSigDocFile && pSigDoc->pSignatures[i]->szId){
				ddocMemAssignData2(&sigBuf, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n", -1);
				strcpy(buf4,&pSigDoc->pSignatures[i]->szId[1]);
				snprintf(szSigFile, sizeof(szSigFile), "META-INF/signature%d.xml", atoi(buf4)+1); 
			}
		 //VS: deny writing orphoned signatures !
			if(pSigDoc->pSignatures[i]->pSigValue && 
					(pSigDoc->pSignatures[i]->pSigValue->mbufSignatureValue.pMem ||
					 pSigDoc->pSignatures[i]->mbufOrigContent.pMem))  {
				  // VS: if Signature has been read from file then 
				  // use the original content
				  if(pSigDoc->pSignatures[i]->mbufOrigContent.pMem) {
						if(szSigDocFile){
							ddocMemAppendData(&sigBuf, pSigDoc->pSignatures[i]->mbufOrigContent.pMem, -1);
						}
						if(pMBufXML)
							err = ddocMemAppendData(pMBufXML, (const char*)pSigDoc->pSignatures[i]->mbufOrigContent.pMem, pSigDoc->pSignatures[i]->mbufOrigContent.nLen);
				  } else {
      			err = addSignatureInfoXMLBDoc(&mbuf1, pSigDoc, pSigDoc->pSignatures[i]);
				if(szSigDocFile)
					ddocMemAppendData(&sigBuf, (char*)mbuf1.pMem, -1);
					
				if(pMBufXML)
					err = ddocMemAppendData(pMBufXML, (const char*)mbuf1.pMem, mbuf1.nLen);
				}
					if(szSigDocFile){
						err = addFileToZip(pSigDoc->szFileName, NULL, szSigFile, 2, (char*)sigBuf.pMem);
						if (err != ERR_OK) return err;}
				ddocMemAppendData(&manifBuf, "<manifest:file-entry manifest:media-type=\"", -1);
				ddocMemAppendData(&manifBuf,"signature/", -1);			
				ddocMemAppendData(&manifBuf, "bdoc", -1);
				ddocMemAppendData(&manifBuf,"-", -1);
				ddocMemAppendData(&manifBuf,pSigDoc->szFormatVer, -1);
#ifdef WITH_TS
				if(ConfigItem_lookup_bool("DIGIDOC_WITH_TS", 0) || pSigDoc->pSignatures[i]->pTimestamps){
					ddocMemAppendData(&manifBuf,"/TS", -1); 
				}else{
#endif
					ddocMemAppendData(&manifBuf,"/TM", -1);
#ifdef WITH_TS
				}
#endif
				ddocMemAppendData(&manifBuf,"\" manifest:full-path=\"",-1);
				strcpy(buf3,buf2);
				strcat(buf3,getSimpleFileName(szSigFile));
				ddocMemAppendData(&manifBuf,buf3, -1);
				ddocMemAppendData(&manifBuf,"\" />\n", -1);
				ddocMemBuf_free(&mbuf1);
				ddocMemBuf_free(&sigBuf);
			}
    } 
	ddocMemAppendData(&manifBuf, "</manifest:manifest>", -1);

	if (err != ERR_OK) return err;

	err = addFileToZip(pSigDoc->szFileName, NULL, "META-INF/manifest.xml", 2,(char*)manifBuf.pMem);

	ddocMemBuf_free(&manifBuf);
	ddocMemBuf_free(&hashBuf);
	ddocMemBuf_free(&digestBuf);
	//meta.xml buf
	//ddocMemBuf_free(&metaBuf);
  } else
    err = ERR_FILE_WRITE;
  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}

//--------------------------------------------------
// Creates a new signed document
// pSigDoc - signed doc info
// returns error code or ERR_OK for success
//--------------------------------------------------
EXP_OPTION int createSignedBDoc(SignedDoc* pSigDoc, const char* szOldFile, const char* szOutputFile)
{
  int err = ERR_OK;

  RETURN_IF_NULL_PARAM(pSigDoc);
  RETURN_IF_NULL_PARAM(szOutputFile);
  RETURN_IF_NULL_PARAM(pSigDoc->szFormat);
	
  clearErrors();	
  if(!strcmp(pSigDoc->szFormat, BDOC_XML_1_NAME)) { 
    err = createSignedXMLBDoc(pSigDoc, szOldFile, szOutputFile, NULL,0);
  }
  else
    err = ERR_UNSUPPORTED_FORMAT;
  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}

//--------------------------------------------------
// Creates a new signed document in memory buffer
// pSigDoc - signed doc info
// szOldFile - name of old file on disk to copy DataFile contents
// pMBuf - buffer for new digidoc document
// returns error code or ERR_OK for success
//--------------------------------------------------
EXP_OPTION int createSignedBDocInMemory(SignedDoc* pSigDoc, const char* szOldFile, DigiDocMemBuf* pMBuf)
{
  int err = ERR_OK;

  RETURN_IF_NULL_PARAM(pSigDoc);
  RETURN_IF_NULL_PARAM(pMBuf);
  RETURN_IF_NULL_PARAM(pSigDoc->szFormat);
  clearErrors();	
  if(!strcmp(pSigDoc->szFormat, BDOC_XML_1_NAME)) { 
    err = createSignedXMLBDoc(pSigDoc, szOldFile, NULL, pMBuf,0);
  }
  else
    err = ERR_UNSUPPORTED_FORMAT;
  if (err != ERR_OK) SET_LAST_ERROR(err);
  return err;
}

//--------------------------------------------------
// Generates XML for <CompleteCertificateRefs> element
// pSigDoc - SignedDoc object [REQUIRED]
// pBuf - memory buffer for storing xml [REQUIRED]
// returns error code or ERR_OK
//--------------------------------------------------
int bdocCompleteCertificateRefs_toXML(const SignedDoc* pSigDoc, 
				      const SignatureInfo* pSigInfo, DigiDocMemBuf* pBuf)
{
  int err = ERR_OK, i, n;
  CertID* pCertID;
 
  RETURN_IF_NULL_PARAM(pBuf)
  RETURN_IF_NULL_PARAM(pSigDoc)
  RETURN_IF_NULL_PARAM(pSigInfo)
  RETURN_IF_NULL(pSigInfo->pCertIDs)
  // <CompleteCertificateRefs>
  err = ddocGen_startElemBegin(pBuf, "CompleteCertificateRefs");
  if(err) return err;
  // end of element start tag
  err = ddocGen_startElemEnd(pBuf);
  if(err) return err;
  err = ddocGen_startElem(pBuf, "CertRefs");
  if(err) return err;
 
  n = ddocCertIDList_GetCertIDsCount(pSigInfo->pCertIDs);
  for(i = 0; i < n; i++) {
    pCertID = ddocCertIDList_GetCertID(pSigInfo->pCertIDs, i);
    if(pCertID && pCertID->nType == CERTID_TYPE_CA_CERTID /*&& pCertID->nType == CERTID_TYPE_TSA_CERTID  */)
      ddocCertID_toXML(pSigDoc, pCertID, pBuf);
  }
  // </CertRefs> (not used in 1.0, 1.1 and 1.2
  err = ddocGen_endElem(pBuf, "CertRefs");
  if(err) return err;
  err = ddocGen_endElem(pBuf, "CompleteCertificateRefs");
  return err;
}
