//==================================================
// FILE:	DigiDocSAXParser.c
// PROJECT:     Digi Doc
// DESCRIPTION: Digi Doc functions for xml parsing using SAX interface
// AUTHOR:  Ants Madisson, Business IT Partner
//==================================================
// Copyright (C) AS Sertifitseerimiskeskus
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// GNU Lesser General Public Licence is available at
// http://www.gnu.org/copyleft/lesser.html
//==========< HISTORY >=============================
//      25.02.2008      Ants Madisson
//                      Creation
//==================================================

#include <libdigidoc/DigiDocDefs.h>
#include <libdigidoc/BDocParser.h>
#include <libdigidoc/DigiDocError.h>
#include <libdigidoc/DigiDocDebug.h>
#include <libdigidoc/DigiDocConvert.h>
#include <libdigidoc/DigiDocMem.h>
#include <libdigidoc/DigiDocLib.h>
#include <libdigidoc/DigiDocCert.h>
#include <libdigidoc/DigiDocConfig.h>
#include <libdigidoc/DigiDocOCSP.h>
#include <libdigidoc/BDocZip.h>
#include <libdigidoc/DigiDocSAXParser.h>
#include <libdigidoc/DigiDocStack.h>
#include <libdigidoc/DigiDocGen.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <ctype.h>

#include <libxml/globals.h>
#include <libxml/xmlerror.h>
#include <libxml/parser.h>
#include <libxml/parserInternals.h> /* only for xmlNewInputFromFile() */
#include <libxml/tree.h>
#include <libxml/debugXML.h>
#include <libxml/xmlmemory.h>
#include <libxml/c14n.h>

// string used to force parser to flush it's buffers
static char g_szDataFileFlush1[] = "</DataFile>";
static char g_szDataFileFlush2[] = "<DataFile Id=\"%s\" ContentType=\"%s\">";

#ifdef WIN32
  #define snprintf  _snprintf
#endif

extern int ddocCheckFormatAndVer(const char* format, const char* version);
extern char* canonicalizeXML(char* source, int len);
extern int escapeXMLSymbols(const char* src, int srclen, char** dest);

#define FLAG_XML_ELEM   1
#define FLAG_SIGNATURE  2
#define FLAG_SIG_PART   3
#define FLAG_TS_INP     4

//===============< SAX handlers >==============================

/*
* Working area for XML parsing with SAX
*/
typedef struct SigDocParse_st {
  SignedDoc* pSigDoc;	// document file to be filled with data
  char tag[300];		// current tag
  char ctx1[300];		// context1
  char ctx2[300];		// context2
  char ctx3[300];		// context3
  char ctx4[300];
  char ctx5[300];
  BIO* bDataFile;
  EVP_ENCODE_CTX ectx;
  SHA_CTX sctx;
  int errcode;
  char* szInputFileName;
  int checkFileDigest;
  int nIgnoreDataFile;  // incremented each time when EMBEDDED XML content is found.
			      // used to ignore such content
  DigiDocMemBuf mbufElemData;
  char bCollectElemData;
  char bNoXMLElemData;
  int checkUTF8;
  DigiDocMemBuf mbufSigData; // <Signature>
  DigiDocMemBuf mbufCanonSigData; // Canonicalized <Signature>
  char bCollectSigData;
  DigiDocMemBuf mbufSigPartData; // <SignedInfo> and <SignedProperties>
#ifdef WITH_TS
  DigiDocMemBuf mbufTsInput;   // total timestamp hash input data
  DigiDocMemBuf mbufTsInpElem;  // one xml elem for ts input before canonicalization
  char bCollectTsInput;
#endif
  char bCollectSigPartData;
  long lMaxDFLen;
  time_t tStartParse;
  long lSize;
  int bCollectDFData;  
  DigiDocMemBuf *pMemBufDF;
  int bKeepBase64;
  char b64line[66];
  int b64pos;
  ElementEntry stack; 
} SigDocParse;

//--------------------------------------------------
// Releases memory that might have been allocated decoding
// the parsing process.
// pctx - pointer to parsing context
//--------------------------------------------------
void bdocSAXCleanup(SigDocParse* pctx)
{
  // field: szInputFileName is not allocated in parser!
  ddocMemBuf_free(&(pctx->mbufElemData));
  ddocMemBuf_free(&(pctx->mbufSigData));
  ddocMemBuf_free(&(pctx->mbufSigPartData));
#ifdef WITH_TS
  ddocMemBuf_free(&(pctx->mbufTsInput));
  ddocMemBuf_free(&(pctx->mbufTsInpElem));
#endif
  if(pctx->pMemBufDF)
    ddocMemBuf_free(pctx->pMemBufDF);
  if(pctx->bDataFile)
    BIO_free(pctx->bDataFile);
  memset(pctx, 0, sizeof(SigDocParse));
}
//--------------------------------------------------
// Finds the desired atribute value. 
// atts - attributes array
// name - name of searched atribute
// defval - default value if not found
//--------------------------------------------------
const char* bdocSaxParseFindAttrib(const xmlChar **atts, 
				   const char* name, const char* defval)
{
  int i;
  
  for (i = 0; (atts != NULL) && (atts[i] != NULL); i += 2) {
    if(!strcmp((const char*)atts[i], name))
      return (const char*)atts[i+1];
  }
  return defval;
}
//--------------------------------------------------
// Starts collecting some element data.
// pctx - pointer to parsing context
// nFlag - 1=element, 2=signature, 3=sigprop, 4=ts-input
// bNoXml - no-xml flag
//--------------------------------------------------
int bdocSaxParseStartCollecting(SigDocParse* pctx, int nFlag, int bNoXml)
{
  RETURN_IF_NULL_PARAM(pctx);
  switch(nFlag) {
  case FLAG_XML_ELEM: 
    pctx->bCollectElemData = 1; 
    pctx->errcode = ddocMemBuf_free(&(pctx->mbufElemData));
    break;
  case FLAG_SIGNATURE: 
    pctx->bCollectSigData = 1; 
    pctx->errcode = ddocMemBuf_free(&(pctx->mbufSigData));
    break;
  case FLAG_SIG_PART: 
    pctx->bCollectSigPartData = 1; 
    pctx->errcode = ddocMemBuf_free(&(pctx->mbufSigPartData));
    break;
#ifdef WITH_TS
  case FLAG_TS_INP:
    pctx->bCollectTsInput = 1; 
    pctx->errcode = ddocMemBuf_free(&(pctx->mbufTsInpElem));
    break;
#endif
  }
  pctx->bNoXMLElemData = bNoXml;
  return pctx->errcode;
}

//--------------------------------------------------
// Ends collecting some element data.
// pctx - pointer to parsing context
// nFlag - 1=element, 2=signature, 3=sigprop, 4=ts-input
// bNoXml - no-xml flag
//--------------------------------------------------
int bdocSaxParseEndCollecting(SigDocParse* pctx, int nFlag, int bNoXml)
{
  RETURN_IF_NULL_PARAM(pctx);
  switch(nFlag) {
  case FLAG_XML_ELEM: 
    pctx->bCollectElemData = 0; 
    pctx->errcode = ddocMemBuf_free(&(pctx->mbufElemData));
    break;
  case FLAG_SIGNATURE: 
    pctx->bCollectSigData = 0; 
    pctx->errcode = ddocMemBuf_free(&(pctx->mbufSigData));
    break;
  case FLAG_SIG_PART: 
    pctx->bCollectSigPartData = 0; 
    pctx->errcode = ddocMemBuf_free(&(pctx->mbufSigPartData));
    break;
#ifdef WITH_TS
  case FLAG_TS_INP:
    pctx->bCollectTsInput = 0; 
    pctx->errcode = ddocMemBuf_free(&(pctx->mbufTsInpElem));
    break;
#endif
  }
  pctx->bNoXMLElemData = bNoXml;
  return pctx->errcode;
}

//--------------------------------------------------
// handles the start of a <SignedDoc> element.
// Creates SignedDoc structure to read in the info.
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartSignedBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  int i;
  const char *key = NULL, *value = NULL;

  //RETURN_VOID_IF_NULL(pctx);
  //RETURN_VOID_IF_NULL(name);
  //RETURN_VOID_IF_NULL(atts);

  strncpy(pctx->ctx1, (const char*)name, sizeof(pctx->ctx1)-1);
  for (i = 0; (atts != NULL) && (atts[i] != NULL); i += 2) {
    key = (const char*)atts[i];
    value = (const char*)atts[i+1];
    if(!strcmp(key, "format"))
      setString(&(pctx->pSigDoc->szFormat), value, -1);
    if(!strcmp(key, "version"))
      setString(&(pctx->pSigDoc->szFormatVer), value, -1);
  }
  if(ddocCheckFormatAndVer(pctx->pSigDoc->szFormat,
			   pctx->pSigDoc->szFormatVer))
    SET_LAST_ERROR(ERR_UNSUPPORTED_FORMAT);
}

//--------------------------------------------------
// handles the start of a <Signature> element.
// Adds a SignatureInfo struct to SignedDoc structure.
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartSignatureBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  int i;
  SignatureInfo *pSignatureInfo;
  for (i = 0; (atts != NULL) && (atts[i] != NULL); i += 2) {
    if(!strcmp((const char*)atts[i], "Id")) {
	  memset(pctx->ctx1, 0, sizeof(pctx->ctx1));
      memset(pctx->ctx2, 0, sizeof(pctx->ctx2));
      memset(pctx->ctx3, 0, sizeof(pctx->ctx3));
      strncpy(pctx->ctx1, (const char*)atts[i+1], /*sizeof(pctx->ctx1)-*/50);
      SignatureInfo_new(&pSignatureInfo, pctx->pSigDoc, (const char*)atts[i+1]); // MEMLEAK: ???
    }
  }
}
const char HEX2DEC[256] = 
{
    /*       0  1  2  3   4  5  6  7   8  9  A  B   C  D  E  F */
    /* 0 */ -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    /* 1 */ -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    /* 2 */ -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    /* 3 */  0, 1, 2, 3,  4, 5, 6, 7,  8, 9,-1,-1, -1,-1,-1,-1,
    
    /* 4 */ -1,10,11,12, 13,14,15,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    /* 5 */ -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    /* 6 */ -1,10,11,12, 13,14,15,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    /* 7 */ -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    
    /* 8 */ -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    /* 9 */ -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    /* A */ -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    /* B */ -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    
    /* C */ -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    /* D */ -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    /* E */ -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    /* F */ -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1
};
//--------------------------------------------------
// Decodes a URI value
// into its components
// uri - URI value
//--------------------------------------------------
//char* decodeURIBdoc(const char* uri, int src_len)
#define SPC_BASE16_TO_10(x) (((x) >= '0' && (x) <= '9') ? ((x) - '0') : \
                             (toupper((x)) - 'A' + 10))
   
char *decodeURIBdoc(const char *url) {
  char       *out, *ptr;
  const char *c;
  char  buf1[3];
   
  if (!(out = ptr = strdup(url))) return 0;
  for (c = url;  *c;  c++) {
	if(*c == '%' && isxdigit(c[1]) && isxdigit(c[2])) {
      *ptr++ = (SPC_BASE16_TO_10(c[1]) * 16) + (SPC_BASE16_TO_10(c[2]));
      c += 2;
    } else if(c[0] == '&' && c[1] == '#' && isxdigit(c[2]) && isxdigit(c[3]) && c[4] == ';') {
	  buf1[0] = c[2]; buf1[1] = c[3]; buf1[2] = 0;
      *ptr++ = atoi(buf1);
      c += 4;
    } else  // in all other cases
		*ptr++ = *c;
  }
  *ptr = 0;
  ddocDebug(3, "decodeURI", "URI: %s -> %s", url, out);
  return out;
}




//--------------------------------------------------
// handles the start of a <Reference> element.
// Records the reference URI attribute
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartReferenceBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  char p[300];

  strncpy(p, bdocSaxParseFindAttrib(atts, "URI", ""), sizeof(p)-1);
  strncpy(pctx->ctx2, decodeURIBdoc(p), sizeof(pctx->ctx2)-1);
  
}

//--------------------------------------------------
// handles the start of a <SignatureMethod> element.
// Records the signature method
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartSignatureMethodBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  SignatureInfo *pSigInfo;
  const char *alg;

  alg = bdocSaxParseFindAttrib(atts, "Algorithm", NULL);
  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  if((alg != NULL)  && (pSigInfo != NULL) &&
     !strcmp(alg+strlen(alg)-8, "rsa-sha1") &&
     (pctx->ctx1[0] == 'S')) {
    ddocSignatureValue_new(&(pSigInfo->pSigValue), 0, SIGN_RSA_NAME, 0, 0);
  }
#ifdef WITH_ECDSA
	else if((alg != NULL)  && (pSigInfo != NULL) &&
     !strcmp(alg+strlen(alg)-10, "ecdsa-sha1") &&
     (pctx->ctx1[0] == 'S')) {
    ddocSignatureValue_new(&(pSigInfo->pSigValue), 0, SIGN_ECDSA_NAME, 0, 0);
  }
#endif
}


//--------------------------------------------------
// handles the start of a <DigestMethod> element.
// Records the digest method
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartDigestMethodBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  SignatureInfo *pSigInfo = NULL;
  const char *alg;
  DataFile* pDf = NULL;

  alg = bdocSaxParseFindAttrib(atts, "Algorithm", NULL);
  if((pctx->ctx2[0] == '/') && (alg != NULL) &&
	(!strcmp(alg+strlen(alg)-4, "sha1") || !strcmp(alg+strlen(alg)-6, "sha256"))){
	pDf = getDataFileWithId(pctx->pSigDoc, pctx->ctx2);
	if(!pDf) {
		ddocDebug(1, "handleStartDigestMethodBDoc", "DataFile: %s not found!", pctx->ctx2);
		pctx->errcode = ERR_FILE_READ;
	}
  }
  ddocDebug(3, "handleStartDigestMethodBDoc", "DF: %s", pctx->ctx2);

  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  if((alg != NULL) && (pSigInfo != NULL) &&
		(!strcmp(alg+strlen(alg)-4, "sha1") || !strcmp(alg+strlen(alg)-6, "sha256"))){
		if(!strcmp(alg+strlen(alg)-6, "sha256"))
			strncpy(pctx->ctx3, DIGEST_SHA256_NAME, sizeof(pctx->ctx3));
		else
			strncpy(pctx->ctx3, DIGEST_SHA1_NAME, sizeof(pctx->ctx3));}
  else
    pctx->ctx3[0] = 0;
}

//--------------------------------------------------
// This function should be called at the end of
// </EncapsulatedOCSPValue> to determine the correct
// types of CertID and CertValue objects.
// pSigInfo - signature object
//--------------------------------------------------
int selectCertIdAndValueTypesBDoc(SignatureInfo* pSigInfo)
{
  int err = ERR_OK, i, j, l1, k;
  CertID* cid;
  CertValue *cval1, *cval2;
  X509* pCert;
  char buf1[300];
  NotaryInfo* pNotary = NULL;
  DigiDocMemBuf mbuf1, mbuf2;
  const DigiDocMemBuf *pMBuf = 0, *pBuf2 = 0;
#ifdef WITH_TS
  TimestampInfo* pTS = NULL;
  TS_RESP* pResp = NULL;
#endif

  RETURN_IF_NULL_PARAM(pSigInfo);
  mbuf1.pMem = 0;
  mbuf1.nLen = 0;
  mbuf2.pMem = 0;
  mbuf2.nLen = 0;
  for(i = 0; pSigInfo->pCertIDs &&
	i < ddocCertIDList_GetCertIDsCount(pSigInfo->pCertIDs); i++) {
    cid = ddocCertIDList_GetCertID(pSigInfo->pCertIDs, i);
    ddocDebug(3, "selectCertIdAndValueTypes", "CID type: %d serial %s - %s", cid->nType, cid->szIssuerSerial, cid->szIssuerName);
    if(cid && cid->nType == CERTID_TYPE_UNKNOWN) {
      ddocDebug(3, "selectCertIdAndValueTypes", "Find type for cid: %s - %s", cid->szIssuerSerial, cid->szIssuerName);
      // find corresponding CertValue
      cval1 = NULL;
      for(j = 0; pSigInfo->pCertValues && 
	    (j < ddocCertValueList_GetCertValuesCount(pSigInfo->pCertValues)); j++) {
	cval2 = ddocCertValueList_GetCertValue(pSigInfo->pCertValues, j);
	if(cval2) {
	  pCert = ddocCertValue_GetCert(cval2);
	  l1 = sizeof(buf1);
	  memset(buf1, 0, l1);
	  ReadCertSerialNumber(buf1, l1, pCert);
	  if(!strcmp(cid->szIssuerSerial, buf1)) {
	    cval1 = cval2;
	    break; // found it
	  }
	  // should I check also hash value?
	}
      } // for certValues
      // if found matching CertID and CertValue
      if(cval1) {
	ddocDebug(3, "selectCertIdAndValueTypes", "CertID: %s - %s -> CertValue: %s", 
		  cid->szIssuerSerial, cid->szIssuerName, cval1->szId);
	// Test1: is this an OCSP responders cert?
	if(pSigInfo->pNotary)
	  pMBuf = (DigiDocMemBuf*)ddocNotInfo_GetResponderId(pSigInfo->pNotary);
	if(pMBuf) {
	  if(pSigInfo->pNotary->nRespIdType == RESPID_NAME_TYPE) {
	    err = ddocCertGetIssuerCN(ddocCertValue_GetCert(cval1), &mbuf1);
	    memset(buf1, 0, sizeof(buf1));
	    findCN((char*)pMBuf->pMem, buf1); // defined in DigiDocConfig.c
	    // yes this is a responders cert
	    ddocDebug(3, "selectCertIdAndValueTypes", "cert: %s responder %s -> RESPONDER", 
			(const char*)mbuf1.pMem, buf1);
		if(!strcmp(buf1, (const char*)mbuf1.pMem)) {
	      cid->nType = CERTID_TYPE_RESPONDERS_CERTID;
	      cval1->nType = CERTID_VALUE_RESPONDERS_CERT;
	      snprintf(buf1, sizeof(buf1), "%s-RESPONDERS_CERTINFO", pSigInfo->szId);
	      ddocCertID_SetId(cid, buf1);	  
	    }
	    ddocMemBuf_free(&mbuf1);
	  } // ByName
	  if(pSigInfo->pNotary->nRespIdType == RESPID_KEY_TYPE) {
	    err = ddocCertGetPubkeyDigest(ddocCertValue_GetCert(cval1), &mbuf1);
	    l1 = sizeof(buf1);
	    memset(buf1, 0, l1);
	    bin2hex((const byte*)mbuf1.pMem, mbuf1.nLen, buf1, &l1);
	    ddocDebug(3, "selectCertIdAndValueTypes", "cert hash: %s", buf1);
	    l1 = sizeof(buf1);
	    memset(buf1, 0, l1);
	    bin2hex((const byte*)pMBuf->pMem, pMBuf->nLen, buf1, &l1);
	    ddocDebug(3, "selectCertIdAndValueTypes", "respid: %s", buf1);
	    
	    ddocEncodeBase64(&mbuf1, &mbuf2);
	    ddocMemBuf_free(&mbuf1);
	    ddocEncodeBase64(pMBuf, &mbuf1);
	    ddocDebug(3, "selectCertIdAndValueTypes", "cert: %s responder %s", 
			(const char*)mbuf2.pMem, (const char*)mbuf1.pMem);
	    if(!strcmp((const char*)mbuf2.pMem, (const char*)mbuf1.pMem)) {
	      // yes this is a responders cert
	      ddocDebug(3, "selectCertIdAndValueTypes", "cert: %s responder %s -> RESPONDER", 
			(const char*)mbuf2.pMem, (const char*)mbuf1.pMem);
	      cid->nType = CERTID_TYPE_RESPONDERS_CERTID;
	      cval1->nType = CERTID_VALUE_RESPONDERS_CERT;
	      snprintf(buf1, sizeof(buf1), "%s-RESPONDERS_CERTINFO", pSigInfo->szId);
	      ddocCertID_SetId(cid, buf1);	  
	    }
	    ddocMemBuf_free(&mbuf1);
	    ddocMemBuf_free(&mbuf2);
	  } // ByKey
	} // if pMBuf
#ifdef WITH_TS
			if(cid->nType == CERTID_TYPE_UNKNOWN) {
				// Test2: is this a TSA cert ?
				for(k = 0; pSigInfo->pTimestamps &&
				k < ddocTimestampInfoList_GetTimestampInfosCount(pSigInfo->pTimestamps); k++) {
				  pTS = ddocTimestampInfoList_GetTimestampInfo(pSigInfo->pTimestamps, k);
				  pBuf2 = (DigiDocMemBuf*)ddocTimestampInfo_GetTS(pTS);
				  if(pBuf2 && pBuf2->pMem) {
				  ddocTsReadTsResp(&pResp, pBuf2);
				  if(pResp) {
						ddocTsResp_GetSignerCertSerial(pResp, &mbuf1);
						ddocDebug(3, "selectCertIdAndValueTypes", "TSA cert: %s this cert %s -> TSA", 
					(const char*)mbuf1.pMem, cid->szIssuerSerial);
						if(!strcmp(cid->szIssuerSerial, (const char*)mbuf1.pMem)) {
				// yes this is a TSA cert
				cid->nType = CERTID_TYPE_TSA_CERTID;
				cval1->nType = CERTID_VALUE_TSA_CERT;
				snprintf(buf1, sizeof(buf1), "%s-TSA_CERTINFO", pSigInfo->szId);
				ddocCertID_SetId(cid, buf1);
				break;
						}
						ddocMemBuf_free(&mbuf1);
						TS_RESP_free(pResp);
						pResp = 0;
				  }
				  }
				  ddocDebug(3, "selectCertIdAndValueTypes", "done with TS: %s", pTS->szId);
				}
			} // if not a responders cert
#endif
      }
    } // for certids
  }
  return err;
}


//--------------------------------------------------
// handles the start of a <EncapsulatedOCSPValue> element.
// Records the OCSP response data
// pctx - pointer to XML parsing work-area
// name - tag name
//--------------------------------------------------
// FIXME : error handling
void handleEndEncapsulatedOCSPValueBDOC(SigDocParse* pctx, const xmlChar *name)
{
  OCSP_RESPONSE* pResp;
  NotaryInfo* pNotInfo = 0;
  SignatureInfo* pSig = 0;

  // convert the X509 cert data to
  // cert and replace the pointer value
  pNotInfo = ddocGetLastNotaryInfo(pctx->pSigDoc);
  RETURN_VOID_IF_NULL(pNotInfo); // VS: 1.76 correct the check
  pSig = ddocGetLastSignature(pctx->pSigDoc);
  RETURN_VOID_IF_NULL(pSig);
  /*pctx->errcode =*/ ddocDecodeOCSPResponsePEMData(&pResp, 
	  (const byte*)pctx->mbufElemData.pMem, (int)pctx->mbufElemData.nLen);
  // cleanup
  bdocSaxParseEndCollecting(pctx, FLAG_XML_ELEM, 0);
  if(pResp)
    pctx->errcode = initializeNotaryInfoWithOCSP(pctx->pSigDoc, pNotInfo, pResp, ddocSigInfo_GetOCSPRespondersCert(pSig), 0);
		OCSP_RESPONSE_free(pResp);
	if(pctx->errcode)
    return;
  selectCertIdAndValueTypesBDoc(pSig);
}


//--------------------------------------------------
// handles the start of a <SignedProperties> element.
// Records the id and target attributes 
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartSignedPropertiesBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  if(!pctx->nIgnoreDataFile)
    bdocSaxParseStartCollecting(pctx, FLAG_SIG_PART, 0);
  strncpy(pctx->ctx2, bdocSaxParseFindAttrib(atts, "Id", ""), sizeof(pctx->ctx2)-1);
  strncpy(pctx->ctx3, bdocSaxParseFindAttrib(atts, "Target", ""), sizeof(pctx->ctx3)-1);
}

//--------------------------------------------------
// handles the end of a <SignedProperties> element.
// Records the digest of this element 
// pctx - pointer to XML parsing work-area
// name - tag name
//--------------------------------------------------
void handleEndSignedPropertiesBDoc(SigDocParse* pctx, const xmlChar *name,const xmlChar *prefix, const xmlChar *nsUri)
{
  char* pTmp2;
  char buf[DIGEST_LEN256+3], buf2[40];
  int l2;
  SignatureInfo* pSigInfo;
	SHA256_CTX ctx;

  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  if(pctx->mbufSigPartData.pMem && pSigInfo) {
    //ddocDebug(5, "handleEndSignedProperties", "DATA: %s ->%s", 
	     // pctx->mbufSigPartData.nLen, (char*)pctx->mbufSigPartData.pMem);
    //ddocDebugWriteFile(0, "sigprop-can1.txt", &(pctx->mbufSigPartData));
    pTmp2 = canonicalizeXMLBlock((char*)pctx->mbufCanonSigData.pMem, pctx->mbufCanonSigData.nLen, 
		"(//xs:SignedProperties/descendant-or-self::* | //xs:SignedProperties/descendant-or-self::text() | //xs:SignedProperties//attribute::*  | //xs:SignedProperties//namespace::* )[not(self::comment())]", NULL, NULL);
	/*if(pTmp2) {
		mbuf1.pMem = pTmp2;
		mbuf1.nLen = strlen(pTmp2);
		ddocDebugWriteFile(0, "sigprop-can2.txt", &mbuf1);
	}*/
    //dumpInFile("sigprop-can2.txt", pTmp2);
	ddocDebug(5, "handleEndSignedProperties", "canonicalizeXMLBlock finished");
    if(pTmp2) {
      //mbuf1.pMem = pTmp2;
      //mbuf1.nLen = strlen(pTmp2);
      //ddocDebugWriteFile(0, "sigprop-can2.txt", &mbuf1);
      //ddocDebug(5, "handleEndSignedProperties", "HASH over: \n---\n%s\n---\n", 
	     // pTmp2);
			if(!strcmp(pSigInfo->szDigestType,DIGEST_SHA256_NAME)){
				SHA256_Init(&ctx);
				SHA256_Update(&ctx, pTmp2, strlen(pTmp2));
				SHA256_Final((unsigned char*)buf,&ctx);
				ddocSigInfo_SetSigPropRealDigest(pSigInfo, buf, DIGEST_LEN256);
			}else{
				SHA1_Init(&(pctx->sctx));
				SHA1_Update(&(pctx->sctx), pTmp2, strlen(pTmp2));
				SHA1_Final((unsigned char*)buf,&(pctx->sctx));
				ddocSigInfo_SetSigPropRealDigest(pSigInfo, buf, DIGEST_LEN);}
	  l2 = 40;
	  encode((const byte*)buf, DIGEST_LEN, (byte*)buf2, &l2);
	  ddocDebug(5, "handleEndSignedProperties", "SigProp hash: %s", buf2);
      free(pTmp2);
    }
    bdocSaxParseEndCollecting(pctx, FLAG_SIG_PART, 0);
    pctx->checkUTF8 = 0;
  }
}


//--------------------------------------------------
// handles the start of a <SignedInfo> element.
// Starts recording data for digest calculation 
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartSignedInfoBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  if(!pctx->nIgnoreDataFile) {
    // mark the start of data collect mode
    pctx->errcode = bdocSaxParseStartCollecting(pctx, FLAG_SIG_PART, 0);
    pctx->checkUTF8 = 1;
  }
}

//--------------------------------------------------
// handles the end of a <SignedInfo> element.
// Records the digest of this element 
// pctx - pointer to XML parsing work-area
// name - tag name
//--------------------------------------------------
void handleEndSignedInfoBDoc(SigDocParse* pctx, const xmlChar *name,const xmlChar *prefix, const xmlChar *nsUri)
{
  char* pTmp = 0;
  char buf[DIGEST_LEN256+3], buf1[300], buf2[100];
  int l2;
  SignatureInfo* pSigInfo;
  DigiDocMemBuf mbuf1;
  mbuf1.pMem = 0;
  mbuf1.nLen = 0;
  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  if(pctx->mbufSigPartData.pMem && pSigInfo) {
	ddocDebug(3, "handleEndSignedInfoBDoc", "Tag: %s uri %s can-len: %d", name, (prefix ? (char*)prefix : ""), pctx->mbufCanonSigData.nLen);
	ddocDebug(5, "handleEndSignedInfoBDoc", "DATA: %d ->%s", 
	      pctx->mbufSigPartData.nLen, (char*)pctx->mbufSigPartData.pMem);
	//ddocDebugWriteFile(0, "siginf1.txt", &(pctx->mbufSigPartData));
	//AM 18.11.2009 change for Smartlink bdoc
	snprintf(buf1, sizeof(buf1), "(//%s/descendant-or-self::* | //%s/descendant-or-self::text() | //%s//attribute::*  | //%s//namespace::* )[not(self::comment())]", name, name, name, name);	
    if(strcmp(name,"SignedInfo")) {
	  ddocDebug(5, "handleEndSignedInfoBDoc", "canonicalizeXMLBlock len: %d xpath:\n---\n%s\n---\n data:\n---\n%s\n---\n",
		  pctx->mbufCanonSigData.nLen, buf1, (char*)pctx->mbufCanonSigData.pMem);
	  pTmp = canonicalizeXMLBlock((char*)pctx->mbufCanonSigData.pMem, pctx->mbufCanonSigData.nLen, buf1, (char*)prefix, (char*)nsUri);
	} else {
	  ddocDebug(5, "handleEndSignedInfoBDoc", "canonicalizeXML len: %d data:\n---\n%s\n---\n", pctx->mbufSigPartData.nLen, (char*)pctx->mbufSigPartData.pMem);
	  pTmp = canonicalizeXML((char*)pctx->mbufSigPartData.pMem, pctx->mbufSigPartData.nLen);
	  ddocDebugWriteFile(3, "siginf-sigpart.txt", &pctx->mbufSigPartData);
	}
	if(pTmp) {
		mbuf1.pMem = pTmp;
		mbuf1.nLen = strlen(pTmp);
		ddocDebugWriteFile(3, "siginf-can2.txt", &mbuf1);
		ddocDebug(5, "handleEndSignedInfoBDoc", "CANONICALIZED: len: %d data:\n---\n%s\n---\n", strlen(pTmp), (char*)pTmp);
	}
	else
		ddocDebug(1, "handleEndSignedInfoBDoc", "CANONICALIZED: NULL");
    if(pTmp) {//AM change when rsa-sha256 is supported
		l2 = sizeof(buf2);
		if(!strcmp(pctx->ctx3,"aadda")){//DIGEST_SHA256_NAME
				SHA256_CTX ctx;
				SHA256_Init(&ctx);
				SHA256_Update(&ctx, pTmp, strlen(pTmp));
				SHA256_Final((unsigned char*)buf,&ctx);
				ddocSigInfo_SetSigInfoRealDigest(pSigInfo, buf, DIGEST_LEN256);
				encode((const byte*)buf, DIGEST_LEN256, (byte*)buf2, &l2);
		} else {
				SHA1_Init(&(pctx->sctx));
				SHA1_Update(&(pctx->sctx), pTmp, strlen(pTmp));
				SHA1_Final((unsigned char*)buf,&(pctx->sctx));
				ddocSigInfo_SetSigInfoRealDigest(pSigInfo, buf, DIGEST_LEN);
				encode((const byte*)buf, DIGEST_LEN, (byte*)buf2, &l2);
		}
      ddocMemBuf_free(&mbuf1);
	  ddocDebug(3, "handleEndSignedInfoBDoc", "SigedInfo hash: %s", buf2);
    }
    pctx->errcode = bdocSaxParseEndCollecting(pctx, FLAG_SIG_PART, 0);
    pctx->checkUTF8 = 0;
  }
}


//--------------------------------------------------
// handles the start of a <Cert> element.
// Records the id attribute
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartCertBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  char *p = 0, buf1[50];
  SignatureInfo* pSigInfo = NULL;
  CertID *pCertID = NULL;

  // used in old formats
  p = (char*)bdocSaxParseFindAttrib(atts, "Id", NULL);
  if(p)
    strncpy(pctx->ctx2, p, sizeof(pctx->ctx2)-1);
  pSigInfo = ddocGetLastSignature(pctx->pSigDoc); 
  ddocDebug(3, "handleStartCert", "Sig: %s", (pSigInfo ? pSigInfo->szId : "NULL"));
  RETURN_VOID_IF_NULL(pSigInfo);  
  // only if we are in Notary since signers cert-id is handled by <SigningCertificate>
  if(pSigInfo->pNotary) {
    // don't know what cert-id it's going to be
    ddocCertID_new(&pCertID, CERTID_TYPE_UNKNOWN, 0, 0, 0, 0, 0);
    if(!pSigInfo->pCertIDs)
      ddocCertIDList_new(&(pSigInfo->pCertIDs));
    ddocCertIDList_addCertID(pSigInfo->pCertIDs, pCertID);
    snprintf(buf1, sizeof(buf1), "%s-UNKNOWN_CERTINFO", pSigInfo->szId);
    strncpy(pctx->ctx2, buf1, sizeof(pctx->ctx2)-1);
  }
}


//--------------------------------------------------
// handles the start of a <SigningCertificate> element.
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartSigningCertificateBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  SignatureInfo* pSigInfo = NULL;
  CertID *pCertID = NULL;
	
  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  ddocDebug(3, "handleStartSigningCertificate", "Sig: %s", 
	    (pSigInfo ? pSigInfo->szId : "NULL"));
  RETURN_VOID_IF_NULL(pSigInfo);
  pCertID = ddocSigInfo_GetOrCreateCertIDOfType(pSigInfo, CERTID_TYPE_SIGNERS_CERTID); 
  snprintf(pctx->ctx2, sizeof(pctx->ctx2)-1, "%s-CERTINFO", pSigInfo->szId);
}

//--------------------------------------------------
// handles the start of a <UnsignedSignatureProperties> element.
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartUnsignedSignaturePropertiesBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  SignatureInfo* pSigInfo = NULL;
  NotaryInfo *pNotaryInfo;
	
  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  ddocDebug(3, "handleStartUnsignedSignatureProperties", "Sig: %s", 
	    (pSigInfo ? pSigInfo->szId : "NULL"));
  RETURN_VOID_IF_NULL(pSigInfo);
  (void)NotaryInfo_new(&pNotaryInfo, pctx->pSigDoc, pSigInfo);
}


//--------------------------------------------------
// handles the start of a <CompleteCertificateRefs> element.
// Records the signature id so that we can later
// capture responders cert digest
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartCompleteCertificateRefsBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  SignatureInfo* pSigInfo = 0;

  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  ddocDebug(3, "handleStartCompleteCertificateRefs", "Sig: %s", (pSigInfo ? pSigInfo->szId : "NULL"));
  RETURN_VOID_IF_NULL(pSigInfo);
  snprintf(pctx->ctx2, sizeof(pctx->ctx2)-1, "%s-RESPONDER_CERTINFO", pSigInfo->szId);
}


//--------------------------------------------------
// handles the start of a <OCSPIdentifier> element.
// Records the id attribute
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartOCSPIdentifierBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  strncpy(pctx->ctx2, bdocSaxParseFindAttrib(atts, "URI", ""), sizeof(pctx->ctx2)-1);
}

//--------------------------------------------------
// handles the start of a <X509Certificate> element.
// Records the digest method
// pctx - pointer to XML parsing work-area
// name - tag name
//--------------------------------------------------
void handleEndX509CertificateBDoc(SigDocParse* pctx, const xmlChar *name)
{
  X509* x509 = 0;
  SignatureInfo* pSigInfo = 0;

  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  // convert the X509 cert data to
  // cert and replace the pointer value
  if(pSigInfo && pctx->mbufElemData.pMem) {
    pctx->errcode = ddocDecodeX509PEMData(&x509, 
	 (const char*)pctx->mbufElemData.pMem, (int)pctx->mbufElemData.nLen);
    if(x509)
      ddocSigInfo_SetSignersCert(pSigInfo, x509);
    // cleanup
    pctx->errcode = bdocSaxParseEndCollecting(pctx, FLAG_XML_ELEM, 0);
  } 
}

//--------------------------------------------------
// handles the start of a <EncapsulatedX509Certificate> element.
// Records the signature id so that we can later
// capture responders cert digest
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartEncapsulatedX509CertificateBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  SignatureInfo* pSigInfo = 0;
  const char *p1 = 0;
  CertValue *pCertValue;

  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  p1 = bdocSaxParseFindAttrib(atts, "Id", NULL);

  ddocDebug(3, "handleStartEncapsulatedX509Certificate", "Sig: %s, type: %s", 
	    (pSigInfo ? pSigInfo->szId : "NULL"), (p1 ? p1 : "NULL"));
  RETURN_VOID_IF_NULL(pSigInfo);
  if(p1) {
    if(strstr(p1, "TSA_CERT") || strstr(p1, "TSA-CERT"))
      pCertValue = ddocSigInfo_GetOrCreateCertValueOfType(pSigInfo, CERTID_VALUE_TSA_CERT);
    if(strstr(p1, "RESPONDER_CERT") || strstr(p1, "RESPONDER-CERT")){
      pCertValue = ddocSigInfo_GetOrCreateCertValueOfType(pSigInfo, CERTID_VALUE_RESPONDERS_CERT);
	  
	}
    if(strstr(p1, "CA_CERT") || strstr(p1, "CA-CERT"))
      pCertValue = ddocSigInfo_GetOrCreateCertValueOfType(pSigInfo, CERTID_VALUE_CA_CERT);
  }
  snprintf(pctx->ctx2, sizeof(pctx->ctx2)-1, "%s", p1);
  //AM test
  ddocCertValue_SetId(pCertValue,p1);
  pctx->errcode = bdocSaxParseStartCollecting(pctx, FLAG_XML_ELEM, 1);
}

//--------------------------------------------------
// handles the end of a <EncapsulatedX509Certificate> element.
// Records the digest method
// pctx - pointer to XML parsing work-area
// name - tag name
//--------------------------------------------------
void handleEndEncapsulatedX509CertificateBDoc(SigDocParse* pctx, const xmlChar *name)
{
  X509* x509 = 0;
  NotaryInfo* pNotInf = 0;
  SignatureInfo* pSigInfo = 0;
  CertValue* pCertValue = 0;

  // convert the X509 cert data to cert<
  pNotInf = ddocGetLastNotaryInfo(pctx->pSigDoc);
  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  pctx->errcode = ddocDecodeX509PEMData(&x509, 
       (const char*)pctx->mbufElemData.pMem, (int)pctx->mbufElemData.nLen);
  if(x509 && pSigInfo) {
    pCertValue = ddocSigInfo_GetLastCertValue(pSigInfo);
    if(pCertValue)
      ddocCertValue_SetCert(pCertValue, x509);
    else
      X509_free(x509); // not found, free it
  }
  // cleanup
  bdocSaxParseEndCollecting(pctx, FLAG_XML_ELEM, 0);
}

#ifdef WITH_TS
//--------------------------------------------------
// handles the start of a <SignatureTimeStamp> element.
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartSignatureTimeStampBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  SignatureInfo* pSigInfo = 0;
  const char *p1 = 0;
  TimestampInfo *pTS;

  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  p1 = bdocSaxParseFindAttrib(atts, "Id", NULL);
  ddocDebug(3, "handleStartSignatureTimeStamp", "Sig: %s, type: %s", 
	    (pSigInfo ? pSigInfo->szId : "NULL"), (p1 ? p1 : "NULL"));
  RETURN_VOID_IF_NULL(pSigInfo);
  pTS = ddocSigInfo_GetOrCreateTypestampOfType(pSigInfo, TS_TYPE_SIGNATURE);
  if(p1)
    ddocTimestampInfo_SetId(pTS, p1);
  // free possible old xml data here
  //ddocMemBuf_free(&(pctx->mbufTsInput));
}

//--------------------------------------------------
// handles the start of a <SigAndRefsTimeStamp> element.
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartSigAndRefsTimeStampBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  SignatureInfo* pSigInfo = 0;
  const char *p1 = 0;
  TimestampInfo *pTS;

  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  p1 = bdocSaxParseFindAttrib(atts, "Id", NULL);
  ddocDebug(3, "handleStartSigAndRefsTimeStamp", "Sig: %s, type: %s", 
	    (pSigInfo ? pSigInfo->szId : "NULL"), (p1 ? p1 : "NULL"));
  RETURN_VOID_IF_NULL(pSigInfo);
  pTS = ddocSigInfo_GetOrCreateTypestampOfType(pSigInfo, TS_TYPE_SIG_AND_REFS);
  if(p1)
    ddocTimestampInfo_SetId(pTS, p1);
}


//--------------------------------------------------
// handles the start of a <Include> element.
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleStartIncludeBDoc(SigDocParse* pctx, const xmlChar *name, const xmlChar **atts)
{
  SignatureInfo* pSigInfo = 0;
  const char *p1 = 0;
  TimestampInfo *pTS;

  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  pTS = ddocSigInfo_GetLastTypestamp(pSigInfo);
  p1 = bdocSaxParseFindAttrib(atts, "URI", NULL);
  ddocDebug(3, "handleStartInclude", "Sig: %s, TS: %s URI: %s", 
	    (pSigInfo ? pSigInfo->szId : "NULL"), (pTS ? pTS->szId : ""), 
	    (p1 ? p1 : "NULL"));
  RETURN_VOID_IF_NULL(pSigInfo);
  RETURN_VOID_IF_NULL(pTS);
  RETURN_VOID_IF_NULL(p1);
  ddocTimestampInfo_AddIncludeInfo(pTS, p1);
}

//--------------------------------------------------
// handles the start of a <EncapsulatedTimeStamp> element.
// Records the timestamp data
// pctx - pointer to XML parsing work-area
// name - tag name
//--------------------------------------------------
void handleEndEncapsulatedTimeStampBDoc(SigDocParse* pctx, const xmlChar *name)
{
  SignatureInfo* pSigInfo = 0;
  TimestampInfo *pTS;
  DigiDocMemBuf mbuf1;

  mbuf1.pMem = 0;
  mbuf1.nLen = 0;
  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  pTS = ddocSigInfo_GetLastTypestamp(pSigInfo);
  RETURN_VOID_IF_NULL(pSigInfo);
  RETURN_VOID_IF_NULL(pTS);
  ddocDecodeBase64Data(pctx->mbufElemData.pMem, pctx->mbufElemData.nLen, &mbuf1);
  ddocTimestampInfo_SetTS(pTS, (const char*)mbuf1.pMem, mbuf1.nLen);
  // cleanup
  bdocSaxParseEndCollecting(pctx, FLAG_XML_ELEM, 0);
  ddocMemBuf_free(&mbuf1);
}

//--------------------------------------------------
// handles the end of a <SignatureTimeStamp> element.
// Records the xml data used to calculate timesamp imprint.
// pctx - pointer to XML parsing work-area
// name - tag name
//--------------------------------------------------
void handleEndSignatureTimeStampBDoc(SigDocParse* pctx, const xmlChar *name)
{
  char *p1 = 0;

  p1 = (char*)canonicalizeXML((char*)pctx->mbufTsInpElem.pMem, pctx->mbufTsInpElem.nLen);
  ddocDebugWriteFile(4, "ts-sigts-txt.txt", &(pctx->mbufTsInpElem));
  if(p1) {
	  // keep the canonicalized xml as we need it also for SigAndRefsTimeStamp hash
	  ddocMemAppendData(&(pctx->mbufTsInput), p1, -1);
	  ddocMemAppendData(&(pctx->mbufTsInput), "\n", -1);
	  free(p1);
      p1 = 0;
  }
  // cleanup
  bdocSaxParseEndCollecting(pctx, FLAG_TS_INP, 0);
}

//--------------------------------------------------
// handles the end of a <CompleteCertificateRefs> element.
// Records the xml data used to calculate timesamp imprint.
// pctx - pointer to XML parsing work-area
// name - tag name
//--------------------------------------------------
void handleEndCompleteCertificateRefsBDoc(SigDocParse* pctx, const xmlChar *name)
{
  char *p1 = 0;

  p1 = (char*)canonicalizeXML((char*)pctx->mbufTsInpElem.pMem, pctx->mbufTsInpElem.nLen);
  ddocDebugWriteFile(4, "ts-completecertrefs-txt.txt", &(pctx->mbufTsInpElem));  
  if(p1) {
	  // keep the canonicalized xml as we need it also for SigAndRefsTimeStamp hash
	  ddocMemAppendData(&(pctx->mbufTsInput), p1, -1);
	  free(p1);
      p1 = 0;
  }
  // cleanup
  bdocSaxParseEndCollecting(pctx, FLAG_TS_INP, 0);
}

//--------------------------------------------------
// handles the end of a <CompleteRevocationRefs> element.
// Records the xml data used to calculate timesamp imprint.
// pctx - pointer to XML parsing work-area
// name - tag name
//--------------------------------------------------
void handleEndCompleteRevocationRefsBDoc(SigDocParse* pctx, const xmlChar *name)
{
  char *p1 = 0;
  int l1;
  TimestampInfo *pTsInfo = 0;
  SignatureInfo *pSigInfo = 0;
  DigiDocMemBuf mbuf1, mbuf2;

  mbuf1.pMem = 0;
  mbuf1.nLen = 0;
  mbuf2.pMem = 0;
  mbuf2.nLen = 0;
  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  if(pctx->pSigDoc->szProfile) {
		if(!strcmp(pctx->pSigDoc->szProfile,"TS")){
		  p1 = (char*)canonicalizeXML((char*)pctx->mbufTsInpElem.pMem, pctx->mbufTsInpElem.nLen);
			ddocDebugWriteFile(4, "ts-completerevocrefs-txt.txt", &(pctx->mbufTsInpElem));
			if(p1) {
			// keep the canonicalized xml as we need it also for SigAndRefsTimeStamp hash
			ddocMemAppendData(&(pctx->mbufTsInput), p1, -1);
			free(p1);
				p1 = 0;
			// debug write hash input to a file
			ddocDebugWriteFile(4, "ts-sigandrefs-inp.txt", &(pctx->mbufTsInput));
			// calculate hash
			l1 = 25;
			ddocMemSetLength(&mbuf1, l1);
			calculateDigest((const byte*)pctx->mbufTsInput.pMem, pctx->mbufTsInput.nLen, DIGEST_SHA1, (byte*)mbuf1.pMem, &l1);
			mbuf1.nLen = l1;
			pTsInfo = ddocSigInfo_GetOrCreateTypestampOfType(pSigInfo, TS_TYPE_SIG_AND_REFS);
			if(pTsInfo)
			ddocTimestampInfo_SetRealDigest(pTsInfo, mbuf1.pMem, mbuf1.nLen);
			ddocEncodeBase64(&mbuf1, &mbuf2);
			ddocDebug(3, "handleEndCompleteRevocationRefs", "Calculated SigAndRefsTimeStamp hash: %s", (const char*)mbuf2.pMem);
			ddocMemBuf_free(&mbuf1);
			ddocMemBuf_free(&mbuf2);
			}
		}
  }
  // cleanup
  bdocSaxParseEndCollecting(pctx, FLAG_TS_INP, 0);
}

#endif // WITH_TS


//--------------------------------------------------
// handles the content of a <DigestValue> element.
// Decodes and reads in digest value
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
void handleEndDigestValueBDoc(SigDocParse* pctx, const xmlChar *name) 
{
  char id[300], type[20];
  SignatureInfo* pSigInfo = NULL;
  NotaryInfo* pNotInfo = NULL;
  DocInfo* pDocInfo = NULL;
  CertID* pCertID = NULL;
  DigiDocMemBuf mbuf1;

  mbuf1.pMem = 0;
  mbuf1.nLen = 0;
  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  ddocDebug(4, "handleDigestValue", "DF: %s value: %s len: %d", 
	  pctx->ctx2, (char*)pctx->mbufElemData.pMem, pctx->mbufElemData.nLen);
  // decode digest value
  ddocDecodeBase64(&(pctx->mbufElemData), &mbuf1);
  ddocDebug(4, "handleDigestValue", "decoded len: %ld", mbuf1.nLen);
  // cleanup
  bdocSaxParseEndCollecting(pctx, FLAG_XML_ELEM, 0);
  // find current signature
  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  RETURN_VOID_IF_NULL(pSigInfo);
	if(pctx->ctx2[0] == '/')
	  strncpy(id,pctx->ctx2,sizeof(id));
	else
	  decodeURI(pctx->ctx2, id, sizeof(id), type, sizeof(type));
  ddocDebug(4, "handleDigestValue", "ctx2: %s id: %s type: %s", pctx->ctx2, id, type);
  if(id[0] == '/' || !strcmp(id, "null")) {
    RETURN_VOID_IF_NULL(pSigInfo);
    pDocInfo = getDocInfoWithId(pSigInfo, id);
    if(!strcmp(type, "MimeType") || 
       !strcmp(type, "MIME")) {
      if(pDocInfo == NULL) {
	    addDocInfo(&pDocInfo, pSigInfo, id, pctx->ctx3,
		   NULL, 0, (const byte*)mbuf1.pMem, mbuf1.nLen);
      } else
	    setDocInfoMimeDigest(pDocInfo, (const byte*)mbuf1.pMem, mbuf1.nLen);
    }
    else {
      if(pDocInfo == NULL) {
	     addDocInfo(&pDocInfo, pSigInfo, id, pctx->ctx3,
		   (const byte*)mbuf1.pMem, mbuf1.nLen, NULL, 0); // MEMLEAK: ??? //AM memleak of 4 bytes 
      } else
	     setDocInfoDigest(pDocInfo, (const byte*)mbuf1.pMem, mbuf1.nLen, pctx->ctx3);
    }
  }
  if(id[0] == 'S') {
    if(!strcmp(type, "SignedProperties")) {
      RETURN_VOID_IF_NULL(pSigInfo);
      ddocSigInfo_SetSigPropDigest(pSigInfo, (const char*)mbuf1.pMem, mbuf1.nLen);
			if(!strcmp(pctx->ctx3,DIGEST_SHA256_NAME))
				setString((char**)&(pSigInfo->szDigestType), DIGEST_SHA256_NAME, -1);
			else
				setString((char**)&(pSigInfo->szDigestType), DIGEST_SHA1_NAME, -1);
    }
    if(!strcmp(type, "CERTINFO")) {
      RETURN_VOID_IF_NULL(pSigInfo);
      pCertID = ddocSigInfo_GetOrCreateCertIDOfType(pSigInfo, CERTID_TYPE_SIGNERS_CERTID);
      RETURN_VOID_IF_NULL(pCertID);
      ddocCertID_SetDigestValue(pCertID, (const char*)mbuf1.pMem, mbuf1.nLen);
			if(!strcmp(pctx->ctx3,DIGEST_SHA256_NAME))
				setString((char**)&(pCertID->szDigestType), DIGEST_SHA256_NAME, -1);
			else
				setString((char**)&(pCertID->szDigestType), DIGEST_SHA1_NAME, -1);
    }
    if(!strcmp(type, "UNKNOWN_CERTINFO")) {
      pCertID = ddocCertIDList_GetLastCertID(pSigInfo->pCertIDs);
      RETURN_VOID_IF_NULL(pCertID);
      ddocCertID_SetDigestValue(pCertID, (const char*)mbuf1.pMem, mbuf1.nLen);
			if(!strcmp(pctx->ctx3,DIGEST_SHA256_NAME))
				setString((char**)&(pCertID->szDigestType), DIGEST_SHA256_NAME, -1);
			else
				setString((char**)&(pCertID->szDigestType), DIGEST_SHA1_NAME, -1);
    }    
  }
  if(id[0] == 'N') {
    pNotInfo = getNotaryWithId(pctx->pSigDoc, id);
    RETURN_VOID_IF_NULL(pNotInfo);
    ddocNotInfo_SetOcspDigest(pNotInfo, (const char*)mbuf1.pMem, mbuf1.nLen);
		if(!strcmp(pctx->ctx3,DIGEST_SHA256_NAME))
			setString((char**)&(pNotInfo->szDigestType), DIGEST_SHA256_NAME, -1);
		else
			setString((char**)&(pNotInfo->szDigestType), DIGEST_SHA1_NAME, -1);
  }
  ddocMemBuf_free(&mbuf1);
}


//--------------------------------------------------
// handles the content of a <SigningTime> element.
// Reads in timestamp data
// pctx - pointer to XML parsing work-area
// value - character values read from file
// len - length of chars ???
//--------------------------------------------------
void handleEndSigningTimeBDoc(SigDocParse* pctx, const xmlChar *name) 
{
  SignatureInfo *pSigInfo;

  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  RETURN_VOID_IF_NULL(pSigInfo);
  if(pctx->mbufElemData.pMem) {
    setString(&(pSigInfo->szTimeStamp), (char*)pctx->mbufElemData.pMem, -1);
    bdocSaxParseEndCollecting(pctx, FLAG_XML_ELEM, 0);
  }	
}

#define ADR_ENTRY_CITY   1
#define ADR_ENTRY_STATE  2
#define ADR_ENTRY_COUNTRY 3
#define ADR_ENTRY_ZIP    4

//--------------------------------------------------
// Handles address entry
// pctx - pointer to XML parsing work-area
// 
// value - character values read from file
// len - length of chars ???
//--------------------------------------------------
void handleAdrEntryBDoc(SigDocParse* pctx, int nAdr)
{
  SignatureInfo* pSigInfo;
	
  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
  if(pSigInfo && pctx->mbufElemData.pMem) {
    switch(nAdr) {
    case ADR_ENTRY_CITY: 
      setString(&(pSigInfo->sigProdPlace.szCity), (const char*)pctx->mbufElemData.pMem, -1); 
      break;
    case ADR_ENTRY_STATE: 
      setString(&(pSigInfo->sigProdPlace.szStateOrProvince), (const char*)pctx->mbufElemData.pMem, -1); 
      break;
    case ADR_ENTRY_COUNTRY: 
      setString(&(pSigInfo->sigProdPlace.szCountryName), (const char*)pctx->mbufElemData.pMem, -1); 
      break;
    case ADR_ENTRY_ZIP: 
      setString(&(pSigInfo->sigProdPlace.szPostalCode), (const char*)pctx->mbufElemData.pMem, -1); 
      break;
    }
  }
  bdocSaxParseEndCollecting(pctx, FLAG_XML_ELEM, 0);
}

//--------------------------------------------------
// handles the start of a <ClaimedRole> element.
// Stores the collected claimed role
// pctx - pointer to XML parsing work-area
// name - tag name
//--------------------------------------------------
void handleEndClaimedRoleBDoc(SigDocParse* pctx, const xmlChar *name)
{
  SignatureInfo* pSigInfo;

  if(pctx->mbufElemData.pMem) {
    pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
    RETURN_VOID_IF_NULL(pSigInfo);
    addSignerRole(pSigInfo, 0, (const char*)pctx->mbufElemData.pMem, -1, 0);
    bdocSaxParseEndCollecting(pctx, FLAG_XML_ELEM, 0);
    pctx->bNoXMLElemData = 0;
  }
}

//--------------------------------------------------
// handles the start of a <CertifiedRole> element.
// Stores the collected certified role
// pctx - pointer to XML parsing work-area
// name - tag name
//--------------------------------------------------
void handleEndCertifiedRoleBDoc(SigDocParse* pctx, const xmlChar *name)
{
  SignatureInfo* pSigInfo;
	
  if(pctx->mbufElemData.pMem) {
    pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
    RETURN_VOID_IF_NULL(pSigInfo);
    addSignerRole(pSigInfo, 1, (const char*)pctx->mbufElemData.pMem, -1, 0);
    bdocSaxParseEndCollecting(pctx, FLAG_XML_ELEM, 0);
  }
}

//--------------------------------------------------
// handles the end of a <IssuerSerial> element.
// Reads in cert isseur serial number
// pctx - pointer to XML parsing work-area
// value - character values read from file
// len - length of chars ???
//--------------------------------------------------
void handleEndIssuerSerialBDoc(SigDocParse* pctx, const xmlChar *name) 
{
  SignatureInfo* pSigInfo;
  char  id[20], type[20];
  CertID* pCertID;

  decodeURI(pctx->ctx2, id, sizeof(id), type, sizeof(type));
  if(!strcmp(type, "CERTINFO")) {
    pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
    pCertID = ddocSigInfo_GetCertIDOfType(pSigInfo, CERTID_TYPE_SIGNERS_CERTID);
    if(pCertID && pctx->mbufElemData.pMem)
      ddocCertID_SetIssuerSerial(pCertID, (char*)pctx->mbufElemData.pMem);
  }
  if(!strcmp(type, "UNKNOWN_CERTINFO")) {
    pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
    pCertID = ddocCertIDList_GetLastCertID(pSigInfo->pCertIDs);
    if(pCertID && pctx->mbufElemData.pMem)
      ddocCertID_SetIssuerSerial(pCertID, (char*)pctx->mbufElemData.pMem);
  }
  bdocSaxParseEndCollecting(pctx, FLAG_XML_ELEM, 0);
}

//--------------------------------------------------
// handles the end of a <IssuerName> element.
// Reads in cert isseur serial number
// pctx - pointer to XML parsing work-area
// value - character values read from file
// len - length of chars ???
//--------------------------------------------------
void handleEndIssuerNameBDoc(SigDocParse* pctx, const xmlChar *name) 
{
  SignatureInfo* pSigInfo;
  char  id[20], type[20];
  CertID* pCertID;

  decodeURI(pctx->ctx2, id, sizeof(id), type, sizeof(type));
  if(!strcmp(type, "CERTINFO")) {
    pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
    pCertID = ddocSigInfo_GetCertIDOfType(pSigInfo, CERTID_TYPE_SIGNERS_CERTID);
    if(pCertID && pctx->mbufElemData.pMem)
      ddocCertID_SetIssuerName(pCertID, (char*)pctx->mbufElemData.pMem);
  }
  if(!strcmp(type, "UNKNOWN_CERTINFO")) {
    pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
    pCertID = ddocCertIDList_GetLastCertID(pSigInfo->pCertIDs);
    if(pCertID && pctx->mbufElemData.pMem)
      ddocCertID_SetIssuerName(pCertID, (char*)pctx->mbufElemData.pMem);
  }
  bdocSaxParseEndCollecting(pctx, FLAG_XML_ELEM, 0);
}

//--------------------------------------------------
// tests if this is a textform responder id
// we base the test on existence of certain
// key elements in a DN.
// return 1 if it is text form
//--------------------------------------------------
int isTextResponderIdBDoc(const char* szRespId)
{
  return strstr(szRespId, "CN=") && strstr(szRespId, "C=");
}

//--------------------------------------------------
// handles the end of a </ResponderId> or <ByName> element.
// Reads in ResponderId data
// pctx - pointer to XML parsing work-area
// value - character values read from file
// len - length of chars ???
//--------------------------------------------------
void handleEndResponderIDBDoc(SigDocParse* pctx, const xmlChar *name) 
{
  NotaryInfo* pNotInf;
  DigiDocMemBuf mbuf1;

  mbuf1.pMem = 0;
  mbuf1.nLen = 0;
  pNotInf = ddocGetLastNotaryInfo(pctx->pSigDoc);
  ddocDebug(3, "handleEndResponderId", "notary: %s id: %s, len: %d", 
	  (pNotInf ? pNotInf->szId : "NULL"), (const char*)pctx->mbufElemData.pMem, pctx->mbufElemData.nLen);
  if(pNotInf && pctx->mbufElemData.pMem) {
      pctx->errcode = ddocNotInfo_SetResponderId(pNotInf, (const char*)pctx->mbufElemData.pMem, pctx->mbufElemData.nLen);
      pNotInf->nRespIdType = RESPID_NAME_TYPE;
    pctx->errcode = bdocSaxParseEndCollecting(pctx, FLAG_XML_ELEM, 0);
  }	
}

//--------------------------------------------------
// handles the end of a <ByKey> element.
// Reads in ResponderId data
// pctx - pointer to XML parsing work-area
// value - character values read from file
// len - length of chars ???
//--------------------------------------------------
void handleEndByKeyBDoc(SigDocParse* pctx, const xmlChar *name) 
{
  NotaryInfo* pNotInf;
  DigiDocMemBuf mbuf1;

  mbuf1.pMem = 0;
  mbuf1.nLen = 0;
  pNotInf = ddocGetLastNotaryInfo(pctx->pSigDoc);
  ddocDebug(3, "handleEndByKey", "notary: %s id: %s, len: %d", 
	  (pNotInf ? pNotInf->szId : "NULL"), (const char*)pctx->mbufElemData.pMem, pctx->mbufElemData.nLen);
  if(pNotInf && pctx->mbufElemData.pMem) {
    ddocDecodeBase64Data(pctx->mbufElemData.pMem, pctx->mbufElemData.nLen, &mbuf1);
    pctx->errcode = ddocNotInfo_SetResponderId(pNotInf, (const char*)mbuf1.pMem, mbuf1.nLen);
    pctx->errcode = bdocSaxParseEndCollecting(pctx, FLAG_XML_ELEM, 0);
    pNotInf->nRespIdType = RESPID_KEY_TYPE;
  }	
  ddocMemBuf_free(&mbuf1);
}

//--------------------------------------------------
// handles the end of a <SignatureValue> element.<
// Decodes the base64 data in a buffer and assigns
// to signature value
// pctx - pointer to XML parsing work-area
// name - tag name
//--------------------------------------------------
void handleEndSignatureValueBDoc(SigDocParse* pctx, const xmlChar* name)
{
  SignatureInfo* pSigInfo;
  DigiDocMemBuf mbuf1, mbuf2;
  char *p1 = 0, *p2 = 0, buf1[500],*buffer = 0;
  int l1,i,len=0;
  //BIO *b64, *bmem;
#ifdef WITH_TS
  TimestampInfo* pTsInfo = 0;
#endif
  mbuf1.pMem = 0;
  mbuf1.nLen = 0;
  mbuf2.pMem = 0;
  mbuf2.nLen = 0;

  pSigInfo = ddocGetLastSignature(pctx->pSigDoc);

#ifdef WITH_TS
  // calculate hash of caninicalized sig-value xml
  p1 = (char*)canonicalizeXML((char*)pctx->mbufElemData.pMem, pctx->mbufElemData.nLen);
	if(p1 && pctx->pSigDoc->szProfile) {
		if(!strcmp(pctx->pSigDoc->szProfile,"TS")){
			pTsInfo = ddocSigInfo_GetOrCreateTypestampOfType(pSigInfo, TS_TYPE_SIGNATURE);
			if(pTsInfo){
				// free possibly collected old timestamp input data (for earlier signatures)
				ddocMemBuf_free(&(pctx->mbufTsInput));
				// keep the canonicalized xml as we need it also for SigAndRefsTimeStamp hash
				ddocMemAppendData(&(pctx->mbufTsInput), p1, -1);
				// debug write hash input to a file
				ddocDebugWriteFile(4, "ts-sigval-inp.txt", &(pctx->mbufTsInput));
				l1 = 37;
				ddocMemSetLength(&mbuf1, l1);
				calculateDigest((const byte*)p1, strlen(p1), DIGEST_SHA1, (byte*)mbuf1.pMem, &l1);
				mbuf1.nLen = l1;

				ddocTimestampInfo_SetRealDigest(pTsInfo, (const char*)mbuf1.pMem, mbuf1.nLen);
				ddocEncodeBase64(&mbuf1, &mbuf2);
				ddocDebug(3, "handleEndSignatureValue", "Calculated SignatureValueTimeStamp hash: %s", (const char*)mbuf2.pMem);

				ddocMemBuf_free(&mbuf1);
				ddocMemBuf_free(&mbuf2);
				}
		}
	}
	free(p1);
	p1 = 0;
#endif
  // decode signature value
	if((const char*)pctx->mbufElemData.pMem){
  p1 = strchr((const char*)pctx->mbufElemData.pMem, '>');
  if(p1)
  p2 = strchr((const char*)p1, '<');
  if(p1 && p2) {
	  p1++;
      *p2 = 0;
	  l1 = strlen(p1) + 10;
	  ddocMemSetLength(&mbuf1, l1);
	  //AM temp code for smartlink base64
	  for(i = 0; i < strlen(p1); i++) {
		if(isb64char(p1[i])) {
		  buf1[len] = p1[i];
		  len++;}

	  }
	  buf1[len] = 0;
	  //ddocDecodeBase64Data((void*)p1, -1, &mbuf1);	
      decode((const byte*)buf1, strlen(buf1), (byte*)mbuf1.pMem, &l1);
	    /*buffer = (char *)malloc(strlen(p1));
		memset(buffer, 0, strlen(p1));

		b64 = BIO_new(BIO_f_base64());
		bmem = BIO_new_mem_buf((unsigned char*)p1, strlen(p1));
		bmem = BIO_push(b64, bmem);

		BIO_read(bmem, buffer, strlen(p1));

		BIO_free_all(bmem);*/
	  //l1 = strlen(buffer);
	  mbuf1.nLen = l1;
      if(pSigInfo && mbuf1.nLen > 0)
					ddocSigInfo_SetSignatureValue(pSigInfo, mbuf1.pMem, mbuf1.nLen);}
	  ddocMemBuf_free(&mbuf1);
	  //ddocMemBuf_free(&mbuf3);
  }
  // cleanup
  bdocSaxParseEndCollecting(pctx, FLAG_TS_INP, 0);	  
}

//--------------------------------------------------
// Collects elements start-tag
// pctx - pointer to XML parsing work-area
// name - xml element tag name
// atts - xml atributes
// pcFlag - pointer to flag governing collection of this element
// pMBuf - DigiDocMemBuf to collect it in
//--------------------------------------------------
int bdocSaxParseCollectStartTag(SigDocParse* pctx, const xmlChar *name, const xmlChar *prefix, const xmlChar **atts,
				char *pcFlag, DigiDocMemBuf* pMBuf)
{
  int i, addXmlns = 0;
  ddocDebug(4, "bdocSaxParseCollectStartTag", "<%s:%s>", (const char *)prefix, (const char *)name);	
  // if we are in collect data mode then
  // record this tag data
  if(*pcFlag) {
	if(!strcmp((const char*)name, "SignedInfo") &&
		(!prefix || !strlen(prefix))) { // must have this attribute
      addXmlns = 1;
	  ddocDebug(5, "bdocSaxParseCollectStartTag", "add xmlns attr");
	}
   pctx->errcode = ddocMemAppendData(pMBuf, "<", -1);
    if(!pctx->errcode)
      pctx->errcode = ddocMemAppendData(pMBuf, (const char*)name, -1);
    for (i = 0; !pctx->errcode && (atts != NULL) && (atts[i] != NULL); i += 2) {
      pctx->errcode = ddocMemAppendData(pMBuf, " ", -1);
      if(!pctx->errcode)
		pctx->errcode = ddocMemAppendData(pMBuf, (const char*)atts[i], -1);
      if(!pctx->errcode)
		pctx->errcode = ddocMemAppendData(pMBuf, "=\"", -1);
      if(!pctx->errcode)
		pctx->errcode = ddocMemAppendData(pMBuf, (const char*)atts[i+1], -1);
      if(!pctx->errcode)
		pctx->errcode = ddocMemAppendData(pMBuf, "\"", -1);
      if(!strcmp((const char*)atts[i], "xmlns"))
		addXmlns = 0; // already has this atribute
    }
	if(addXmlns && !pctx->errcode)
      pctx->errcode = ddocMemAppendData(pMBuf, " xmlns=\"http://www.w3.org/2000/09/xmldsig#\"", -1);
    if(!pctx->errcode)
	  pctx->errcode = ddocMemAppendData(pMBuf, ">", -1);
	if(addXmlns)
		ddocDebug(5, "bdocSaxParseCollectStartTag", "Collected: %d - %s", pMBuf->nLen, (char*)pMBuf->pMem);
    ddocDebug(6, "bdocSaxParseCollectStartTag", "Element tag collected: %d", pMBuf->nLen);
  }
  return pctx->errcode;
}


/**
 * startElementHandler:
 * @ctxt:  An XML parser context
 * @name:  The element name
 *
 * called when an opening tag has been processed.
 */
static void startElementHandlerBDoc(void *ctx, const xmlChar *name, const xmlChar **atts)
{
  SigDocParse* pctx = (SigDocParse*)ctx;
  char *pTmp1 = NULL;
  DigiDocMemBuf mbuf1;
    //AM 12.02.09
  ElementEntry* pCurrElem = 0;
  ddocStackPushElementSAX(&(pctx->stack), name, atts, &pCurrElem);
  strncpy(pctx->tag, (const char*)pCurrElem->tag, sizeof(pctx->tag));
  // do nothing if error has ocurred
  if(hasUnreadErrors()) return;
  ddocDebug(4, "startElementHandler", "<%s>", (const char*)name);
  //printf("<%s>\n", (const char*)name);
  if(!strcmp((const char*)pCurrElem->tag, "SignedProperties")) 
    handleStartSignedPropertiesBDoc(pctx, pCurrElem->tag, atts);
  if(!strcmp((const char*)pCurrElem->tag, "SignedInfo")) 
    handleStartSignedInfoBDoc(pctx, pCurrElem->tag, atts);
  // if we are not in a <DataFile> and we encounter
  // the <Signature> tag then start collecting original content
  if(!pctx->nIgnoreDataFile && 
     (!strcmp((const char*)pCurrElem->tag, "Signature") || pctx->bCollectSigData)) { 
    if(!strcmp((const char*)pCurrElem->tag, "Signature"))
      pctx->errcode = bdocSaxParseStartCollecting(pctx, FLAG_SIGNATURE, 0);
    // collect general <Signature> data<
    bdocSaxParseCollectStartTag(pctx, name, pCurrElem->prefix, atts, &(pctx->bCollectSigData), &(pctx->mbufSigData));
  }
  // we need the tags here because of hash value calculation
#ifdef WITH_TS
  if((pctx && pctx->pSigDoc && pctx->pSigDoc->szFormatVer) &&
	 (!strcmp((const char*)pCurrElem->tag, "SignatureTimeStamp") ||
	  !strcmp((const char*)pCurrElem->tag, "CompleteCertificateRefs") ||
	  !strcmp((const char*)pCurrElem->tag, "CompleteRevocationRefs") ))
       pctx->errcode = bdocSaxParseStartCollecting(pctx, FLAG_TS_INP, 0);
  bdocSaxParseCollectStartTag(pctx, name, pCurrElem->prefix, atts, &(pctx->bCollectTsInput), &(pctx->mbufTsInpElem));
#endif
  // collect general tag data
  if(!strcmp((const char*)pCurrElem->tag, "SignatureValue"))
       pctx->errcode = bdocSaxParseStartCollecting(pctx, FLAG_XML_ELEM, 0);
  bdocSaxParseCollectStartTag(pctx, name, pCurrElem->prefix, atts, &(pctx->bCollectElemData), &(pctx->mbufElemData));
  // use base64 DataFile parsing optimization
  // only possible if: a) not 1.0 format b) base64 content c) check config settings
#ifdef WITH_BASE64_HASHING_HACK
  if((pctx->bCollectDFData == 1) && 
	  (strcmp(pctx->pSigDoc->szFormat, SK_XML_1_NAME) && strcmp(pctx->pSigDoc->szFormatVer, SK_XML_1_VER)) && 
	 !strcmp(pctx->ctx3, CONTENT_EMBEDDED_BASE64) ) {
	// append end tag
	ddocMemAppendData(&(pctx->mbufElemData), "</DataFile>", -1);
	pTmp1 = canonicalizeXML((char*)pctx->mbufElemData.pMem, pctx->mbufElemData.nLen);
	if(pTmp1) {
	  // remove end tag again after canonicalization
	  pTmp1[strlen(pTmp1) - 11] = 0;
	  ddocDebug(4, "startElementHandler", "Initial sha1 update: \'%s\'", pTmp1);
	  mbuf1.pMem = pTmp1;
	  mbuf1.nLen = strlen(pTmp1);
	  ddocDebugWriteFile(4, "df-data.txt", &mbuf1);
	  SHA1_Update(&(pctx->sctx), pTmp1, strlen(pTmp1));
	  free(pTmp1);
	}
	ddocMemBuf_free(&(pctx->mbufElemData));
  }
#endif
  // collect general <SignedProperties> data
  bdocSaxParseCollectStartTag(pctx, name, pCurrElem->prefix, atts, &(pctx->bCollectSigPartData), &(pctx->mbufSigPartData));
  // check other start-tag-actions
  if(!pctx->nIgnoreDataFile) {
    if(!strcmp((const char*)pCurrElem->tag, "SignedDoc")) 
      handleStartSignedBDoc(pctx, pCurrElem->tag, atts);
    if(!strcmp((const char*)pCurrElem->tag, "Signature")) 
      handleStartSignatureBDoc(pctx, pCurrElem->tag, atts);
    if(!strcmp((const char*)pCurrElem->tag, "Reference")) 
      handleStartReferenceBDoc(pctx, pCurrElem->tag, atts);
    if(!strcmp((const char*)pCurrElem->tag, "SignatureMethod")) 
      handleStartSignatureMethodBDoc(pctx, pCurrElem->tag, atts);
    if(!strcmp((const char*)pCurrElem->tag, "DigestMethod")) 
      handleStartDigestMethodBDoc(pctx, pCurrElem->tag, atts);
    if(!strcmp((const char*)pCurrElem->tag, "Cert")) 
      handleStartCertBDoc(pctx, pCurrElem->tag, atts);
    /*if(!strcmp((const char*)name, "ResponderID"))   // TODO: do we need it ???
      handleStartResponderId(pctx, name, atts);*/
    /*    if(!strcmp((const char*)name, "Certificate"))  // TODO: do we need it ???
	  handleStartCertificate(pctx, name, atts);*/
    if(!strcmp((const char*)name, "UnsignedSignatureProperties")) 
      handleStartUnsignedSignaturePropertiesBDoc(pctx, pCurrElem->tag, atts);
			#ifdef WITH_TS
					if(!strcmp((const char*)pCurrElem->tag, "SignatureTimeStamp")) 
						handleStartSignatureTimeStampBDoc(pctx, pCurrElem->tag, atts);
					if(!strcmp((const char*)pCurrElem->tag, "SigAndRefsTimeStamp")) 
						handleStartSigAndRefsTimeStampBDoc(pctx, pCurrElem->tag, atts);
					if(!strcmp((const char*)pCurrElem->tag, "Include")) 
						handleStartIncludeBDoc(pctx, pCurrElem->tag, atts);
			#endif
    if(!strcmp((const char*)pCurrElem->tag, "OCSPIdentifier")) 
      handleStartOCSPIdentifierBDoc(pctx, pCurrElem->tag, atts);
    if(!strcmp((const char*)pCurrElem->tag, "SigningCertificate")) 
      handleStartSigningCertificateBDoc(pctx, pCurrElem->tag, atts);
    if(!strcmp((const char*)pCurrElem->tag, "CompleteCertificateRefs")) 
      handleStartCompleteCertificateRefsBDoc(pctx, pCurrElem->tag, atts);
    if(!strcmp((const char*)pCurrElem->tag, "EncapsulatedX509Certificate")) 
      handleStartEncapsulatedX509CertificateBDoc(pctx, pCurrElem->tag, atts);
    // start collecting but release old if exists
    if(!strcmp((const char*)pCurrElem->tag, "X509SerialNumber") ||
       !strcmp((const char*)pCurrElem->tag, "X509IssuerName")) {
      bdocSaxParseEndCollecting(pctx, FLAG_XML_ELEM, 0);
      pctx->errcode = bdocSaxParseStartCollecting(pctx, FLAG_XML_ELEM, 1);
    }
    // start collecting data of these elements
    if(!strcmp((const char*)pCurrElem->tag, "CertifiedRole") ||
       !strcmp((const char*)pCurrElem->tag, "ClaimedRole") ||
       !strcmp((const char*)pCurrElem->tag, "DigestValue") ||
       !strcmp((const char*)pCurrElem->tag, "EncapsulatedOCSPValue") ||
       !strcmp((const char*)pCurrElem->tag, "EncapsulatedTimeStamp") ||
       !strcmp((const char*)pCurrElem->tag, "X509Certificate") ||
       !strcmp((const char*)pCurrElem->tag, "SigningTime") ||
       !strcmp((const char*)pCurrElem->tag, "ResponderID") ||
       !strcmp((const char*)pCurrElem->tag, "ByName") ||
       !strcmp((const char*)pCurrElem->tag, "ByKey") ||
       !strcmp((const char*)pCurrElem->tag, "City") ||
       !strcmp((const char*)pCurrElem->tag, "StateOrProvince") ||
       !strcmp((const char*)pCurrElem->tag, "PostalCode") ||
       !strcmp((const char*)pCurrElem->tag, "CountryName") ||
       !strcmp((const char*)pCurrElem->tag, "IssuerSerial")
       )
      pctx->errcode = bdocSaxParseStartCollecting(pctx, FLAG_XML_ELEM, 1);
  }
}

//--------------------------------------------------
// Collects elements end-tag
// pctx - pointer to XML parsing work-area
// name - xml element tag name
// pcFlag - pointer to flag governing collection of this element
// pMBuf - DigiDocMemBuf to collect it in
//--------------------------------------------------
int bdocSaxParseCollectEndElement(SigDocParse* pctx, const xmlChar *name, 
				  char* pcFlag, DigiDocMemBuf* pMBuf)
{
  if(*pcFlag) {
    if(!pctx->errcode)
      pctx->errcode = ddocMemAppendData(pMBuf, "</", -1);
    if(!pctx->errcode)
      pctx->errcode = ddocMemAppendData(pMBuf, (const char*)name, -1);
    if(!pctx->errcode)
      pctx->errcode = ddocMemAppendData(pMBuf, ">", -1);
  }
  return pctx->errcode;
}

/**
 * endElementHandler:
 * @ctxt:  An XML parser context
 * @name:  The element name
 *
 * called when the end of an element has been detected.
 */
static void endElementHandlerBDoc(void *ctx, const xmlChar *name/*, const xmlChar **atts*/)
{
  SignatureInfo* pSigInfo = NULL;
  SigDocParse* pctx = (SigDocParse*)ctx;
  
  //AM 04.03.09 namespace support
  ElementEntry* pCurrElem = 0;
  pCurrElem = ddocStackFindEnd(&(pctx->stack));
  // do nothing if error has ocurred
  if(hasUnreadErrors()) return;
  ddocDebug(4, "endElementHandler", "</%s>", (const char*)pCurrElem->tag);
  // if we are in collect data mode then
  // collect the tag end
  if(!pctx->bNoXMLElemData)
    bdocSaxParseCollectEndElement(pctx, name, &(pctx->bCollectElemData), &(pctx->mbufElemData));

  // collect separately <Signature> and <SignedProperties> data
  bdocSaxParseCollectEndElement(pctx, name, &(pctx->bCollectSigData), &(pctx->mbufSigData));
  bdocSaxParseCollectEndElement(pctx, name, &(pctx->bCollectSigPartData), &(pctx->mbufSigPartData));
#ifdef WITH_TS
  bdocSaxParseCollectEndElement(pctx, name, &(pctx->bCollectTsInput), &(pctx->mbufTsInpElem));
#endif
  if(pctx->mbufSigData.pMem && !strcmp((const char*)pCurrElem->tag, "Signature")) {
    pSigInfo = ddocGetLastSignature(pctx->pSigDoc);
    if(pSigInfo) {
      // MEMLEAK: possible memleak if old content would not be released ???
      ddocDebug(3, "endElementHandler", "Set orig-content %s old-mem-used: %s new-content: %d", 
		pSigInfo->szId, (pSigInfo->mbufOrigContent.pMem ? "TRUE" : "FALSE"), pctx->mbufSigData.nLen);
      ddocMemBuf_free(&(pSigInfo->mbufOrigContent));
      pSigInfo->mbufOrigContent.pMem = (byte*)pctx->mbufSigData.pMem;
      pSigInfo->mbufOrigContent.nLen = pctx->mbufSigData.nLen;
      pctx->mbufSigData.pMem = 0;
      pctx->mbufSigData.nLen = 0;
    }
  }
  if(!pctx->nIgnoreDataFile) {
    if(!strcmp((const char*)pCurrElem->tag, "EncapsulatedX509Certificate")) 
      handleEndEncapsulatedX509CertificateBDoc(pctx, name);
    if(!strcmp((const char*)pCurrElem->tag, "EncapsulatedOCSPValue")) 
      handleEndEncapsulatedOCSPValueBDOC(pctx, name);
#ifdef WITH_TS
    if(!strcmp((const char*)pCurrElem->tag, "EncapsulatedTimeStamp")) 
      handleEndEncapsulatedTimeStampBDoc(pctx, name);
    if(!strcmp((const char*)pCurrElem->tag, "SignatureTimeStamp")) 
      handleEndSignatureTimeStampBDoc(pctx, name);
    if(!strcmp((const char*)pCurrElem->tag, "CompleteCertificateRefs")) 
      handleEndCompleteCertificateRefsBDoc(pctx, name);
    if(!strcmp((const char*)pCurrElem->tag, "CompleteRevocationRefs")) 
      handleEndCompleteRevocationRefsBDoc(pctx, name);
#endif
    if(!strcmp((const char*)pCurrElem->tag, "ClaimedRole")) 
      handleEndClaimedRoleBDoc(pctx, pCurrElem->tag);
    if(!strcmp((const char*)pCurrElem->tag, "CertifiedRole")) 
      handleEndCertifiedRoleBDoc(pctx, pCurrElem->tag);
    if(!strcmp((const char*)pCurrElem->tag, "X509Certificate")) 
      handleEndX509CertificateBDoc(pctx, pCurrElem->tag);
    if(!strcmp((const char*)pCurrElem->tag, "SignedProperties")) 
      handleEndSignedPropertiesBDoc(pctx, name, pCurrElem->prefix, pCurrElem->nsUri);
    if(!strcmp((const char*)pCurrElem->tag, "SignedInfo")) 
      handleEndSignedInfoBDoc(pctx, name, pCurrElem->prefix, pCurrElem->nsUri);
    if(!strcmp((const char*)pCurrElem->tag, "SignatureValue")) 
      handleEndSignatureValueBDoc(pctx, pCurrElem->tag);
    if(!strcmp((const char*)pCurrElem->tag, "ResponderID")) 
      handleEndResponderIDBDoc(pctx, pCurrElem->tag);
    if(!strcmp((const char*)pCurrElem->tag, "ByName")) 
      handleEndResponderIDBDoc(pctx, pCurrElem->tag);
    if(!strcmp((const char*)pCurrElem->tag, "ByKey")) 
      handleEndByKeyBDoc(pctx, pCurrElem->tag);
    if(!strcmp((const char*)pCurrElem->tag, "DigestValue")) 
      handleEndDigestValueBDoc(pctx, pCurrElem->tag);
    if(!strcmp((const char*)pCurrElem->tag, "SigningTime")) 
      handleEndSigningTimeBDoc(pctx, pCurrElem->tag);
    if(!strcmp((const char*)pCurrElem->tag, "IssuerSerial")) 
      handleEndIssuerSerialBDoc(pctx, pCurrElem->tag);
    if(!strcmp((const char*)pCurrElem->tag, "X509SerialNumber")) 
      handleEndIssuerSerialBDoc(pctx, pCurrElem->tag);
    if(!strcmp((const char*)pCurrElem->tag, "X509IssuerName")) 
      handleEndIssuerNameBDoc(pctx, pCurrElem->tag);
    //if(!strcmp((const char*)name, "Cert")) 
    //  handleEndCert(pctx, name);
    if(!strcmp((const char*)pCurrElem->tag, "City"))
       handleAdrEntryBDoc(pctx, ADR_ENTRY_CITY);
    if(!strcmp((const char*)pCurrElem->tag, "StateOrProvince"))
       handleAdrEntryBDoc(pctx, ADR_ENTRY_STATE);
    if(!strcmp((const char*)pCurrElem->tag, "PostalCode"))
       handleAdrEntryBDoc(pctx, ADR_ENTRY_ZIP);
    if(!strcmp((const char*)pCurrElem->tag, "CountryName"))
       handleAdrEntryBDoc(pctx, ADR_ENTRY_COUNTRY);

  }
  // reset tag, but not the context because used
  pctx->tag[0] = 0;
    // pop stack
  pctx->errcode = ddocStackPopElement(&(pctx->stack), 0, NULL);
}

/**
 * charactersHandler:
 * @ctxt:  An XML parser context
 * @ch:  a xmlChar string
 * @len: the number of xmlChar
 *
 * receiving some chars from the parser.
 * Question: how much at a time ???
 */
static void charactersHandlerBDoc(void *ctx, const xmlChar *ch, int len)
{
  SigDocParse* pctx = (SigDocParse*)ctx;
  char *p = 0;
	
  // do nothing if error has ocurred
  if(hasUnreadErrors()) return;
  // if we are in collect data mode then
  // collect this data
  ddocDebug(4, "charactersHandler", "tag: %s len: %d, elem-data: %s, sig-data: %s, collected: %d", 
	    pctx->tag, len, (pctx->mbufElemData.pMem ? "Y" : "N"), 
	    (pctx->mbufSigData.pMem ? "Y" : "N"), pctx->mbufElemData.nLen);
//#ifndef WITH_BASE64_HASHING_HACK
  if(pctx->bCollectElemData) {
    if(!pctx->errcode) {
      if(pctx->bNoXMLElemData) {
	pctx->errcode = ddocMemAppendData(&(pctx->mbufElemData), (const char*)ch, len);
      } else {
	pctx->errcode = escapeXMLSymbols((const char*)ch, len, &p);
	pctx->errcode = ddocMemAppendData(&(pctx->mbufElemData), (const char*)p, -1);
	free(p);
      }
    }
  }
//#endif
  if(pctx->bCollectSigData) {
    if(!pctx->errcode) {
      p = 0;
      pctx->errcode = escapeXMLSymbols((const char*)ch, len, &p);
      pctx->errcode = ddocMemAppendData(&(pctx->mbufSigData), p, -1);
      free(p);
    }
  }
  if(pctx->bCollectSigPartData) {
    if(!pctx->errcode) {
      p = 0;
      pctx->errcode = escapeXMLSymbols((const char*)ch, len, &p);
      pctx->errcode = ddocMemAppendData(&(pctx->mbufSigPartData), p, -1);
      free(p);
    }
  }

#ifdef WITH_TS
  if(pctx->bCollectTsInput) {
    if(!pctx->errcode) {
      p = 0;
      pctx->errcode = escapeXMLSymbols((const char*)ch, len, &p);
      pctx->errcode = ddocMemAppendData(&(pctx->mbufTsInpElem), p, -1);
      free(p);
    }
  }
#endif
  ddocDebug(5, "charactersHandler", "End collecting");
  ddocDebug(5, "charactersHandler", "End");
}
/**
 * startElementHandler:
 * @ctxt:  An XML parser context
 * @name:  The element name
 *
 * called when an opening tag has been processed.
 */
static void extractStartElementHandlerBDoc(void *ctx, const xmlChar *name, const xmlChar **atts)
{
  const char* id = 0, *ctype = 0;
  int i, l1;
  char *p1 = 0;

  SigDocParse* pctx = (SigDocParse*)ctx;	
  strncpy(pctx->tag, (const char*)name, sizeof(pctx->tag));
  ddocDebug(5, "extractStartElementHandler", "tag: %s", pctx->tag);
  // do nothing if error has ocurred
  if(hasUnreadErrors()) return;
  if(!strcmp((const char*)name, "DataFile")) {
    for (i = 0; (atts != NULL) && (atts[i] != NULL); i += 2) {
      if(!strcmp((const char*)atts[i], "Id"))
	id = (const char*)atts[i+1];
      if(!strcmp((const char*)atts[i], "ContentType"))
	ctype = (const char*)atts[i+1];
    } 
    if(!pctx->nIgnoreDataFile) {
      strncpy(pctx->ctx2, id, sizeof(pctx->ctx2));	
      strncpy(pctx->ctx4, ctype, sizeof(pctx->ctx4));
      if(!strcmp(pctx->ctx2, pctx->ctx3)) {
	pctx->bCollectDFData++; // increment bypass mode
	ddocDebug(4, "extractStartElementHandler", "Start DF: %s mode: %s skip: %d", 
		  pctx->ctx3, pctx->ctx1, pctx->bCollectDFData);
	if(pctx->bCollectDFData == 1) { // only the first time
	  ddocDebug(4, "extractStartElementHandler", "Init collecting DF: %s mode: %s", 
		    pctx->ctx3, pctx->ctx1);
	  if(!strcmp(pctx->ctx4, CONTENT_EMBEDDED_BASE64) && !pctx->bKeepBase64) {
	    EVP_DecodeInit(&(pctx->ectx));
	    pctx->b64pos = 0;
	    pctx->lSize = 0;
	  }
	  // open output file if necessary
	  if(!pctx->bDataFile && !pctx->pMemBufDF) {
	    pctx->bDataFile = BIO_new_file(pctx->ctx5, "w");
	    ddocDebug(4, "extractStartElementHandler", "Opening file: %s", pctx->ctx5);
	    if(!pctx->bDataFile)
	      SET_LAST_ERROR(ERR_FILE_WRITE);
	  }
	}
      }
    }		
    if(!strcmp(ctype, CONTENT_EMBEDDED) || pctx->nIgnoreDataFile)
      pctx->nIgnoreDataFile++;
  }
  if(!strcmp(pctx->ctx2, pctx->ctx3) && 
     !strcmp(pctx->ctx4, CONTENT_EMBEDDED) &&
     strcmp((const char*)name, "DataFile")) {
    if(!strcmp(pctx->ctx1, CHARSET_ISO_8859_1)) { // if must convert
      // begining of the tag
      l1 = strlen((char*)name) + 10;
      p1 = (char*)malloc(l1);
      RETURN_VOID_IF_BAD_ALLOC(p1);
      if(pctx->pMemBufDF) {
	ddocMemAppendData(pctx->pMemBufDF, "<", -1);
	ddocMemAppendData(pctx->pMemBufDF, utf82ascii((const char*)name, p1, &l1), -1);
      }
      else
	BIO_printf(pctx->bDataFile, "<%s", utf82ascii((const char*)name, p1, &l1));
      free(p1);
      p1 = 0;
      for (i = 0; (atts != NULL) && (atts[i] != NULL); i += 2) {
	l1 = strlen((char*)atts[i]) + 10;
	p1 = (char*)malloc(l1);
	RETURN_VOID_IF_BAD_ALLOC(p1);
	if(pctx->pMemBufDF) {
	  ddocMemAppendData(pctx->pMemBufDF, " ", -1);
	  ddocMemAppendData(pctx->pMemBufDF, utf82ascii((const char*)atts[i], p1, &l1), -1);
	  ddocMemAppendData(pctx->pMemBufDF, "=", -1);
	}
	else
	  BIO_printf(pctx->bDataFile, " %s=", utf82ascii((const char*)atts[i], p1, &l1));
	free(p1);
	p1 = 0;
	l1 = strlen((char*)atts[i+1]) + 10;
	p1 = (char*)malloc(l1);
	RETURN_VOID_IF_BAD_ALLOC(p1);
	if(pctx->pMemBufDF) {
	  ddocMemAppendData(pctx->pMemBufDF, "\"", -1);
	  ddocMemAppendData(pctx->pMemBufDF, utf82ascii((const char*)atts[i+1], p1, &l1), -1);
	  ddocMemAppendData(pctx->pMemBufDF, "\"", -1);
	}
	else
	BIO_printf(pctx->bDataFile, "\"%s\"", utf82ascii((const char*)atts[i+1], p1, &l1));
	free(p1);
	p1 = 0;
      } // for - atributes
      if(pctx->pMemBufDF)
	ddocMemAppendData(pctx->pMemBufDF, ">", -1);
      else
	BIO_puts(pctx->bDataFile, ">");
    } // if - must convert
    else { // no need to convert
      if(pctx->pMemBufDF) {
	ddocMemAppendData(pctx->pMemBufDF, "<", -1);
	ddocMemAppendData(pctx->pMemBufDF, (const char*)name, -1);
	for (i = 0; (atts != NULL) && (atts[i] != NULL); i += 2) {
	  ddocMemAppendData(pctx->pMemBufDF, " ", -1);
	  ddocMemAppendData(pctx->pMemBufDF, (const char*)atts[i], -1);
	  p1 = 0;
	  ddocMemAppendData(pctx->pMemBufDF, "=\"", -1);
	  pctx->errcode = escapeXMLSymbols((const char*)atts[i+1], 
		  strlen((const char*)atts[i+1]), &p1);
	  ddocMemAppendData(pctx->pMemBufDF, p1, -1);
	  free(p1);
	  ddocMemAppendData(pctx->pMemBufDF, "\"", -1);
	} 
	ddocMemAppendData(pctx->pMemBufDF, ">", -1);
      } else {
	BIO_printf(pctx->bDataFile, "<%s", name);
	for (i = 0; (atts != NULL) && (atts[i] != NULL); i += 2) {
	  BIO_printf(pctx->bDataFile, " %s=\"%s\"", 
		     (const char*)atts[i], (const char*)atts[i+1]);
	} 
	BIO_puts(pctx->bDataFile, ">");
      }      
    } // else no conversion
  } // if strcmp(DataFile)
}

/**
 * extractBodyHandler:
 * @ctxt:  An XML parser context
 * @name:  The element name
 *
 * called when bypassing xml parser for base64 data and extracting to file
 */
void extractBodyHandlerBDoc(SigDocParse* pctx, const char* ch, int len)
{
  ddocDebug(4, "extractBodyHandler", "DF: %s data: %d", pctx->ctx2, len);
  /*if(!pctx->pMemBufDF)
    SET_LAST_ERROR_RETURN_VOID_IF_NOT(pctx->bDataFile, ERR_FILE_WRITE);*
  if(!pctx->bKeepBase64) {
    extractDecodeB64(pctx, ch, len, 0);
  } else {
	  if(pctx->pMemBufDF) {
         ddocMemAppendData(pctx->pMemBufDF, ch, len);
	  }else {
		 ddocDebug(4, "extractBodyHandler", "Writing: %s len: %d", ch, len);
         BIO_write(pctx->bDataFile, ch, len);
	  }
    pctx->lSize += len;
  }*/
}

/**
 * extractBodyHandler:
 * @ctxt:  An XML parser context
 * @name:  The element name
 *
 * called when bypassing xml parser for bas64 data and extracting to file
 */
void extractNoChangeHandlerBDoc(SigDocParse* pctx, const char* ch, int len)
{

  char *p = 0;

  ddocDebug(4, "extractNoChangeHandler", "DF: %s data: %d", pctx->ctx2, len);
  if(hasUnreadErrors()) { return; }
  /*if(!pctx->pMemBufDF)
    SET_LAST_ERROR_RETURN_VOID_IF_NOT(pctx->bDataFile, ERR_FILE_WRITE);
  if(!strcmp(pctx->ctx4, CONTENT_EMBEDDED_BASE64)) {
    BIO_write(pctx->bDataFile, ch, len);
    pctx->lSize += len;
  } else {
    pctx->errcode = escapeXMLSymbols((const char*)ch, len, &p);
    l = strlen(p);
    pctx->lSize += l;
    if(pctx->pMemBufDF) {
      ddocMemAppendData(pctx->pMemBufDF,p, l);
	} else {
	  ddocDebug(5, "extractNoChangeHandler", "Writing: %s len: %d", p, l);
      BIO_write(pctx->bDataFile, p, l);
	}
    free(p);
    p = 0;
  }*/
}

/**
 * charactersHandler:
 * @ctxt:  An XML parser context
 * @ch:  a xmlChar string
 * @len: the number of xmlChar
 *
 * receiving some chars from the parser.
 * Question: how much at a time ???
 */
static void extractCharactersHandlerBDoc(void *ctx, const xmlChar *ch, int len)
{
  static int pos = 0;
  char *p = 0, *p2 = 0; 
  SigDocParse* pctx = (SigDocParse*)ctx;

  ddocDebug(5, "extractCharactersHandler", "tag: %s, data: %d - \'%s\'", pctx->tag, len, (char*)ch);
  // do nothing if error has ocurred
  if(hasUnreadErrors()) { return; }
  ddocDebug(5, "extractCharactersHandler", "done, errs: %d", hasUnreadErrors());
}

/**
 * extractEndElementHandler:
 * @ctxt:  An XML parser context
 * @name:  The element name
 *
 * called when the end of an element has been detected.
 */
static void extractEndElementHandlerBDoc(void *ctx, const xmlChar *name)
{
  SigDocParse* pctx = (SigDocParse*)ctx;

  ddocDebug(5, "extractEndElementHandler", "tag: %s", (char*)name);
  // do nothing if error has ocurred
  if(hasUnreadErrors()) { return; }
  // reset tag
  pctx->tag[0] = 0;
}


/**
 * cdataBlockHandler:
 * @ctx: the user data (XML parser context)
 * @value:  The pcdata content
 * @len:  the block length
 *
 * called when a pcdata block has been parsed
 */
static void cdataBlockHandlerBDoc(void *ctx, const xmlChar *value, int len)
{
    fprintf(stdout, "SAX.pcdata(%.20s, %d)\n", (char *) value, len);
}


/**
 * warningHandler:
 * @ctxt:  An XML parser context
 * @msg:  the message to display/transmit
 * @...:  extra parameters for the message display
 *
 * Display and format a warning messages, gives file, line, position and
 * extra parameters.
 */
static void warningHandlerBDoc(void *ctx, const char *msg, ...)
{
  va_list args;

  va_start(args, msg);
  ddocDebugVaArgs(2, "warningHandler", msg, args);
  va_end(args);
}

/**
 * errorHandler:
 * @ctxt:  An XML parser context
 * @msg:  the message to display/transmit
 * @...:  extra parameters for the message display
 *
 * Display and format a error messages, gives file, line, position and
 * extra parameters.
 */
static void errorHandlerBDoc(void *ctx, const char *msg, ...)
{
  va_list args;
  SigDocParse* pctx = (SigDocParse*)ctx;
	
  va_start(args, msg);	
  //pctx->errcode = ERR_DIGIDOC_PARSE;
  ddocDebugVaArgs(1, "errorHandler", msg, args);
  //addError(pctx->errcode, __FILE__, __LINE__, "XML parsing error");
  va_end(args);
}

/**
 * fatalErrorHandler:
 * @ctxt:  An XML parser context
 * @msg:  the message to display/transmit
 * @...:  extra parameters for the message display
 *
 * Display and format a fatalError messages, gives file, line, position and
 * extra parameters.
 */
static void fatalErrorHandlerBDoc(void *ctx, const char *msg, ...)
{
  va_list args;
  SigDocParse* pctx = (SigDocParse*)ctx;

  va_start(args, msg);
  pctx->errcode = ERR_DIGIDOC_PARSE;
  ddocDebugVaArgs(1, "fatalErrorHandler", msg, args);
  addError(pctx->errcode, __FILE__, __LINE__, "XML parsing error");
  va_end(args);
}


xmlSAXHandler debugSAXHandlerStruct1 = {
    NULL, //internalSubsetHandler,
    NULL, //isStandaloneHandler,
    NULL, //hasInternalSubsetHandler,
    NULL, //hasExternalSubsetHandler,
    NULL, //resolveEntityHandler,
    NULL, //getEntityHandler,
    NULL, //entityDeclHandler,
    NULL, //notationDeclHandler,
    NULL, //attributeDeclHandler,
    NULL, //elementDeclHandler,
    NULL, //unparsedEntityDeclHandler,
    NULL, //setDocumentLocatorHandler,
    NULL, //startDocumentHandler,
    NULL, //endDocumentHandler,
    startElementHandlerBDoc,
    endElementHandlerBDoc,
    NULL, //referenceHandler,
    charactersHandlerBDoc,
    NULL, //ignorableWhitespaceHandler,
    NULL, //processingInstructionHandler,
    NULL, //commentHandler,
    warningHandlerBDoc,
    errorHandlerBDoc,
    fatalErrorHandlerBDoc,
    NULL, //getParameterEntityHandler,
    cdataBlockHandlerBDoc,
    NULL, //externalSubsetHandler,
    1
};


xmlSAXHandlerPtr debugSAXHandlerBDoc = &debugSAXHandlerStruct1;


xmlSAXHandler extractSAXHandlerStruct1 = {
    NULL, //internalSubsetHandler,
    NULL, //isStandaloneHandler,
    NULL, //hasInternalSubsetHandler,
    NULL, //hasExternalSubsetHandler,
    NULL, //resolveEntityHandler,
    NULL, //getEntityHandler,
    NULL, //entityDeclHandler,
    NULL, //notationDeclHandler,
    NULL, //attributeDeclHandler,
    NULL, //elementDeclHandler,
    NULL, //unparsedEntityDeclHandler,
    NULL, //setDocumentLocatorHandler,
    NULL, //startDocumentHandler,
    NULL, //endDocumentHandler,
    extractStartElementHandlerBDoc,
    extractEndElementHandlerBDoc,
    NULL, //referenceHandler,
    extractCharactersHandlerBDoc,
    NULL, //ignorableWhitespaceHandler,
    NULL, //processingInstructionHandler,
    NULL, //commentHandler,
    warningHandlerBDoc,
    errorHandlerBDoc,
    fatalErrorHandlerBDoc,
    NULL, //getParameterEntityHandler,
    NULL, //cdataBlockHandler,
    NULL, //externalSubsetHandler,
    1
};

xmlSAXHandlerPtr extractSAXHandlerBDoc = &extractSAXHandlerStruct1;


//--------------------------------------------------
// handles the start of a <DataFile> element.
// Reads in id, filename, mime_type and embedded attributes
// pctx - pointer to XML parsing work-area
// name - tag name
// atts - attributes
//--------------------------------------------------
EXP_OPTION int addDataFileBDoc(SignedDoc* pSigDoc, char* filename, char* mime, char* ctype, int size, 
							   const char* dtype, byte* dvalue, int dlen,
							   byte* dvalue2, int dlen2)
{
  int err = ERR_OK;
  DataFile* pDataFile = NULL;
  char buf1[300];

  memset(buf1, 0, sizeof(buf1));
  snprintf(buf1,sizeof(buf1), "/%s", filename);
  err = DataFile_new(&pDataFile, pSigDoc, buf1, filename, ctype, 
			 mime, size, NULL, 0, dtype, CHARSET_UTF_8);

  RETURN_IF_NULL(pDataFile);
  ddocDebug(4, "addDataFileBDoc", "Create DF: \'%s\' - err: %d", filename, err);
  if(dvalue) 
    ddocDataFile_SetDigestValue(pDataFile, (const char*)dvalue, dlen);
  if(dvalue2) 
    ddocDataFile_SetWrongDigestValue(pDataFile, (const char*)dvalue2, dlen2);
  

  return 0;
}

//--------------------------------------------------
// Checks if memory buffer contains bdoc-1.0 mime type
// trims spaces from front and end
// pBuf - memory buffer
// return 0 = OK, or error code.
//--------------------------------------------------
int checkBdocMime(DigiDocMemBuf* pBuf)
{
	char *p1;

	if(pBuf && pBuf->pMem) {
		// trim spaces at end
		p1 = strchr(pBuf->pMem, 0); // find end
		while(*(p1-1) && isspace(*(p1-1))) {
			p1--;
			*p1 = 0;
		}
		// trimspaced at the beginning
		p1 = pBuf->pMem;
		while(*p1 && isspace(*p1)) p1++;
		return (strcmp(p1, MIME_BDOC_1) ? ERR_UNSUPPORTED_FORMAT : 0);
	}
	return ERR_UNSUPPORTED_FORMAT;
}

//--------------------------------------------------
// Reads in signed XML document info from digidoc file
// ppSigDoc - pointer to the buffer of newly read info pointer
// szFileName - documents filename
// checkFileDigest - indicates if digests of datafiles referred by the document must be checked
// lMaxDFLen - maximum size for a DataFile whose contents will be
// kept in memory
//--------------------------------------------------
EXP_OPTION int bdocSaxReadSignedDocFromFile(SignedDoc** ppSigDoc, const char* szFileName, 
			     int checkFileDigest, long lMaxDFLen)
{
	int err = ERR_OK, ret, size=0, l1, l2;
	SigDocParse pctx;
	DigiDocMemBuf mbuf1, mbuf2;
	xmlDocPtr doc;
	xmlNodePtr cur;
	byte digest1[DIGEST_LEN256 + 2], digest2[DIGEST_LEN + 2];
	char *mimeType;
	char *fullPath;
	char buf2[200], buf1[100];
	char convFileName[256];
	char * pch;

	mbuf1.pMem = 0;
	mbuf1.nLen = 0;
	mbuf2.pMem = 0;
	mbuf2.nLen = 0;
	ddocDebug(3, "bdocSaxReadSignedDocFromFile", "digidoc: %s, checkDig: %d, maxDF: %ld", 
	szFileName, checkFileDigest, lMaxDFLen);
    RETURN_IF_NULL_PARAM(ppSigDoc);
    RETURN_IF_NULL_PARAM(szFileName);
    clearErrors();
	
	// check archieve mime type
	ddocConvertFileName(convFileName, sizeof(convFileName), szFileName);
    err = do_extract_filemem(convFileName, "mimetype", &mbuf1);
    if(err != ERR_OK) return err;
	ddocMemAppendData(&mbuf2, "mimetype", -1);
	err = checkBdocMime(&mbuf1);
	ddocMemBuf_free(&mbuf1);
	if(err != ERR_OK) return err;
	
	// check file list in manifest.xml
	err = do_extract_filemem(convFileName,"META-INF/manifest.xml", &mbuf1);
	if (err != ERR_OK) {
		err = do_extract_filemem(convFileName,"META-INF\\manifest.xml", &mbuf1);
		if (err != ERR_OK) return err;
		ddocMemAppendData(&mbuf2, "|META-INF\\manifest.xml", -1);
	} else
		ddocMemAppendData(&mbuf2, "|META-INF/manifest.xml", -1);
	// initialize parser
    memset(&pctx, 0, sizeof(pctx));
    pctx.pSigDoc = (SignedDoc*)malloc(sizeof(SignedDoc));
    RETURN_IF_BAD_ALLOC(pctx.pSigDoc);
    memset(pctx.pSigDoc, 0, sizeof(SignedDoc));
	pctx.szInputFileName = 0; 
	pctx.checkFileDigest = 0;
	pctx.lMaxDFLen = lMaxDFLen;
	setString(&pctx.pSigDoc->szFileName, convFileName, -1);
	setString(&(pctx.pSigDoc->szFormat), BDOC_XML_1_NAME, -1);
	setString(&(pctx.pSigDoc->szFormatVer), BDOC_XML_1_VER, -1);

	//reading manifest
	doc = xmlParseMemory((const char*)mbuf1.pMem, (int)mbuf1.nLen);
	if (doc != NULL ) 
		cur = xmlDocGetRootElement(doc);
	else 
		return ERR_MANIFEST;
	if (cur == NULL) {
		xmlFreeDoc(doc);
		return ERR_MANIFEST;
	} else {
		if (xmlStrcmp(cur->name, (const xmlChar *) "manifest")) {
			xmlFreeDoc(doc);
		}
		cur = cur->xmlChildrenNode;
		while (cur != NULL && !err) {
				if ((!xmlStrcmp(cur->name, (const xmlChar *)"file-entry"))) {
					mimeType = xmlGetProp(cur, "media-type");
					fullPath = xmlGetProp(cur, "full-path");
					strncpy(buf2,mimeType,sizeof(buf2));
					pch = strtok (buf2,"/");
					// collect file names
					if(strcmp(fullPath, "/")) {
						ddocMemAppendData(&mbuf2, "|", -1); // separator char
						ddocMemAppendData(&mbuf2, fullPath, -1);
					}
					if(strcmp(fullPath, "/") && strcmp(pch, "signature")){
						//ddocConvertFileName(convDfName, sizeof(convDfName), fullPath);
						size = calculateFileSizeFromZip(convFileName, fullPath);
						calculateFileDigestFromZip(convFileName, fullPath, digest1, BDOC_DIGEST, DIGEST_LEN256);
						calculateFileDigestFromZip(convFileName, fullPath, digest2, DIGEST_SHA1, DIGEST_LEN); // VS: support multi sha1, sha256 bdocs
						l1 = sizeof(buf1);
						encode((const byte*)digest1, DIGEST_LEN256, (byte*)buf1, &l1);
						l2 = sizeof(buf2);
						encode((const byte*)digest2, DIGEST_LEN, (byte*)buf2, &l2);
						ddocDebug(3, "bdocSaxReadSignedDocFromFile", "Ref: %s l1: %d digest1: %s l2: %d digest2: %d ", fullPath, l1, buf1, l2, buf2);
						addDataFileBDoc(pctx.pSigDoc, fullPath,mimeType, CONTENT_BINARY, size, 
							(!BDOC_DIGEST ? DIGEST_SHA1_NAME : DIGEST_SHA256_NAME), digest1, DIGEST_LEN256, digest2, DIGEST_LEN);
					} else if(strcmp(fullPath, "/") && !strcmp(pch, "signature")) {
						if(!xmlStrcmp(mimeType, MIME_BDOCSIG_TM) || !xmlStrcmp(mimeType, MIME_BDOCSIG_TS)) {
							pch = strtok (NULL,"/");
							pch = strtok (NULL,"/");
						    setString(&pctx.pSigDoc->szProfile,  pch, -1);
							err = do_extract_filemem(convFileName,fullPath, &pctx.mbufCanonSigData);
							if(!err) {
								// do a primitive counting because otherwiuse errors happen in inappropriate places
								l1 = 0;
								if(strstr((const char*)pctx.mbufCanonSigData.pMem, "SignatureValueTimeStamp>")) l1++;
								if(strstr((const char*)pctx.mbufCanonSigData.pMem, "SigAndRefsTimeStamp>")) l1++;

								// check profile
								if(!xmlStrcmp(mimeType, MIME_BDOCSIG_TM) && l1 > 0) {
									ddocDebug(1, "bdocSaxReadSignedDocFromMemory", "Signature: %s of type: %s contains: %d timestamps", fullPath, mimeType, l1);
									err = ERR_SIG_INVALID_PROFILE;
									SET_LAST_ERROR(err);
								}
								if(!xmlStrcmp(mimeType, MIME_BDOCSIG_TS) && l1 == 0) {
									ddocDebug(1, "bdocSaxReadSignedDocFromMemory", "Signature: %s of type: %s contains: %d timestamps", fullPath, mimeType, l1);
									err = ERR_SIG_INVALID_PROFILE;
									SET_LAST_ERROR(err);
								}
								ret = xmlSAXUserParseMemory(debugSAXHandlerBDoc, &pctx, (const char*)pctx.mbufCanonSigData.pMem, (int)pctx.mbufCanonSigData.nLen);
							}
							ddocMemBuf_free(&pctx.mbufCanonSigData);
							
						} else {
							err = ERR_UNSUPPORTED_SIGNATURE;
							SET_LAST_ERROR(err);
							ddocDebug(1, "bdocSaxReadSignedDocFromMemory", "Unsupported signature type: %s", (const char*)mimeType);
						}
					}
					xmlFree(mimeType);
					xmlFree(fullPath);
				}
				 
				cur = cur->next;
			}
	}
  ddocDebug(3, "bdocSaxReadSignedDocFromMemory", "BDOC file list: %s", (char*)mbuf2.pMem);
  err = checkFilesInList(convFileName, (const char*)mbuf2.pMem);

  *ppSigDoc = pctx.pSigDoc;
  if (err != ERR_OK) {
    SET_LAST_ERROR(err);
    ddocDebug(1, "bdocSaxReadSignedDocFromMemory", "parsing error: %d", err);
  }
  if(!err)
    err = getLastError();
  // cleanup parser context
	ddocMemBuf_free(&mbuf1);
	ddocMemBuf_free(&mbuf2);
	xmlFreeDoc(doc);
  bdocSAXCleanup(&pctx);
  return err;
}


//--------------------------------------------------
// Reads in signed XML document info
// ppSigDoc - pointer to the buffer of newly read info pointer
// szFileName - documents filename
// checkFileDigest - indicates if digests of datafiles referred by the document must be checked
// lMaxDFLen - maximum size for a DataFile whose contents will be
// kept in memory
//--------------------------------------------------
EXP_OPTION int bdocSaxReadSignedDocFromMemory(SignedDoc** ppSigDoc, const void* pData, 
				       int len, long lMaxDFLen)
{
  int err = ERR_OK;
  int ret;
  SigDocParse pctx;

  ddocDebug(3, "bdocSaxReadSignedDocFromMemory", "data len: %d, maxDF: %ld", len, lMaxDFLen);
  RETURN_IF_NULL_PARAM(ppSigDoc);
  RETURN_IF_NULL_PARAM(pData);

  clearErrors();
  memset(&pctx, 0, sizeof(pctx));
  pctx.pSigDoc = (SignedDoc*)malloc(sizeof(SignedDoc));
  RETURN_IF_BAD_ALLOC(pctx.pSigDoc);
  memset(pctx.pSigDoc, 0, sizeof(SignedDoc));
  //pctx.bDataFile = (BIO*)1;
  pctx.szInputFileName = 0;
  pctx.checkFileDigest = 0;
  pctx.lMaxDFLen = lMaxDFLen;

  ret = xmlSAXUserParseMemory(debugSAXHandlerBDoc, &pctx, (const char*)pData, len);
  *ppSigDoc = pctx.pSigDoc;
  if (err != ERR_OK) {
    SET_LAST_ERROR(err);
    ddocDebug(1, "bdocSaxReadSignedDocFromMemory", "parsing error: %d", err);
  }
  if(!err)
    err = getLastError();
  // cleanup parser context
  bdocSAXCleanup(&pctx);
  return err;
}


//--------------------------------------------------
// Reads new signatures from another digidoc file
// and adds to existing digidoc. Adds only those
// signatures that don't exist in old digidoc.
// pSigDoc - signed document object
// szFileName - name of digidoc file
// returns error code or ERR_OK
//--------------------------------------------------
EXP_OPTION int bdocReadNewSignaturesFromDdoc(SignedDoc* pSigDoc, const char* szFileName)
{
  SignedDoc *pSigDoc2 = 0;
  int err = ERR_OK, n1 = 0, n2 = 0, i;
  SignatureInfo** pSignatures = NULL;

  RETURN_IF_NULL_PARAM(pSigDoc);
  RETURN_IF_NULL_PARAM(szFileName);
  // read in digidoc file
  err = bdocSaxReadSignedDocFromFile(&pSigDoc2, szFileName, 0, 0);
  if(err) {
    if(pSigDoc2) 
      SignedDoc_free(pSigDoc2);
    return err;
  }
  // copy new signatures from digidoc just read in
  // assumes that new signatures are at the end
  n1 = getCountOfSignatures(pSigDoc);
  n2 = getCountOfSignatures(pSigDoc2);
  if(n2 > n1) {
    pSignatures = (SignatureInfo**)realloc(pSigDoc->pSignatures, n2 * sizeof(void *));
    if(!pSignatures) {
      if(pSigDoc2) 
		SignedDoc_free(pSigDoc2);
      RETURN_IF_BAD_ALLOC(pSignatures);
    }
    pSigDoc->pSignatures = pSignatures;
    for(i = n1; i < n2; i++) {
      pSigDoc->pSignatures[i] = pSigDoc2->pSignatures[i]; // take ownership
      pSigDoc2->pSignatures[i] = 0; // set to NULL to prevent deallocation
	  // VS destroy hash - debug
	  ((char*)pSigDoc->pSignatures[i]->pDocs[0]->szDigest)[0] = 0x0A;
    }
    pSigDoc->nSignatures = n2;
    pSigDoc2->nSignatures = n1;
  }
  // delete new digidoc object no longer necessary
  SignedDoc_free(pSigDoc2);

  return err;
}

//--------------------------------------------------
// Reads in signed XML document and extracts the desired data file
// pSigDoc - signed document object if exists. Can be NULL
// szFileName - digidoc filename
// szDataFileName - name of the file where to store embedded data. 
// szDocId - DataFile Id atribute value
// szCharset - convert DataFile content to charset
//--------------------------------------------------
EXP_OPTION int bdocExtractDataFile(const char* szFileName, const char* szDataFileNameInZip, const char* szDataFileName)
{
  int err = ERR_OK;

  RETURN_IF_NULL_PARAM(szFileName);
  RETURN_IF_NULL_PARAM(szDataFileName);

	ddocDebug(3, "bdocExtractDataFile", "DataFileNameInZip: %s, DataFileName: %s", szFileName, szDataFileNameInZip, szDataFileName);
	err = do_extract_onefile(szFileName, szDataFileNameInZip, 1, szDataFileName);
  if (err != ERR_OK){
    ddocDebug(1, "bdocExtractDataFile", "Error reading file: %s", szFileName);
    SET_LAST_ERROR_RETURN_CODE(ERR_FILE_READ);
  }

  return err;
}

//--------------------------------------------------
// Reads in desired data file from BDOC container
// pSigDoc - signed document object if exists. Can be NULL
// szFileName - bdoc digidoc filename
// szDataFileNameInZip - name of the file to get from bdoc container
// pBuf - memory buffer 
//--------------------------------------------------
EXP_OPTION int bdocGetDataFile(const char* szFileName, const char* szDataFileNameInZip, DigiDocMemBuf* pBuf)
{
  int err = ERR_OK;

  RETURN_IF_NULL_PARAM(szFileName);
  RETURN_IF_NULL_PARAM(szDataFileNameInZip);
  RETURN_IF_NULL_PARAM(pBuf);
  ddocDebug(3, "bdocGetDataFile", "DataFileNameInZip: %s", szFileName, szDataFileNameInZip);
  err = do_extract_filemem(szFileName, szDataFileNameInZip, pBuf);
  if (err != ERR_OK){
    ddocDebug(1, "bdocGetDataFile", "Error reading file: %s", szFileName);
    SET_LAST_ERROR_RETURN_CODE(ERR_FILE_READ);
  }

  return err;
}
