//==================================================
// FILE:	BDocZip.c
// PROJECT:     Digi Doc
// DESCRIPTION: Digi Doc functions for zip operations
// AUTHOR:  Ants Madisson, Business IT Partner
//==================================================
// Copyright (C) AS Sertifitseerimiskeskus
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// GNU Lesser General Public Licence is available at
// http://www.gnu.org/copyleft/lesser.html
//==========< HISTORY >=============================
//      20.02.2008      Ants Madisson
//                      Creation
//==================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>

#ifdef WIN32
# include <direct.h>
# include <io.h>
#else
# include <unistd.h>
# include <utime.h>
# include <sys/types.h>
# include <sys/stat.h>
#endif

#include <minizip/zip.h>
#include <minizip/unzip.h>
#include <openssl/sha.h>
#include <libdigidoc/DigiDocDebug.h>
#include <libdigidoc/DigiDocError.h>

#define CASESENSITIVITY (0)

#ifdef WIN32
#define USEWIN32IOAPI
#include <minizip/iowin32.h>
#else
#define MAX_PATH (256)
#endif




#define WRITEBUFFERSIZE (16384)
#define MAXFILENAME (256)

//--------------------------------------------------
//char *f;                /* name of file to get info on */
//tm_zip *tmzip;             /* return value: access, modific. and creation times */
//uLong *dt;             /* dostime */
//--------------------------------------------------
#ifdef WIN32
uLong filetime(const char *f, tm_zip *tmzip, uLong *dt)
{
  int ret = 0;

  FILETIME ftLocal;
  HANDLE hFind;
  WIN32_FIND_DATA  ff32;
	FILETIME tmp;

  hFind = FindFirstFile(f,&ff32);
  if (hFind != INVALID_HANDLE_VALUE)
  {
    FileTimeToLocalFileTime(&(ff32.ftLastWriteTime),&ftLocal);
    FileTimeToDosDateTime(&ftLocal,((LPWORD)dt)+1,((LPWORD)dt)+0);
    FindClose(hFind);
    ret = 1;
  } else {
		GetSystemTimeAsFileTime(&tmp);
		FileTimeToLocalFileTime(&tmp,&ftLocal);
    FileTimeToDosDateTime(&ftLocal,((LPWORD)dt)+1,((LPWORD)dt)+0);
	}

  return ret;
}
#else
uLong filetime(const char *f, tm_zip *tmzip, uLong *dt)
{
  int ret=0;
  struct stat s;        /* results of stat() */
  struct tm filedate;
  time_t tm_t=0;
if(f){
  if (strcmp(f,"-")!=0)
  {
    char name[MAXFILENAME+1];
    int len = strlen(f);

    strncpy(name, f,MAXFILENAME-1);
    /* strncpy doesnt append the trailing NULL, of the string is too long. */
    name[ MAXFILENAME ] = '\0';

    if (name[len - 1] == '/')
      name[len - 1] = '\0';
    /* not all systems allow stat'ing a file with / appended */
    if (stat(name,&s)==0)
    {
      tm_t = s.st_mtime;
      ret = 1;
    }
  }
}else{
  tm_t = time(NULL);
}	
  //filedate = localtime(&tm_t);
  ddocLocalTime(&tm_t,&filedate,0);

  tmzip->tm_sec  = filedate.tm_sec;
  tmzip->tm_min  = filedate.tm_min;
  tmzip->tm_hour = filedate.tm_hour;
  tmzip->tm_mday = filedate.tm_mday;
  tmzip->tm_mon  = filedate.tm_mon ;
  tmzip->tm_year = filedate.tm_year;

  return ret;
}
#endif




int check_exist_file(filename)
    const char* filename;
{
    FILE* ftestexist;
    int ret = 1;
    ftestexist = fopen(filename,"rb");
    if (ftestexist==NULL)
        ret = 0;
    else
        fclose(ftestexist);
    return ret;
}

/* calculate the CRC32 of a file,
   because to encrypt a file, we need known the CRC32 of the file before */
int getFileCrc(const char* filenameinzip,void*buf,unsigned long size_buf,unsigned long* result_crc)
{
   unsigned long calculate_crc=0;
   int err=ZIP_OK;
   FILE * fin = fopen(filenameinzip,"rb");
   unsigned long size_read = 0;
   unsigned long total_read = 0;
   if (fin==NULL)
   {
       err = ZIP_ERRNO;
   }

    if (err == ZIP_OK)
        do
        {
            err = ZIP_OK;
            size_read = (int)fread(buf,1,size_buf,fin);
            if (size_read < size_buf)
                if (feof(fin)==0)
            {
								ddocDebug(3, "getFileCrc", "error in reading %s\n",filenameinzip);
                err = ZIP_ERRNO;
            }

            if (size_read>0)
                calculate_crc = crc32(calculate_crc,buf,size_read);
            total_read += size_read;

        } while ((err == ZIP_OK) && (size_read>0));

    if (fin)
        fclose(fin);

    *result_crc=calculate_crc;
		ddocDebug(6, "getFileCrc","file %s crc %x\n",filenameinzip,calculate_crc);
    return err;
}

void change_file_date(filename,dosdate,tmu_date)
    const char *filename;
    uLong dosdate;
    tm_unz tmu_date;
{
#ifdef WIN32
  HANDLE hFile;
  FILETIME ftm,ftLocal,ftCreate,ftLastAcc,ftLastWrite;

  hFile = CreateFile(filename,GENERIC_READ | GENERIC_WRITE,
                      0,NULL,OPEN_EXISTING,0,NULL);
  GetFileTime(hFile,&ftCreate,&ftLastAcc,&ftLastWrite);
  DosDateTimeToFileTime((WORD)(dosdate>>16),(WORD)dosdate,&ftLocal);
  LocalFileTimeToFileTime(&ftLocal,&ftm);
  SetFileTime(hFile,&ftm,&ftLastAcc,&ftm);
  CloseHandle(hFile);
#else
#ifdef unix
  struct utimbuf ut;
  struct tm newdate;
  newdate.tm_sec = tmu_date.tm_sec;
  newdate.tm_min=tmu_date.tm_min;
  newdate.tm_hour=tmu_date.tm_hour;
  newdate.tm_mday=tmu_date.tm_mday;
  newdate.tm_mon=tmu_date.tm_mon;
  if (tmu_date.tm_year > 1900)
      newdate.tm_year=tmu_date.tm_year - 1900;
  else
      newdate.tm_year=tmu_date.tm_year ;
  newdate.tm_isdst=-1;

  ut.actime=ut.modtime=mktime(&newdate);
  utime(filename,&ut);
#endif
#endif
}


int mymkdir(dirname)
    const char* dirname;
{
    int ret=0;
#ifdef WIN32
    ret = mkdir(dirname);
#else
#ifdef unix
    ret = mkdir (dirname,0775);
#endif
#endif
    return ret;
}

int makedir (newdir)
    char *newdir;
{
  char *buffer ;
  char *p;
  int  len = (int)strlen(newdir);

  if (len <= 0)
    return 0;

  buffer = (char*)malloc(len+1);
  strcpy(buffer,newdir);

  if (buffer[len-1] == '/') {
    buffer[len-1] = '\0';
  }
  if (mymkdir(buffer) == 0)
    {
      free(buffer);
      return ZIP_ERRNO;
    }

  p = buffer+1;
  while (1)
    {
      char hold;

      while(*p && *p != '\\' && *p != '/')
        p++;
      hold = *p;
      *p = 0;
      if ((mymkdir(buffer) == -1) && (errno == ENOENT))
        {
					ddocDebug(4,"makedir", "couldn't create directory %s\n",buffer);
          free(buffer);
          return 0;
        }
      if (hold == 0)
        break;
      *p++ = hold;
    }
  free(buffer);
  return ZIP_ERRNO;
}

//--------------------------------------------------
// Deletes file from zip file
// zip_name - zip filename
// del_file - name of file,which will be deleted in zip
// returns error code or ERR_OK.
//--------------------------------------------------
EXP_OPTION int deleteFileFromZIP(const char* zip_name, const char* del_file)
{
	int some_was_del = 0;
	zipFile szip;
	zipFile dzip;
	int n_files = 0;
	int rv;
	unz_file_info unzfi;
	char fn[MAX_PATH];
	char* extrafield = NULL;
	char* commentary = NULL;
	int method;
	int level;
	int size_local_extra;
	void* local_extra = NULL;
	void* buf;
	zip_fileinfo zfi;
	zlib_filefunc_def ffunc;
	int err = 0;

	// change name for temp file
	char* tmp_name = (char*)malloc((strlen(zip_name) + 5)*sizeof(char));
	strcpy(tmp_name, zip_name);
	strncat(tmp_name, ".tmp", 5);

	ddocDebug(4, "deleteFileFromZIP", "Deleting %s from %s", del_file, zip_name);

	// open source and destination file
#     ifdef USEWIN32IOAPI
      fill_win32_filefunc(&ffunc);
			szip = unzOpen2(zip_name, &ffunc);
			if (szip==NULL) { free(tmp_name); return ERR_ZIP_FILE_READ; }
			dzip = zipOpen2(tmp_name, APPEND_STATUS_CREATE, NULL, &ffunc);
			if (dzip==NULL) { unzClose(szip); free(tmp_name); return ERR_ZIP_FILE_READ; }
#     else
			szip = unzOpen(zip_name);
			if (szip==NULL) { free(tmp_name); return ERR_ZIP_FILE_READ; }
			dzip = zipOpen(tmp_name, APPEND_STATUS_CREATE);
			if (dzip==NULL) { unzClose(szip); free(tmp_name); return ERR_ZIP_FILE_READ; }
#     endif

		//file not found
    if (unzLocateFile(szip,del_file,CASESENSITIVITY)!=UNZ_OK)
    {
			zipClose(dzip, NULL);
			unzClose(szip);
			_unlink(tmp_name);		
			free(tmp_name);
			ddocDebug(4, "deleteFileFromZIP","file %s not found in the zipfile\n",del_file);
      return UNZ_OK;
    }

	// copying files

	rv = unzGoToFirstFile(szip);
	while (rv == UNZ_OK)
	{
		// get zipped file info
		if (unzGetCurrentFileInfo(szip, &unzfi, fn, MAX_PATH, NULL, 0, NULL, 0) != UNZ_OK) 
		{
				err = unzGetCurrentFileInfo(szip, &unzfi, fn, MAX_PATH, NULL, 0, NULL, 0);
				ddocDebug(4, "deleteFileFromZIP", "unzGetCurrentFileInfo: %s and %d", fn, err);
				break;
		}

		// if not need delete this file
		ddocDebug(4, "deleteFileFromZIP", "Comparing %s and %s", fn, del_file);
		ddocDebug(3, "deleteFileFromZIP", "strcmp: %d", !strcmp(fn, del_file));
		if (!strcmp(fn, del_file) )
			some_was_del = 1;	
		else
		{
			extrafield = (char*)malloc(unzfi.size_file_extra);
			if ((extrafield==NULL)&&(unzfi.size_file_extra!=0)) break;
			commentary = (char*)malloc(unzfi.size_file_comment);
			if ((commentary==NULL)&&(unzfi.size_file_comment!=0)) {free(extrafield); break;}

			if (unzGetCurrentFileInfo(szip, &unzfi, fn, MAX_PATH, extrafield, unzfi.size_file_extra, commentary, unzfi.size_file_comment) != UNZ_OK) {free(extrafield); free(commentary); break;}

			// open file for RAW reading

			if (unzOpenCurrentFile3(szip, &method, &level, 1,NULL)!=UNZ_OK) {free(extrafield); free(commentary); break;}

			size_local_extra = unzGetLocalExtrafield(szip, NULL, 0);
			if (size_local_extra<0) {free(extrafield); free(commentary); break;}
			local_extra = malloc(size_local_extra);
			if ((local_extra==NULL)&&(size_local_extra!=0)) {free(extrafield); free(commentary); break;}
			if (unzGetLocalExtrafield(szip, local_extra, size_local_extra)<0) {free(extrafield); free(commentary); free(local_extra); break;}
			
			// this malloc may fail if file very large
			buf = malloc(unzfi.compressed_size);
			if ((buf==NULL)&&(unzfi.compressed_size!=0)) {free(extrafield); free(commentary); free(local_extra); break;}

			// open destination file
			memcpy (&zfi.tmz_date, &unzfi.tmu_date, sizeof(tm_unz));
			zfi.dosDate = unzfi.dosDate;
			zfi.internal_fa = unzfi.internal_fa;
			zfi.external_fa = unzfi.external_fa;
			
			if (zipOpenNewFileInZip3(dzip, fn, &zfi, NULL,0,NULL,0,NULL /* comment*/, method, level, 1,-MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY,NULL,0)!=UNZ_OK) {free(extrafield); free(commentary); free(local_extra); free(buf); break;}
			
			do
      {
        err = unzReadCurrentFile(szip,buf,WRITEBUFFERSIZE);
        if (err<0)
        {
					ddocDebug(4, "do_extract_currentfile", "error %d with zipfile in unzReadCurrentFile\n",err);
          break;
        }
        if (err>0)
          if (zipWriteInFileInZip(dzip, buf, err)!=UNZ_OK) {
						free(extrafield); free(commentary); free(local_extra); free(buf); break;}
        } while (err>0);

			if (zipCloseFileInZipRaw(dzip, unzfi.uncompressed_size, unzfi.crc)!=UNZ_OK) {free(extrafield); free(commentary); free(local_extra); free(buf); break;}

			if (unzCloseCurrentFile(szip)==UNZ_CRCERROR) {free(extrafield); free(commentary); free(local_extra); free(buf); break;}
			free(commentary);
			free(buf);
			free(extrafield);
			free(local_extra);

			n_files ++;
		}

		rv = unzGoToNextFile(szip);
	}
	ddocDebug(3, "deleteFileFromZIP", "rv: %d", rv);

	zipClose(dzip, NULL);
	unzClose(szip);
		
	// if fail
	if ( (!some_was_del) || (rv!=UNZ_END_OF_LIST_OF_FILE) )
	{
		_unlink(tmp_name);		
		free(tmp_name);
		return ERR_ZIP;
	}

	err = _unlink(zip_name);
	ddocDebug(4, "deleteFileFromZIP", "Remove: %s Error: %d", zip_name, err);
	err = rename(tmp_name, zip_name);
	ddocDebug(5, "deleteFileFromZIP", "Rename %s to %s Error: %d", tmp_name, zip_name, err);
	if (err != 0)
	{
		_unlink(tmp_name);	
		free(tmp_name);
		return ERR_ZIP;
	}

	// if all files were deleted
	if (n_files==0) _unlink(zip_name);

	_unlink(tmp_name);	
	free(tmp_name);
	return 0;
}


//--------------------------------------------------
// Add file to zip file
// filename_try - zip filename
// filename - filename in filesystem, not used if memBuf is specified
// filenameinzip - file name in zip
// opt_overwrite - append flag [CREATE (0), CREATEAFTER (1), ADDINZIP (2)]
// memBuf - memory buffer of data, which will be added to zip, should be NULL if filename is specified
// returns error code or ERR_OK.
//--------------------------------------------------
EXP_OPTION int addFileToZip(const char* filename_try, const char* filename, const char* filenameinzip, int opt_overwrite, char* memBuf)
{
    int opt_compress_level=Z_DEFAULT_COMPRESSION;
    int zipfilenamearg = 1;
    int err=0;
    int size_buf=0;
    void* buf=NULL;
    const char* password=NULL;
		FILE * fin = NULL;
		int size_read;
		zip_fileinfo zi;
		unsigned long crcFile=0;
		unzFile uf = NULL;
		char *filenameinzipBuf = 0;
#        ifdef USEWIN32IOAPI
        zlib_filefunc_def ffunc;
#        endif


#        ifdef USEWIN32IOAPI
        fill_win32_filefunc(&ffunc);
        uf = unzOpen2(filename_try,&ffunc);
#        else
        uf = unzOpen(filename_try);
#        endif
    
		if(uf){
			ddocDebug(4, "addFileToZip", "Adding %s to %s", filenameinzip, filename_try);
			if (unzLocateFile(uf,filenameinzip,CASESENSITIVITY)==UNZ_OK)
			{
				unzClose(uf);	
				ddocDebug(4, "addFileToZip","file %s found in the zipfile\n",filenameinzip);
				uf = 0;
				if(opt_overwrite < 2)
					return UNZ_OK;
				else
					deleteFileFromZIP(filename_try, filenameinzip);
			}
			if(uf)
			unzClose(uf);
		} else {
			opt_overwrite = 1;
		}
		
		size_buf = WRITEBUFFERSIZE;
    buf = (void*)malloc(size_buf);
    if (buf==NULL)
    {
        ddocDebug(4, "addFileToZip", "Error allocating memory\n");
        return ZIP_INTERNALERROR;
    }


    if (err==0)
    {
        zipFile zf = NULL;
        int errclose;
#        ifdef USEWIN32IOAPI
        zlib_filefunc_def ffunc;
        fill_win32_filefunc(&ffunc);
        zf = zipOpen2(filename_try,(opt_overwrite==2) ? 2 : 0,NULL,&ffunc);
#        else
        zf = zipOpen(filename_try,(opt_overwrite==2) ? 2 : 0);
#        endif

        if (zf == NULL)
        {
						ddocDebug(4,"addFileToZip", "error opening %s\n",filename_try, opt_overwrite);
            err= ZIP_ERRNO;
        }

		zi.tmz_date.tm_sec = zi.tmz_date.tm_min = zi.tmz_date.tm_hour =
		zi.tmz_date.tm_mday = zi.tmz_date.tm_mon = zi.tmz_date.tm_year = 0;
		zi.dosDate = 0;
		zi.internal_fa = 0;
		zi.external_fa = 0;
		filetime(filename,&zi.tmz_date,&zi.dosDate);


		if ((password != NULL) && (err==ZIP_OK))
			err = getFileCrc(filename,buf,size_buf,&crcFile);

		err = zipOpenNewFileInZip3(zf,filenameinzip,&zi,
						 NULL,0,NULL,0,NULL /* comment*/,
						 (opt_compress_level != 0) ? Z_DEFLATED : 0,
						 opt_compress_level,0,
						 -MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY,
						 password,crcFile);
		if (err != ZIP_OK)
			ddocDebug(4,"addFileToZip", "error in opening %s in the zipfile\n",filenameinzip);
		else if(filename == NULL)
		{
			size_read = strlen(memBuf);
			err = zipWriteInFileInZip (zf,memBuf, size_read);
			if (err<0)
			{
				ddocDebug(4,"addFileToZip", "error in writing %s in the zipfile\n",filenameinzip);
			}
		}
		else
		{
			fin = fopen(filename,"rb");
			if (fin==NULL)
			{
				err=ZIP_ERRNO;
				ddocDebug(4, "addFileToZip", "error in opening %s for reading\n",filenameinzip);
			}

		if (err == ZIP_OK)
			do
			{
				err = ZIP_OK;
				size_read = (int)fread(buf,1,size_buf,fin);
				if (size_read < size_buf)
					if (feof(fin)==0)
				{
					ddocDebug(4, "addFileToZip", "error in reading %s\n",filename);
					err = ZIP_ERRNO;
				}

				if (size_read>0)
				{
					err = zipWriteInFileInZip (zf,buf,size_read);
					if (err<0)
					{
						ddocDebug(4, "addFileToZip", "error in writing %s in the zipfile\n", filenameinzip);
					}

				}
			} while ((err == ZIP_OK) && (size_read>0));
		}

		if (fin)
			fclose(fin);

		if (err<0)
			err=ZIP_ERRNO;
		else
		{
			err = zipCloseFileInZip(zf);
			if (err!=ZIP_OK)
				ddocDebug(4, "addFileToZip", "error in closing %s in the zipfile\n",filenameinzip);
		}
 
    errclose = zipClose(zf,NULL);
    if (errclose != ZIP_OK)
			ddocDebug(4, "addFileToZip", "error in closing %s\n",filename_try);
  }

  free(buf);
  return 0;
}


//--------------------------------------------------
// Deletes all files, except datafiles.
// zip_name - zip filename
// tmp_name - new zip file, which contains only datafiles
// returns error code or ERR_OK.
//--------------------------------------------------
EXP_OPTION int makeDfZIP(const char* zip_name, const char* tmp_name)
{
	int some_was_del = 0;
	zipFile szip;
	zipFile dzip;
	int n_files = 0;
	int rv;
	unz_file_info unzfi;
	char fn[MAX_PATH];
	char* extrafield = NULL;
	char* commentary = NULL;
	int method;
	int level;
	int size_local_extra;
	void* local_extra = NULL;
	void* buf;
	zip_fileinfo zfi;
	zlib_filefunc_def ffunc;
	int err = 0;

	// open source and destination file
#     ifdef USEWIN32IOAPI
      fill_win32_filefunc(&ffunc);
			szip = unzOpen2(zip_name, &ffunc);
			if (szip==NULL) { return ERR_ZIP_FILE_READ; }
			dzip = zipOpen2(tmp_name, APPEND_STATUS_CREATE, NULL, &ffunc);
			if (dzip==NULL) { unzClose(szip); return ERR_ZIP_FILE_READ; }
#     else
			szip = unzOpen(zip_name);
			if (szip==NULL) { return ERR_ZIP_FILE_READ; }
			dzip = zipOpen(tmp_name, APPEND_STATUS_CREATE);
			if (dzip==NULL) { unzClose(szip); return ERR_ZIP_FILE_READ; }
#     endif

	//file not found
    if (unzLocateFile(szip,"mimetype",CASESENSITIVITY)!=UNZ_OK)
    {
			zipClose(dzip, NULL);
			unzClose(szip);
			ddocDebug(4, "deleteFileFromZIP","mimetype not found in the zipfile\n");
      return UNZ_OK;
    }

	// copying files

	rv = unzGoToFirstFile(szip);
	while (rv == UNZ_OK)
	{
		// get zipped file info
		if (unzGetCurrentFileInfo(szip, &unzfi, fn, MAX_PATH, NULL, 0, NULL, 0) != UNZ_OK) 
		{
				err = unzGetCurrentFileInfo(szip, &unzfi, fn, MAX_PATH, NULL, 0, NULL, 0);
				ddocDebug(4, "deleteFileFromZIP", "unzGetCurrentFileInfo: %s and %d", fn, err);
				break;
		}

		// if not need delete this file
		if (!strcmp(fn, "mimetype") || strstr(fn, "META-INF"))
			some_was_del = 1;	
		else
		{
			extrafield = (char*)malloc(unzfi.size_file_extra);
			if ((extrafield==NULL)&&(unzfi.size_file_extra!=0)) break;
			commentary = (char*)malloc(unzfi.size_file_comment);
			if ((commentary==NULL)&&(unzfi.size_file_comment!=0)) {free(extrafield); break;}

			if (unzGetCurrentFileInfo(szip, &unzfi, fn, MAX_PATH, extrafield, unzfi.size_file_extra, commentary, unzfi.size_file_comment) != UNZ_OK) {free(extrafield); free(commentary); break;}

			// open file for RAW reading

			if (unzOpenCurrentFile3(szip, &method, &level, 1,NULL)!=UNZ_OK) {free(extrafield); free(commentary); break;}

			size_local_extra = unzGetLocalExtrafield(szip, NULL, 0);
			if (size_local_extra<0) {free(extrafield); free(commentary); break;}
			local_extra = malloc(size_local_extra);
			if ((local_extra==NULL)&&(size_local_extra!=0)) {free(extrafield); free(commentary); break;}
			if (unzGetLocalExtrafield(szip, local_extra, size_local_extra)<0) {free(extrafield); free(commentary); free(local_extra); break;}
			
			// this malloc may fail if file very large
			buf = malloc(unzfi.compressed_size);
			if ((buf==NULL)&&(unzfi.compressed_size!=0)) {free(extrafield); free(commentary); free(local_extra); break;}

			// open destination file
			memcpy (&zfi.tmz_date, &unzfi.tmu_date, sizeof(tm_unz));
			zfi.dosDate = unzfi.dosDate;
			zfi.internal_fa = unzfi.internal_fa;
			zfi.external_fa = unzfi.external_fa;
			
			if (zipOpenNewFileInZip3(dzip, fn, &zfi, NULL,0,NULL,0,NULL /* comment*/, method, level, 1,-MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY,NULL,0)!=UNZ_OK) {free(extrafield); free(commentary); free(local_extra); free(buf); break;}
			
			do
      {
        err = unzReadCurrentFile(szip,buf,WRITEBUFFERSIZE);
        if (err<0)
        {
					ddocDebug(4, "do_extract_currentfile", "error %d with zipfile in unzReadCurrentFile\n",err);
          break;
        }
        if (err>0)
          if (zipWriteInFileInZip(dzip, buf, err)!=UNZ_OK) {
						free(extrafield); free(commentary); free(local_extra); free(buf); break;}
        } while (err>0);

			if (zipCloseFileInZipRaw(dzip, unzfi.uncompressed_size, unzfi.crc)!=UNZ_OK) {free(extrafield); free(commentary); free(local_extra); free(buf); break;}

			if (unzCloseCurrentFile(szip)==UNZ_CRCERROR) {free(extrafield); free(commentary); free(local_extra); free(buf); break;}
			free(commentary);
			free(buf);
			free(extrafield);
			free(local_extra);

			n_files ++;
		}

		rv = unzGoToNextFile(szip);
	}
	ddocDebug(3, "deleteFileFromZIP", "rv: %d", rv);

	zipClose(dzip, NULL);
	unzClose(szip);

	// if fail
	if ( (!some_was_del) || (rv!=UNZ_END_OF_LIST_OF_FILE) )
	{
		return ERR_ZIP;
	}

	if (err != 0)
	{
		return ERR_ZIP;
	}

	return 0;
}

//--------------------------------------------------
// Checks if zip file contains files not listed in fileNames list
// zip_name - zip filename
// fileNames - list of known files names separated by |
// returns error code or ERR_OK.
//--------------------------------------------------
EXP_OPTION int checkFilesInList(const char* zip_name, const char* fileNames)
{
	zipFile szip;
	int rv;
	unz_file_info unzfi;
	char fn[MAX_PATH];
	char* p1;
	zlib_filefunc_def ffunc;

	// open source and destination file
#     ifdef USEWIN32IOAPI
		fill_win32_filefunc(&ffunc);
		szip = unzOpen2(zip_name, &ffunc);
		if (szip==NULL) { return ERR_ZIP_FILE_READ; }
#     else
		szip = unzOpen(zip_name);
		if (szip==NULL) { return ERR_ZIP_FILE_READ; }
#     endif

	// checking files

	rv = unzGoToFirstFile(szip);
	while (rv == UNZ_OK) {
		// get zipped file info
		if(unzGetCurrentFileInfo(szip, &unzfi, fn, MAX_PATH, NULL, 0, NULL, 0) == UNZ_OK) 	{
			p1 = strstr(fileNames, fn);
			ddocDebug(4, "checkFilesInList", "File: %s exists: %s", fn, (p1 ? "OK" : "NO"));
			if(!p1)
				return ERR_DATAFILE_NOT_MANIFEST;
		}
		rv = unzGoToNextFile(szip);
	}
	unzClose(szip);

	return 0;
}


//--------------------------------------------------
// Deletes all files, except datafiles
// zip_name - zip filename
// returns error code or ERR_OK.
//--------------------------------------------------
EXP_OPTION int makeDfZIP2(const char* zip_name)
{
	int some_was_del = 0;
	zipFile szip;
	zipFile dzip;
	int n_files = 0;
	int rv;
	unz_file_info unzfi;
	char fn[MAX_PATH];
	char* extrafield = NULL;
	char* commentary = NULL;
	int method;
	int level;
	int size_local_extra;
	void* local_extra = NULL;
	void* buf;
	zip_fileinfo zfi;
	zlib_filefunc_def ffunc;
	int err = 0;

	// change name for temp file
	char* tmp_name = (char*)malloc((strlen(zip_name) + 5)*sizeof(char));
	strcpy(tmp_name, zip_name);
	strncat(tmp_name, ".tmp", 5);

	// open source and destination file
#     ifdef USEWIN32IOAPI
      fill_win32_filefunc(&ffunc);
			szip = unzOpen2(zip_name, &ffunc);
			if (szip==NULL) { free(tmp_name); return ERR_ZIP_FILE_READ; }
			dzip = zipOpen2(tmp_name, APPEND_STATUS_CREATE, NULL, &ffunc);
			if (dzip==NULL) { unzClose(szip); free(tmp_name); return ERR_ZIP_FILE_READ; }
#     else
			szip = unzOpen(zip_name);
			if (szip==NULL) { free(tmp_name); return ERR_ZIP_FILE_READ; }
			dzip = zipOpen(tmp_name, APPEND_STATUS_CREATE);
			if (dzip==NULL) { unzClose(szip); free(tmp_name); return ERR_ZIP_FILE_READ; }
#     endif

	// copying files
	rv = unzGoToFirstFile(szip);
	while (rv == UNZ_OK)
	{
		// get zipped file info
		if (unzGetCurrentFileInfo(szip, &unzfi, fn, MAX_PATH, NULL, 0, NULL, 0) != UNZ_OK) 
		{
				err = unzGetCurrentFileInfo(szip, &unzfi, fn, MAX_PATH, NULL, 0, NULL, 0);
				ddocDebug(4, "deleteFileFromZIP", "unzGetCurrentFileInfo: %s and %d", fn, err);
				break;
		}


		if (!strcmp(fn, "mimetype") || strstr(fn, "META-INF"))
			some_was_del = 1;	
		else
		{
			extrafield = (char*)malloc(unzfi.size_file_extra);
			if ((extrafield==NULL)&&(unzfi.size_file_extra!=0)) break;
			commentary = (char*)malloc(unzfi.size_file_comment);
			if ((commentary==NULL)&&(unzfi.size_file_comment!=0)) {free(extrafield); break;}

			if (unzGetCurrentFileInfo(szip, &unzfi, fn, MAX_PATH, extrafield, unzfi.size_file_extra, commentary, unzfi.size_file_comment) != UNZ_OK) {free(extrafield); free(commentary); break;}

			// open file for RAW reading

			if (unzOpenCurrentFile3(szip, &method, &level, 1,NULL)!=UNZ_OK) {free(extrafield); free(commentary); break;}

			size_local_extra = unzGetLocalExtrafield(szip, NULL, 0);
			if (size_local_extra<0) {free(extrafield); free(commentary); break;}
			local_extra = malloc(size_local_extra);
			if ((local_extra==NULL)&&(size_local_extra!=0)) {free(extrafield); free(commentary); break;}
			if (unzGetLocalExtrafield(szip, local_extra, size_local_extra)<0) {free(extrafield); free(commentary); free(local_extra); break;}
			
			// this malloc may fail if file very large
			buf = malloc(unzfi.compressed_size);
			if ((buf==NULL)&&(unzfi.compressed_size!=0)) {free(extrafield); free(commentary); free(local_extra); break;}

			// open destination file
			memcpy (&zfi.tmz_date, &unzfi.tmu_date, sizeof(tm_unz));
			zfi.dosDate = unzfi.dosDate;
			zfi.internal_fa = unzfi.internal_fa;
			zfi.external_fa = unzfi.external_fa;
	
			if (zipOpenNewFileInZip3(dzip, fn, &zfi, NULL,0,NULL,0,NULL /* comment*/, method, level, 1,-MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY,NULL,0)!=UNZ_OK) {free(extrafield); free(commentary); free(local_extra); free(buf); break;}
			
			do
      {
        err = unzReadCurrentFile(szip,buf,WRITEBUFFERSIZE);
        if (err<0)
        {
					ddocDebug(4, "do_extract_currentfile", "error %d with zipfile in unzReadCurrentFile\n",err);
          break;
        }
        if (err>0)
          if (zipWriteInFileInZip(dzip, buf, err)!=UNZ_OK) {
						free(extrafield); free(commentary); free(local_extra); free(buf); break;}
        } while (err>0);

			if (zipCloseFileInZipRaw(dzip, unzfi.uncompressed_size, unzfi.crc)!=UNZ_OK) {free(extrafield); free(commentary); free(local_extra); free(buf); break;}

			if (unzCloseCurrentFile(szip)==UNZ_CRCERROR) {free(extrafield); free(commentary); free(local_extra); free(buf); break;}
			free(commentary);
			free(buf);
			free(extrafield);
			free(local_extra);

			n_files ++;
		}

		rv = unzGoToNextFile(szip);
	}
	ddocDebug(3, "deleteFileFromZIP", "rv: %d", rv);

	zipClose(dzip, NULL);
	unzClose(szip);
		
	// if fail
	if ( (!some_was_del) || (rv!=UNZ_END_OF_LIST_OF_FILE) )
	{
		_unlink(tmp_name);		
		free(tmp_name);
		return ERR_ZIP;
	}

	err = _unlink(zip_name);
	ddocDebug(4, "deleteFileFromZIP", "Remove: %s Error: %d", zip_name, err);
	err = rename(tmp_name, zip_name);
	ddocDebug(5, "deleteFileFromZIP", "Rename %s to %s Error: %d", tmp_name, zip_name, err);
	if (err != 0)
	{
		_unlink(tmp_name);	
		free(tmp_name);
		return ERR_ZIP;
	}

	// if all files were deleted
	if (n_files==0) _unlink(zip_name);

	_unlink(tmp_name);	
	free(tmp_name);
	return 0;
}

int do_extract_currentfile(unzFile uf, int* popt_overwrite, const char* write_filename)
{
    char filename_inzip[256];
    char* p;
    int err=UNZ_OK;
    FILE *fout=NULL;
    void* buf;
    uInt size_buf;
		int skip=0;

    unz_file_info file_info;
    uLong ratio=0;
    err = unzGetCurrentFileInfo(uf,&file_info,filename_inzip,sizeof(filename_inzip),NULL,0,NULL,0);

    if (err!=UNZ_OK)
    {
				ddocDebug(4,"do_extract_currentfile", "error %d with zipfile in unzGetCurrentFileInfo\n",err);
        return err;
    }

    size_buf = WRITEBUFFERSIZE;
    buf = (void*)malloc(size_buf);
    if (buf==NULL)
    {
        return UNZ_INTERNALERROR;
    }

    p = filename_inzip;

    err = unzOpenCurrentFilePassword(uf,NULL);
    if (err!=UNZ_OK)
    {
			ddocDebug(4,"do_extract_currentfile", "error %d with zipfile in unzOpenCurrentFilePassword\n",err);
    }

    if (((*popt_overwrite)==0) && (err==UNZ_OK))
    {
        char rep=0;
        FILE* ftestexist;
        ftestexist = fopen(write_filename,"rb");
        if (ftestexist!=NULL)
        {
            fclose(ftestexist);
            skip = 1;
				}
    }

    if ((skip==0) && (err==UNZ_OK))
    {
        fout=fopen(write_filename,"wb");
        if (fout==NULL)
        {
           ddocDebug(4,"do_extract_currentfile", "error opening %s\n",write_filename);
        }
    }

    if (fout!=NULL)
    {
        ddocDebug(3, "do_extract_currentfile", " extracting: %s\n",write_filename);
        do
        {
            err = unzReadCurrentFile(uf,buf,size_buf);
            if (err<0)
            {
							ddocDebug(4, "do_extract_currentfile", "error %d with zipfile in unzReadCurrentFile\n",err);
              break;
            }
            if (err>0)
                if (fwrite(buf,err,1,fout)!=1)
                {
                    err=UNZ_ERRNO;
                    break;
                }
        }
        while (err>0);
        if (fout)
                fclose(fout);

        if (err==0)
            change_file_date(write_filename,file_info.dosDate,
                             file_info.tmu_date);
    }

    if (err==UNZ_OK)
    {
        err = unzCloseCurrentFile (uf);
        if (err!=UNZ_OK)
        {
					ddocDebug(4, "do_extract_currentfile","error %d with zipfile in unzCloseCurrentFile\n",err);
        }
    }
    else
        unzCloseCurrentFile(uf); /* don't lose the error */
		
		unzClose(uf);
    free(buf);
    return err;
}

//--------------------------------------------------
// Extracts file from zip
// zipfilename - zip filename
// filenameinzip - filename in zip
// opt_overwrite - overwrite flag
// filename - filename in filesystem
// returns error code or ERR_OK.
//--------------------------------------------------
EXP_OPTION int do_extract_onefile(const char* zipfilename,const char* filenameinzip, int opt_overwrite, const char* filename)
{   
		unzFile uf;
		int err = UNZ_OK;
#        ifdef USEWIN32IOAPI
        zlib_filefunc_def ffunc;
#        endif


#        ifdef USEWIN32IOAPI
        fill_win32_filefunc(&ffunc);
        uf = unzOpen2(zipfilename,&ffunc);
#        else
        uf = unzOpen(zipfilename);
#        endif
	

    if (unzLocateFile(uf,filenameinzip,CASESENSITIVITY)!=UNZ_OK)
    {
				ddocDebug(4, "do_extract_onefile","file %s not found in the zipfile\n",filenameinzip);
        unzClose(uf);
				return ERR_ZIP_FILE_READ;
    }
		
    if (do_extract_currentfile(uf, &opt_overwrite, filename) == UNZ_OK)
        return 0;
    else
        return ERR_ZIP;

}

//--------------------------------------------------
// Extracts file from zip into memory buffer
// zipfilename - zip filename
// filename - filename in zip
// pBuf - DigiDocMemBuf, where file content will be stored
// returns error code or ERR_OK.
//--------------------------------------------------
EXP_OPTION int do_extract_filemem(const char* zipfilename,const char* filename, DigiDocMemBuf* pBuf)
{   
		unzFile uf;
		int err = UNZ_OK;
#        ifdef USEWIN32IOAPI
        zlib_filefunc_def ffunc;
#        endif


#        ifdef USEWIN32IOAPI
        fill_win32_filefunc(&ffunc);
        uf = unzOpen2(zipfilename,&ffunc);
#        else
        uf = unzOpen(zipfilename);
#        endif

    if (unzLocateFile(uf,filename,CASESENSITIVITY)!=UNZ_OK)
    {
				ddocDebug(4, "do_extract_filemem","file %s not found in the zipfile\n",filename);
        unzClose(uf);
				return ERR_ZIP_FILE_READ;
    }
		
    if (do_extract_file(uf, pBuf) == UNZ_OK)
        return 0;
    else
        return ERR_ZIP;

}


int do_extract_file(unzFile uf, DigiDocMemBuf* pBuf)
{   
		int err = UNZ_OK;
    char filename_inzip[256];
    FILE *fout=NULL;
    void* buf;
    uInt size_buf;
    unz_file_info file_info;
    uLong ratio=0;
		int popt_extract_without_path = 0; 
		int popt_overwrite = 2;
		int i=0;

    err = unzGetCurrentFileInfo(uf,&file_info,filename_inzip,sizeof(filename_inzip),NULL,0,NULL,0);

    if (err!=UNZ_OK)
    {
			ddocDebug(4, "do_extract_file","error %d with zipfile in unzGetCurrentFileInfo\n",err);
      return err;
    }

    size_buf = WRITEBUFFERSIZE;
    buf = (void*)malloc(file_info.uncompressed_size);
    if (buf==NULL)
    {
        return UNZ_INTERNALERROR;
    }

        err = unzOpenCurrentFilePassword(uf,NULL);
        if (err!=UNZ_OK)
        {
					ddocDebug(4, "do_extract_file","error %d with zipfile in unzOpenCurrentFilePassword\n",err);
        }

        if (err==UNZ_OK)
        {
            do
            {
                err = unzReadCurrentFile(uf,buf,size_buf);
                if (err<0)
                {
                  ddocDebug(4, "do_extract_file","error %d with zipfile in unzReadCurrentFile\n",err);  
                  break;
                }
                if (err>0)
										ddocMemAppendData(pBuf, buf, err);
						}
            while (err>0);
        }
				
        if (err==UNZ_OK)
        {
            err = unzCloseCurrentFile (uf);
            if (err!=UNZ_OK)
            {
							ddocDebug(4, "do_extract_file","error %d with zipfile in unzCloseCurrentFile\n",err);  
            }
        }
        else
            unzCloseCurrentFile(uf); /* don't lose the error */
 
		unzClose(uf);
    free(buf);
    return err;
}

//--------------------------------------------------
// Calculates file size ja digest
// zipfilename - zip filename
// filename - filename in zip
// size - DigiDocMemBuf, where file size will be stored
// digest - buffer for digest
// returns error code or ERR_OK.
//--------------------------------------------------
EXP_OPTION int calculateFileDigestAndSizeZip(const char* zipfilename,const char* filename, DigiDocMemBuf* size, byte* digest)
{  
		unzFile uf;
		int err = UNZ_OK;

#        ifdef USEWIN32IOAPI
        zlib_filefunc_def ffunc;
#        endif


#        ifdef USEWIN32IOAPI
        fill_win32_filefunc(&ffunc);
        uf = unzOpen2(zipfilename,&ffunc);
#        else
        uf = unzOpen(zipfilename);
#        endif

    if (unzLocateFile(uf,filename,CASESENSITIVITY)!=UNZ_OK)
    {
				ddocDebug(4, "do_extract_filemem","file %s not found in the zipfile\n",filename);
        unzClose(uf);
				return ERR_ZIP_FILE_READ;
    }

    if (getSizeAndDigest(uf, size, digest) == UNZ_OK)
        err = UNZ_OK;
    else
        err = ERR_ZIP;
		unzClose(uf);
		return err;
}

int getSizeAndDigest(unzFile uf, DigiDocMemBuf* size, byte* digest)
{   
		int err = UNZ_OK;
    char filename_inzip[256];
    FILE *fout=NULL;
    void* buf;
    uInt size_buf;
    unz_file_info file_info;
    uLong ratio=0;
		int i=0;
		SHA_CTX ctx;
		char tmp[10];

    err = unzGetCurrentFileInfo(uf,&file_info,filename_inzip,sizeof(filename_inzip),NULL,0,NULL,0);
    if (err!=UNZ_OK)
    {
			ddocDebug(4, "getSizeAndDigest","error %d with zipfile in unzGetCurrentFileInfo\n",err);
      return err;
    }

    size_buf = WRITEBUFFERSIZE;
    buf = (void*)malloc(size_buf);
		snprintf(tmp, sizeof(tmp), "%d", (int)file_info.uncompressed_size);
		ddocMemAssignData(size, tmp,-1);
    if (buf==NULL)
    {
        return UNZ_INTERNALERROR;
    }

        err = unzOpenCurrentFilePassword(uf,NULL);
        if (err!=UNZ_OK)
        {
					ddocDebug(4, "getSizeAndDigest","error %d with zipfile in unzOpenCurrentFilePassword\n",err);
        }


        if (err==UNZ_OK)
        {
            memset(digest, 0, BDOC_DIGEST); // not sizeof(digest) which is 4 bytes!
			SHA1_Init(&ctx);
			do
            {
                err = unzReadCurrentFile(uf,buf,size_buf);
                if (err<0)
                {
					ddocDebug(4, "getSizeAndDigest","error %d with zipfile in unzReadCurrentFile\n",err);
                    break;
                }
                if (err>0)
					SHA1_Update(&ctx, buf, err);
			}
            while (err>0);
			SHA1_Final(digest,&ctx);
        }
				
        if (err==UNZ_OK)
        {
            err = unzCloseCurrentFile (uf);
            if (err!=UNZ_OK)
            {
							ddocDebug(4, "getSizeAndDigest","error %d with zipfile in unzCloseCurrentFile\n",err);
            }
        }
        else
            unzCloseCurrentFile(uf); /* don't lose the error */

			
    free(buf);
    return err;
}

//--------------------------------------------------
// Calculates file size ja digest
// zipfilename - zip filename
// filename - filename in zip
// method - SHA1 0, SHA256 1
// digest - buffer for digest
// returns error code or ERR_OK.
//--------------------------------------------------
EXP_OPTION int calculateFileDigestFromZip(const char* zipfilename,const char* filename, byte* digest, int method, int len)
{  
	// Cannot figure out why it breaks build on windows
	//ddocDebug(3, "calculateFileDigestFromZip", "Calc file: %s digest: %d", filename, method);
		unzFile uf;
		int err = UNZ_OK;

#        ifdef USEWIN32IOAPI
        zlib_filefunc_def ffunc;
#        endif


#        ifdef USEWIN32IOAPI
        fill_win32_filefunc(&ffunc);
        uf = unzOpen2(zipfilename,&ffunc);
#        else
        uf = unzOpen(zipfilename);
#        endif

    if(unzLocateFile(uf,filename,CASESENSITIVITY)!=UNZ_OK)
    {
		ddocDebug(4, "calculateFileDigestFromZip","file %s not found in the zipfile\n",filename);
        unzClose(uf);
		return ERR_ZIP_FILE_READ;
    }

    if (getDigest(uf, digest, method, len) == UNZ_OK)
        err = UNZ_OK;
    else
        err = ERR_ZIP;
		unzClose(uf);
		return err;
}

int getDigest(unzFile uf, byte* digest, int method, int len)
{   
	int err = UNZ_OK;
    char filename_inzip[256];
    FILE *fout=NULL;
    void* buf;
	uInt size_buf;
    unz_file_info file_info;
    uLong ratio=0;
	int i=0;
	SHA256_CTX ctx256;
	SHA_CTX ctx;

		

    err = unzGetCurrentFileInfo(uf,&file_info,filename_inzip,sizeof(filename_inzip),NULL,0,NULL,0);
    if (err!=UNZ_OK)
    {
			ddocDebug(4, "getSizeAndDigest","error %d with zipfile in unzGetCurrentFileInfo\n",err);
      return err;
    }

    size_buf = WRITEBUFFERSIZE;
		buf = (void*)malloc(size_buf);
    if (buf==NULL)
    {
        return UNZ_INTERNALERROR;
    }

        err = unzOpenCurrentFilePassword(uf,NULL);
        if (err!=UNZ_OK)
        {
			ddocDebug(4, "getSizeAndDigest","error %d with zipfile in unzOpenCurrentFilePassword\n",err);
        }


        if (err==UNZ_OK)
        {
            memset(digest, 0, len); // not sizeof(digest) which is 4 bytes!
			if(method == DIGEST_SHA256) {
				SHA256_Init(&ctx256);
			} else {
				SHA1_Init(&ctx);
			}
			do {
                err = unzReadCurrentFile(uf,buf,size_buf);
                if (err<0)
                {
					ddocDebug(4, "getSizeAndDigest","error %d with zipfile in unzReadCurrentFile\n",err);
                    break;
                }
                if (err>0) {
					if(method == DIGEST_SHA256)
						SHA256_Update(&ctx256, buf, err);
					else
						SHA1_Update(&ctx, buf, err);
				}
			}
            while (err>0);
			if(method == DIGEST_SHA256)
				SHA256_Final(digest,&ctx256);
			else
			SHA1_Final(digest,&ctx);
        }
				
        if (err==UNZ_OK)
        {
            err = unzCloseCurrentFile (uf);
            if (err!=UNZ_OK)
            {
				ddocDebug(4, "getSizeAndDigest","error %d with zipfile in unzCloseCurrentFile\n",err);
            }
        }
        else
            unzCloseCurrentFile(uf); /* don't lose the error */

			
    free(buf);
    return err;
}

//--------------------------------------------------
// Calculates file size ja digest
// zipfilename - zip filename
// filename - filename in zip
// returns file size
//--------------------------------------------------
EXP_OPTION int calculateFileSizeFromZip(const char* zipfilename,const char* filename)
{  
		unzFile uf;
		int err = UNZ_OK;
		//char tmp[10];
		int tmp;
		unz_file_info file_info;
		char filename_inzip[256];


#        ifdef USEWIN32IOAPI
        zlib_filefunc_def ffunc;
#        endif


#        ifdef USEWIN32IOAPI
        fill_win32_filefunc(&ffunc);
        uf = unzOpen2(zipfilename,&ffunc);
#        else
        uf = unzOpen(zipfilename);
#        endif

    if (unzLocateFile(uf,filename,CASESENSITIVITY)!=UNZ_OK)
    {
				ddocDebug(4, "do_extract_filemem","file %s not found in the zipfile\n",filename);
        unzClose(uf);
				return ERR_ZIP_FILE_READ;
    }


    err = unzGetCurrentFileInfo(uf,&file_info,filename_inzip,sizeof(filename_inzip),NULL,0,NULL,0);
    if (err!=UNZ_OK)
    {
			ddocDebug(4, "getSizeAndDigest","error %d with zipfile in unzGetCurrentFileInfo\n",err);
      return err;
    }

		tmp = file_info.uncompressed_size;
		unzClose(uf);
		return tmp;
}
